@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -14,53 +14,21 @@
 /* WARNING: This may not work in BSD, because it was written for Linux! */
 
 #include <stdio.h>
-#include <termios.h>
 #include <stdlib.h>
-#include <time.h>
-#include <string.h>
-#include <ctype.h>
-#include <sys/times.h>
-#include <sys/time.h>
 #include <sys/types.h>
-#if X_GRAPHICS
-#include <sys/mman.h>           /* root@sjoerd*/
-#endif /* X_GRAPHICS */
+#include <signal.h>
+#include <unistd.h>
+#include <asm/page.h>
 
 #include "emu.h"
-#include "bios.h"
-#include "mouse.h"
-#include "serial.h"
-#include "xms.h"
-#include "timers.h"
-#include "cmos.h"
-#include "memory.h"
-#include "termio.h"
-#include "config.h"
-#include "port.h"
 #include "int.h"
-#include "hgc.h"
-#include "dosio.h"
 
-#include "video.h"
 #if X_GRAPHICS
 #include "vgaemu.h" /* root@zaphod */
 #endif /* X_GRAPHICS */
 
-#include "pic.h"
-
 #include "dpmi.h"
 
-#ifdef USING_NET
-#include "ipx.h"
-#endif /* USING_NET */
-
-/* Needed for DIAMOND define */
-#include "vc.h"
-
-/* #include "sound.h" */
-
-#include "dma.h"
-
 
 /* Function prototypes */
 void print_exception_info(struct sigcontext_struct *scp);
@@ -86,6 +54,24 @@ int signal, struct sigcontext_struct *scp
   unsigned char *csp;
   int i;
 
+#if 0
+  _eflags &= ~(AC|ID);
+  REG(eflags) &= ~(AC|ID);
+#endif
+
+  if (fault_cnt > 1) {
+    error("Fault handler re-entered! signal=%i _trapno=0x%lX\n",
+      signal, _trapno);
+    if (!in_vm86 && _cs == UCODESEL) {
+      /* TODO - we can start gdb here */
+      /* start_gdb() */
+    } else {
+      error("BUG: Fault handler re-entered not within dosemu code! in_vm86=%i\n",
+        in_vm86);
+    }
+    goto bad;
+  }
+
   if (in_vm86) {
     in_vm86 = 0;
     switch (_trapno) {
@@ -95,6 +81,11 @@ int signal, struct sigcontext_struct *scp
       case 0x04: /* overflow */
       case 0x05: /* bounds */
       case 0x07: /* device_not_available */
+#ifdef TRACE_DPMI
+		 if (_trapno==1) {
+	           t_printf("\n%s",e_scp_disasm(scp,0));
+		 }
+#endif
 		 return (void) do_int(_trapno);
 
       case 0x10: /* coprocessor error */
@@ -102,17 +93,6 @@ int signal, struct sigcontext_struct *scp
 		 return;
 
       case 0x11: /* alignment check */
-		 /*
-		  * FIRST thing to do - to avoid being trapped into int0x11
-		  * forever, we must clear AC before doing anything else!
-		  */
-		 __asm__ __volatile__ ("
-			pushfl
-			popl	%%eax
-			andl	$0xfffbffff,%%eax
-			pushl	%%eax
-			popfl"
-			: : : "%eax");
 		 /* we are now safe; nevertheless, fall into the default
 		  * case and exit dosemu, as an AC fault in vm86 is(?) a
 		  * catastrophic failure.
@@ -152,7 +132,7 @@ int signal, struct sigcontext_struct *scp
       case 0x0e:
                 if(config.X)
                   {
-                    if(VGA_EMU_FAULT(scp,code)==True)
+                    if(VGA_EMU_FAULT(scp,code,0)==True)
                       return;
                   }
                 /* fall into default case if not X */
@@ -171,75 +151,127 @@ sgleave:
 
 		 print_exception_info(scp);
 #else
-		 error("unexpected CPU exception 0x%02lx errorcode: 0x%08lx while in vm86 (DOS)\n",
-	  	 _trapno, _err);
+		 error("unexpected CPU exception 0x%02lx err=0x%08lx cr2=%08lx while in vm86 (DOS)\n",
+	  	 _trapno, _err, _cr2);
 		{
-		  extern FILE *dbg_fd;
-		  int auxg = d.general;
+		  int auxg = debug_level('g');
 		  FILE *aux = dbg_fd;
 		  flush_log();  /* important! else we flush to stderr */
 		  dbg_fd = stderr;
-		  d.general =1;
+		  set_debug_level('g',1);
 		  show_regs(__FILE__, __LINE__);
-		  d.general = auxg;
+		  set_debug_level('g', auxg);
 		  flush_log();
 		  dbg_fd = aux;
 		}
 #endif
 
  		 show_regs(__FILE__, __LINE__);
-		 if (d.network)		/* XXX */
+		 if (debug_level('n'))		/* XXX */
 		     abort();
+		 flush_log();
  		 leavedos(4);
     }
   }
-
-
-#if X_GRAPHICS				/* only for debug ?*/ /*root@sjoerd*/
-/* The reason it comes here instead of inside_VM86 is
- * char_out() in ./video/int10.c (and the memory is protected).
- * We want to protect the video memory and the VGA BIOS.
- * This function is called from the following functions:
+#define VGA_ACCESS_HACK 1
+#if VGA_ACCESS_HACK
+#if X_GRAPHICS
+  if(_trapno==0x0e && config.X && _cs==UCODESEL) {
+/* Well, there are currently some dosemu functions that touches video memory
+ * without checking the permissions. This is a VERY BIG BUG.
+ * Must be fixed ASAP.
+ * Known offensive functions are:
  * dosemu/utilities.c:     char_out(*s++, READ_BYTE(BIOS_CURRENT_SCREEN_PAGE));
  * video/int10.c:    char_out(*(char *) &REG(eax), READ_BYTE(BIOS_CURRENT_SCREEN_PAGE));
+ * EMS and XMS memory transfer functions may also touch video mem.
+ *  but if only the protection needs to be adjusted (no instructions emulated)
+ *  we should be able to handle it in DOSEMU 
  */
+    v_printf("BUG: dosemu touched protected video mem, but trying to recover\n");
+    if(VGA_EMU_FAULT(scp,code,1)==True)
+      return;
+  }
+#endif
+#endif
 
-  if(_trapno==0x0e)
-    if(config.X)
-      {
-/*
-        printf("ERROR: cpu exception in dosemu code outside of VM86()!\n"
-               "trapno: 0x%02lx  errorcode: 0x%08lx  cr2: 0x%08lx\n"
-               "eip: 0x%08lx  esp: 0x%08lx  eflags: 0x%08lx\n"
-               "cs: 0x%04x  ds: 0x%04x  es: 0x%04x  ss: 0x%04x\n",
-               _trapno, scp->err, scp->cr2,
-               _eip, _esp, _eflags, _cs, _ds, _es, _ss);
-*/
-        if(VGA_EMU_FAULT(scp,code)==True)
+  if (in_dpmi) {
+    /* At first let's find out where we came from */
+    if (_cs==UCODESEL) {
+      /* Fault in dosemu code */
+      /* Now see if it is HLT */
+      csp = (unsigned char *) SEL_ADR(_cs, _eip);
+      if (*csp == 0xf4) {
+	/* Well, must come from dpmi_control() */
+        _eip += 1;
+        /* Note: when using DIRECT_DPMI_CONTEXT_SWITCH, we only come
+         * here if we have set the trap-flags (TF)
+         * ( needed for dosdebug only )
+         */
+	indirect_dpmi_switch(scp);
+	return;
+      }
+      else { /* No, not HLT, too bad :( */
+	error("Fault in dosemu code, in_dpmi=%i\n", in_dpmi);
+        /* TODO - we can start gdb here */
+        /* start_gdb() */
+
+	/* Going to die from here */
+	goto bad;	/* well, this goto is unnecessary but I like gotos:) */
+      }
+    } /*_cs==UCODESEL*/
+    else {
+    /* Not in dosemu code */
+
+    /* If this is an exception 0x11, we have to ignore it. The reason is that
+     * under real DOS the AM bit of CR0 is not set.
+     * Also clear the AC flag to prevent it from re-occuring.
+     */
+     if (_trapno == 0x11) {
+       g_printf("Exception 0x11 occured, clearing AC\n");
+       _eflags &= ~AC;
+       return;
+     }
+
+#if X_GRAPHICS
+      if(_trapno==0x0e && config.X) {
+        if(VGA_EMU_FAULT(scp,code,1)==True) {
+	  if (dpmi_eflags & VIP) {
+	    dpmi_sigio(scp);
+	  }
           return;
+	}
       }
 #endif /* X_GRAPHICS */
 
+      /* dpmi_fault() will handle that */
+      dpmi_fault(scp);
+      if (_cs==UCODESEL) {
+        /* context was switched to dosemu's, return ASAP */
+        return;
+      }
 
-  if (in_dpmi)
-#ifdef __linux__
-    return dpmi_fault(scp);
-#endif /* __linux__ */
+      CheckSelectors(scp);
+      /* now we are safe */
+      return;
+    }
+  } /*in_dpmi*/
 
-  csp = (char *) _eip;
-
-  /* This has been added temporarily as most illegal sigsegv's are attempt
-     to call Linux int routines */
+bad:
+/* All recovery attempts failed, going to die :( */
 
 #if 0
+  csp = (char *) _eip;
+  /* This was added temporarily as most illegal sigsegv's are attempt
+     to call Linux int routines */
   if (!(csp[-2] == 0xcd && csp[-1] == 0x80 && csp[0] == 0x85)) {
 #else
   {
 #endif /* 0 */
-    error("cpu exception in dosemu code outside of VM86()!\n"
+    error("cpu exception in dosemu code outside of %s!\n"
 	  "trapno: 0x%02lx  errorcode: 0x%08lx  cr2: 0x%08lx\n"
 	  "eip: 0x%08lx  esp: 0x%08lx  eflags: 0x%08lx\n"
 	  "cs: 0x%04x  ds: 0x%04x  es: 0x%04x  ss: 0x%04x\n",
+	  (in_dpmi ? "DPMI client" : "VM86()"),
 	  _trapno, _err, _cr2,
 	  _eip, _esp, _eflags, _cs, _ds, _es, _ss);
 
@@ -281,7 +313,7 @@ sgleave:
     /* display the 10 bytes before and after CS:EIP.  the -> points
      * to the byte at address CS:EIP
      */
-    dbug_printf("OPS  : ");
+    dbug_printf("OOPS : ");
     csp = (unsigned char *) _eip - 10;
     for (i = 0; i < 10; i++)
       dbug_printf("%02x ", *csp++);
@@ -300,7 +332,27 @@ sgleave:
 #ifdef __linux__
 void dosemu_fault(int signal, struct sigcontext_struct context)
 {
-    dosemu_fault1 (signal, &context);
+  restore_eflags_fs_gs();
+  fault_cnt++;
+
+  if (fault_cnt > 2) {
+   /*
+    * At this point we already tried leavedos(). Now try _exit()
+    * and NOT exit(3), because glibc is probably malfunctions if
+    * we are here.
+    */
+    _exit(255);
+  }
+
+  if (debug_level('g')>7)
+    g_printf("Entering fault handler, signal=%i _trapno=0x%lX\n",
+      signal, context.trapno);
+
+  dosemu_fault1 (signal, &context);
+
+  if (debug_level('g')>8)
+    g_printf("Returning from the fault handler\n");
+  fault_cnt--;
 }
 #endif /* __linux__ */
 
@@ -461,14 +513,39 @@ void print_exception_info(struct sigcontext_struct *scp)
 
       error("@Exception was caused by ");
       if(_err & 0x01)
-	error("@insufficient privelege\n");
+	error("@insufficient privilege\n");
       else
 	error("@non-available page\n");
       break;
 
-   case 0x10:
-      error ("Coprocessor Error:\n");
-      break;
+   case 0x10: {
+      struct _fpstate *p = scp->fpstate;
+      int i, n;
+      error ("@Coprocessor Error:\n");
+      error ("@cw=%04x sw=%04x tag=%04x\n",
+	*((unsigned short *)&(p->cw)),*((unsigned short *)&(p->sw)),
+	*((unsigned short *)&(p->tag)));
+      n = (p->sw >> 11) & 7;
+      error ("@cs:eip=%04x:%08lx ds:data=%04x:%08lx\n",
+	*((unsigned short *)&(p->cssel)),p->ipoff,
+	*((unsigned short *)&(p->datasel)),p->dataoff);
+      if ((p->sw&0x80)==0) error("@No error summary bit,why?\n");
+      else {
+	if (p->sw&0x20) error("@Precision\n");
+	if (p->sw&0x10) error("@Underflow\n");
+	if (p->sw&0x08) error("@Overflow\n");
+	if (p->sw&0x04) error("@Divide by 0\n");
+	if (p->sw&0x02) error("@Denormalized\n");
+	if ((p->sw&0x41)==0x01) error("@Invalid op\n");
+	  else if ((p->sw&0x41)==0x41) error("@Stack fault\n");
+      }
+      for (i=0; i<8; i++) {
+	unsigned long long *r = (unsigned long long *)&(p->_st[i].significand);
+	unsigned short *e = (unsigned short *)&(p->_st[i].exponent);
+	error ("@fpr[%d] = %04x:%016Lx\n",n,*e,*r);
+	n = (n+1) & 7;
+      }
+      } break;
 
     default:
       error("@Unknown exception\n");
