@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -9,12 +9,14 @@
 #ifndef _EMULATION_H
 #define _EMULATION_H
 
-#include"midid.h"
+#include "device.h"
 
 /* Current instrument mapping needed for emulation */
-int *imap;
+extern int *imap;
+
+extern int imap_default[128], imap_mt2gm[128];
 
 /* Sets emulation mode to new_mode */
-void emulation_set(Emumode new_mode);
+void emulation_set(Device *dev, Emumode new_mode);
 
 #endif
