@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -8,7 +8,7 @@
   null (flush) device -- mainly for testing
  ***********************************************************************/
 
-#include"midid.h"
+#include "device.h"
 
 bool null_detect(void)
 {
@@ -53,5 +53,6 @@ void register_null(Device * dev)
 	dev->channelpressure = null_doall2;
 	dev->bender = null_doall2;
 	dev->program = null_doall2;
+	dev->setmode = null_setmode;
 }
 
