@@ -1,27 +1,23 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
 
-#include <features.h>
 #include <stdio.h>
 #include <string.h>
 #include <termios.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <sys/times.h>
 #include <sys/time.h>
 #include <sys/stat.h>
 #include <sys/utsname.h>
 #include <errno.h>
-#include <sys/utsname.h>
 
 #include "config.h"
 #include "emu.h"
 #include "memory.h"
-#include "config.h"
 #include "bios.h"
 #include "int.h"
 #include "timers.h"
@@ -30,6 +26,7 @@
 #include "vc.h"
 #include "mouse.h"
 #include "port.h"
+#include "joystick.h"
 #ifdef USING_NET
 #include "pktdrvr.h"
 #include "ipx.h"
@@ -46,16 +43,11 @@
 #include "speaker.h"
 
 #include "keyb_clients.h"
+#include "keyb_server.h"
 
 #include "mapping.h"
 
 
-#ifdef USING_NET
-extern void pkt_check_receive_quick(void);
-/* flag to activate use of pic by packet driver */
-#define PICPKT 1
-#endif
-
 #if 0
 static inline void dbug_dumpivec(void)
 {
@@ -106,6 +98,8 @@ dosemu_banner(void)
  */
 void stdio_init(void)
 {
+  setbuf(stdout, NULL);
+
   if(dbg_fd) {
     warn("DBG_FD already set\n");
     return;
@@ -119,6 +113,8 @@ void stdio_init(void)
 	      config.debugout);
       exit(1);
     }
+    free(config.debugout);
+    config.debugout = NULL;
   }
   else
   {
@@ -126,7 +122,6 @@ void stdio_init(void)
     warn("No debug output file specified, debugging information will not be printed");
   }
   sync();  /* for safety */
-  setbuf(stdout, NULL);
 }
 
 /*
@@ -155,10 +150,7 @@ void timer_interrupt_init(void)
   struct itimerval itv;
   int delta;
 
-  /* DANG_FIXTHIS config.timer is undocumented and probably obsolete
-   */
   delta = (config.update / TIMER_DIVISOR);
-  config.wantdelta = delta;	/* save it */
   /* Check that the kernel actually supports such a frequency - we
    * can't go faster than jiffies with setitimer()
    */
@@ -167,7 +159,6 @@ void timer_interrupt_init(void)
     config.update = 54925; config.freq = 18;
     delta = 54925 / TIMER_DIVISOR;
   }
-  config.realdelta = delta;
 
   itv.it_interval.tv_sec = 0;
   itv.it_interval.tv_usec = delta;
@@ -175,7 +166,7 @@ void timer_interrupt_init(void)
   itv.it_value.tv_usec = delta;
   c_printf("TIME: using %d usec for updating ALRM timer\n", delta);
 
-  setitimer(TIMER_TIME, &itv, NULL);
+  setitimer(ITIMER_REAL, &itv, NULL);
 }
 
 /*
@@ -188,37 +179,27 @@ void timer_interrupt_init(void)
  */
 void hardware_setup(void)
 {
-  extern void  do_irq1(void);
-  extern void rtc_int8(void);
   /* PIC init */
-  pic_seti(PIC_IRQ0, timer_int_engine, 0);  /* do_irq0 in pic.c */
+  pic_seti(PIC_IRQ0, timer_int_engine, 0, NULL);  /* do_irq0 in pic.c */
   pic_unmaski(PIC_IRQ0);
   pic_request(PIC_IRQ0);  /* start timer */
-  pic_seti(PIC_IRQ1, do_irq1, 0); /* do_irq1 in dosio.c   */
+  pic_seti(PIC_IRQ1, do_irq1, 0, NULL); /* do_irq1 in dosio.c   */
   pic_unmaski(PIC_IRQ1);
-  pic_seti(PIC_IRQ8, rtc_int8, 0);
+  pic_seti(PIC_IRQ8, rtc_int8, 0, NULL);
   pic_unmaski(PIC_IRQ8);
-  if (mice->intdrv || mice->type == MOUSE_PS2 || mice->type == MOUSE_IMPS2) {
-    pic_seti(PIC_IMOUSE, DOSEMUMouseEvents, 0);
-    pic_unmaski(PIC_IMOUSE);
-  }
 #ifdef USING_NET
 #ifdef IPX
-  pic_seti(PIC_IPX, IPXCallRel, 0);
+  pic_seti(PIC_IPX, do_irq, 0, IPXCallRel);
   pic_unmaski(PIC_IPX);
 #endif
-#ifdef PICPKT
-  pic_seti(PIC_NET, pkt_check_receive_quick, 0x61);
-#else
-  pic_seti(PIC_NET, pkt_check_receive_quick, 0);
-#endif
+  pic_seti(PIC_NET, pkt_check_receive_quick, 0, NULL);
   pic_unmaski(PIC_NET);
 #endif
   
   /* DMA Init */
   /* dma_init(); - Currently in dev_list */
 
-  g_printf("PIC,mouse,IPX initialized\n");
+  g_printf("PIC,IPX initialized\n");
 }
 
 /*
@@ -229,9 +210,11 @@ void hardware_setup(void)
  * 
  * DANG_END_FUNCTION
  */
-static inline void map_video_bios(void)
+void map_video_bios(void)
 {
-  extern int load_file(char *name, int foffset, char *mstart, int msize);
+  uint32_t int_area[256];
+  uint16_t seg, off;
+  int i;
 
   v_printf("Mapping VBIOS = %d\n",config.mapped_bios);
 
@@ -256,6 +239,25 @@ static inline void map_video_bios(void)
 
     memcheck_addtype('V', "Video BIOS");
     memcheck_reserve('V', VBIOS_START, VBIOS_SIZE);
+    video_ints[0x1f] = 1;
+    video_ints[0x43] = 1;    
+    if (!config.vbios_post) {
+      load_file("/dev/mem", 0, (char *)int_area, sizeof(int_area));
+      for (i = 0; i < 256; i++) {
+        seg = int_area[i] >> 16;
+        off = int_area[i] & 0xffff;
+        g_printf("int0x%x was 0x%04x:0x%04x\n", i, seg, off);
+        if (seg == VBIOS_START >> 4) {
+          g_printf("Setting int0x%x to 0x%04x:0x%04x\n", i, seg, off);
+          SETIVEC(i, seg, off);
+          video_ints[i] = 1;
+        } else video_ints[i] = 0;
+      }
+      g_printf("Now initialising 0x40:49-66, 84-8a and a8-ab\n");
+      load_file("/dev/mem", (0x40 << 4) + 0x49, (char *)(0x40 << 4) + 0x49, 0x67-0x49);
+      load_file("/dev/mem", (0x40 << 4) + 0x84, (char *)(0x40 << 4) + 0x84, 0x8b-0x84);
+      load_file("/dev/mem", (0x40 << 4) + 0xa8, (char *)(0x40 << 4) + 0xa8, 0xac-0xa8);
+    }
   }
 }
 
@@ -267,12 +269,11 @@ static inline void map_video_bios(void)
  * 
  * DANG_END_FUNCTION
  */
-static inline void map_hardware_ram(void)
+void map_hardware_ram(void)
 {
   int i, j;
   unsigned int addr, size;
 
-  open_mapping(MAPPING_INIT_HWRAM);
   if (!config.must_spare_hardware_ram)
     return;
   if (!can_do_root_stuff) {
@@ -333,15 +334,21 @@ static inline void bios_mem_setup(void)
 {
   int b;
 
-  /* show 0 serial ports and 3 parallel ports, maybe a mouse, and the
+  /* show 0 serial ports and 3 parallel ports, maybe a mouse, game card and the
    * configured number of floppy disks
    */
   CONF_NFLOP(configuration, config.fdisks);
   CONF_NSER(configuration, config.num_ser);
   CONF_NLPT(configuration, config.num_lpt);
-  if ((mice->type == MOUSE_PS2) || (mice->type == MOUSE_IMPS2) || (mice->intdrv))
+  if (config.mouse.intdrv)
     configuration |= CONF_MOUSE;
 
+  /* 
+   * real BIOS' don't seem to set this bit but setting it can't do any harm
+   * and improves joystick detection for the game, Alley Cat, anyway :)
+   */
+  if (joy_exist ()) configuration |= CONF_GAME;
+
   if (config.mathco)
     configuration |= CONF_MATHCO;
 
@@ -370,15 +377,11 @@ void memory_init(void)
   /* first_call is a truly awful hack to keep some of these procedures from
      being called twice.  It should be fixed sometime. */
 
-  map_video_bios();            /* map the video bios */
-  if (first_call)
-    map_hardware_ram();        /* map the direct hardware ram */
   map_custom_bios();           /* map the DOSEMU bios */
   setup_interrupts();          /* setup interrupts */
 
   {
     /* update boot drive in Banner-code */
-    extern void bios_f000_bootdrive(), bios_f000();
     u_char *ptr;
 
     ptr = (u_char *)((BIOSSEG << 4) + ((long)bios_f000_bootdrive - (long)bios_f000));
@@ -391,7 +394,8 @@ void memory_init(void)
 #endif
 
 #ifdef USING_NET
-  pkt_init(0x60);              /* Install the new packet driver interface */
+  if (config.pktdrv)
+    pkt_init(0x60);              /* Install the new packet driver interface */
 #endif
 
   bios_mem_setup();            /* setup values in BIOS area */
@@ -410,11 +414,6 @@ void memory_init(void)
     init_vga_card();
   }
 
-  if (config.exitearly) {
-    dbug_printf("Leaving DOS before booting\n");
-    leavedos(0);
-  }
-
   if (first_call) {
     ems_init();                /* initialize ems */
     xms_init();                /* initialize xms */
@@ -434,30 +433,32 @@ void memory_init(void)
  */
 void device_init(void)
 {
-  extern void check_console(void);
+  hardware_setup();		/* setup any hardware */
+
   /* check whether we are running on the console */
   check_console();
 
   scr_state_init();
+#ifdef HAVE_KEYBOARD_V1
   if (!keyb_server_init()) {
     error("can't init keyboard server\n");
-    leavedos(19);
+    config.exitearly = 1;
   }
   if (!keyb_client_init()) {
     error("can't open keyboard client\n");
-    leavedos(19);
+    config.exitearly = 1;
   }
+#endif
    
   if (!config.vga)
-    config.allowvideoportaccess = 0;
- 
+    config.vbios_post = 0;
+
   video_config_init();
   if (config.console && (config.speaker == SPKR_EMULATED)) {
     register_speaker((void *)console_fd,
 		     console_speaker_on, console_speaker_off);
   }
   iodev_init();
-  mouse_init();
   printer_init();
   disk_init();
 }
@@ -474,17 +475,14 @@ void low_mem_init(void)
 {
   char *result;
 
-  open_mapping(MAPPING_INIT_LOWRAM);
   g_printf ("DOS memory area being mapped in\n");
   result = mmap_mapping(MAPPING_INIT_LOWRAM | MAPPING_SCRATCH, 0,
   		0x100000, PROT_EXEC | PROT_READ | PROT_WRITE, 0);
   if (result != NULL)
     {
       perror ("anonymous mmap");
-      leavedos (1);
+      config.exitearly = 1;
     }
-
-  HMA_init();
 }
 
 /*
@@ -505,7 +503,6 @@ void version_init(void) {
   running_kversion = atoi(strtok(version,".")) *1000000;
   running_kversion += atoi(strtok(NULL,".")) *1000;
   running_kversion += atoi(strtok(NULL,"."));
-  use_sigio=FASYNC;
 #endif
   
 
@@ -516,7 +513,6 @@ void version_init(void) {
   if (unames.release[0] > 0 ) {
     if ((unames.release[2] == 1  && unames.release[3] > 1 ) ||
          unames.release[2] > 1 ) {
-      use_sigio=FASYNC;
     }
   }
 #endif
