@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -14,7 +14,7 @@
 #include <unistd.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <malloc.h>
+#include <stdlib.h>
 #include <ctype.h>
 #include <string.h>
 #include <sys/ioctl.h>
@@ -28,26 +28,24 @@
 
 #include "emu.h"
 #include "disks.h"
+#include "timers.h"
+#include "utilities.h"
 #include "priv.h"
-#include "dosemu_select.h"
 #include "int.h"
 #include "fatfs.h"
 #include "utilities.h"
-
-#ifdef NEED_LLSEEK_PROTOTYPE
-  /* well, if we don't have llseek prototype,
-   * we most likely won't have __loff_t too, hence using long long
-   */
-  extern long long libless_llseek(int fd, long long offset, int origin);
-  #define llseek libless_llseek
+#ifdef X86_EMULATOR
+#include "cpu-emu.h"
 #endif
 
+#define llseek libless_llseek
+
 static int disks_initiated = 0;
 
 #define FDISKS config.fdisks
 #define HDISKS config.hdisks
 
-inline void disk_close(void);
+static void set_part_ent(struct disk *dp, char *tmp_mbr);
 
 #define USE_FSYNC 1
 
@@ -143,7 +141,7 @@ read_sectors(struct disk *dp, char *buffer, long head, long sector,
   d_printf("DISK: %s: Trying to read %ld sectors at T/S/H %ld/%ld/%ld",
 	   dp->dev_name,count,track,sector,head);
 #ifdef __linux__
-  d_printf("%+Ld at pos %Ld\n", dp->header, pos);
+  d_printf("%+lld at pos %lld\n", dp->header, pos);
 #else
   d_printf(" at pos %ld\n", pos);
 #endif
@@ -215,14 +213,20 @@ read_sectors(struct disk *dp, char *buffer, long head, long sector,
       error("Sector not found in read_sector, error = %s!\n", strerror(errno));
       return -DERR_NOTFOUND;
     }
+#ifdef X86_EMULATOR
+    if (config.cpuemu>1)
+	tmpread = e_dos_read(dp->fdesc, buffer, count * SECTOR_SIZE - already);
+    else
+#endif
     tmpread = RPT_SYSCALL(read(dp->fdesc, buffer, count * SECTOR_SIZE - already));
   }
 
   if(tmpread != -1) {
-    if(d.disk > 8) dump_disk_blks(buffer, count - already / SECTOR_SIZE, SECTOR_SIZE);
+    if (debug_level('d') > 8) dump_disk_blks(buffer, count - already / SECTOR_SIZE, SECTOR_SIZE);
     return tmpread + already;
   }
   else {
+/**/ dbug_printf("disks.c: read failed\n");
     return -DERR_ECCERR;
   }
 }
@@ -249,7 +253,7 @@ write_sectors(struct disk *dp, char *buffer, long head, long sector,
   d_printf("DISK: %s: Trying to write %ld sectors T/S/H %ld/%ld/%ld",
 	   dp->dev_name,count,track,sector,head);
 #ifdef __linux__
-  d_printf(" at pos %Ld\n", pos);
+  d_printf(" at pos %lld\n", pos);
 #else
   d_printf(" at pos %ld\n", pos);
 #endif
@@ -468,8 +472,6 @@ void dir_setup(struct disk *dp)
 
   dp->fatfs = NULL;
   fatfs_init(dp);
-
-  if(dp->fatfs && !redir_state) set_int21_revectored(redir_state = 1);
 }
 
 /* XXX - relies upon a file of SECTOR_SIZE in PARTITION_PATH that which
@@ -484,11 +486,9 @@ void dir_setup(struct disk *dp)
 void
 partition_setup(struct disk *dp)
 {
-  PRIV_SAVE_AREA
   int part_fd, i;
   unsigned char tmp_mbr[SECTOR_SIZE];
   char *hd_name;
-  void set_part_ent(struct disk *, char *);
 
 #define PART_BYTE(p,b)  *((unsigned char *)tmp_mbr + PART_INFO_START + \
 			  (PART_INFO_LEN * (p-1)) + b)
@@ -505,19 +505,19 @@ partition_setup(struct disk *dp)
   hd_name[8] = '\0';			/* i.e.  /dev/hda6 -> /dev/hda */
 #endif
 
-  enter_priv_on();
   part_fd = SILENT_DOS_SYSCALL(open(hd_name, O_RDONLY));
-  leave_priv_setting();
   if (part_fd == -1) {
     if (dp->removeable) return;
-    error("opening device %s to read MBR for PARTITION %s\n",
-	  hd_name, dp->dev_name);
-    leavedos(22);
+    PNUM = 1;
+    set_part_ent(dp, tmp_mbr);
+    tmp_mbr[0x1fe] = 0x55;
+    tmp_mbr[0x1ff] = 0xaa;
+  } else {
+    RPT_SYSCALL(read(part_fd, tmp_mbr, SECTOR_SIZE));
+    close(part_fd);
+    d_printf("Using MBR from %s for PARTITION %s (part#=%d).\n",
+             hd_name, dp->dev_name, PNUM);
   }
-  RPT_SYSCALL(read(part_fd, tmp_mbr, SECTOR_SIZE));
-  close(part_fd);
-  d_printf("Using MBR from %s for PARTITION %s (part#=%d).\n",
-           hd_name, dp->dev_name, PNUM);
   free(hd_name);
 
   /* check for logical partition, if so simulate as primary part#1 */
@@ -587,8 +587,7 @@ partition_setup(struct disk *dp)
  *       put in the dp->part_info.number'th entry in the part table.
  */
 
-void
-set_part_ent(struct disk *dp, char *tmp_mbr)
+static void set_part_ent(struct disk *dp, char *tmp_mbr)
 {
   long	length;		/* partition length in sectors		*/
   long	end;		/* last sector number offset		*/
@@ -614,16 +613,26 @@ set_part_ent(struct disk *dp, char *tmp_mbr)
 	HEAD(end), SECT(end), CYL(end));
 
   /* get address of where to put new part table entry */
-  p = tmp_mbr + PART_INFO_START + (PART_INFO_LEN * (dp->part_info.number-1)); 
+  p = tmp_mbr + PART_INFO_START + (PART_INFO_LEN * (dp->part_info.number-1));
 
+  /* initialize with LBA values */
+  p[1] = p[5] = 254;
+  p[2] = p[6] = 255;
+  p[3] = p[7] = 255;
+  p[4] = 0xe;
+    
   p[0] = PART_BOOT;						/* bootable  */
-  p[1] = HEAD(dp->start);					/* beg head  */
-  p[2] = SECT(dp->start) | ((CYL(dp->start) >> 2) & 0xC0);	/* beg sect  */
-  p[3] = CYL(dp->start) & 0xFF;					/* beg cyl   */
-  p[4] = (length < 32*1024*1024/SECTOR_SIZE)? 0x04 : 0x06;	/* dos sysid */
-  p[5] = HEAD(end);						/* end head  */
-  p[6] = SECT(end) | ((CYL(end) >> 2) & 0xC0);			/* end sect  */
-  p[7] = CYL(end) & 0xFF;					/* end cyl   */
+  if (CYL(dp->start) <= 1023) {
+    p[1] = HEAD(dp->start);					/* beg head  */
+    p[2] = SECT(dp->start) | ((CYL(dp->start) >> 2) & 0xC0);	/* beg sect  */
+    p[3] = CYL(dp->start) & 0xFF;				/* beg cyl   */
+  }
+  if (CYL(end) <= 1023) {
+    p[4] = (length < 32*1024*1024/SECTOR_SIZE)? 0x04 : 0x06;	/* dos sysid */
+    p[5] = HEAD(end);						/* end head  */
+    p[6] = SECT(end) | ((CYL(end) >> 2) & 0xC0);		/* end sect  */
+    p[7] = CYL(end) & 0xFF;					/* end cyl   */
+  }
   *((long *)(p+8)) = dp->start;					/* pre sects */
   *((long *)(p+12)) = length;					/* len sects */
 }
@@ -645,7 +654,7 @@ disk_close(void)
   if (!disks_initiated) return;  /* just to be safe */
   for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
     if (dp->removeable && dp->fdesc >= 0) {
-      d_printf("DISK: %s: Closing a disk\n",dp->dev_name);
+      d_printf("DISK: Closing disk %s\n",dp->dev_name);
       (void) close(dp->fdesc);
       dp->fdesc = -1;
     }
@@ -657,15 +666,12 @@ disk_close(void)
 void
 disk_open(struct disk *dp)
 {
-  PRIV_SAVE_AREA
   struct floppy_struct fl;
 
   if (dp == NULL || dp->fdesc >= 0)
     return;
     
-  enter_priv_on();
   dp->fdesc = SILENT_DOS_SYSCALL(open(dp->type == DIR_TYPE ? "/dev/null" : dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0));
-  leave_priv_setting();
 
   /* FIXME:
    * Why the hell was the below handling restricted to non-removeable disks?
@@ -674,25 +680,16 @@ disk_open(struct disk *dp)
    */
   if ( /*!dp->removeable &&*/ (dp->fdesc < 0)) {
     if (errno == EROFS || errno == ENODEV) {
-      enter_priv_on();
       dp->fdesc = DOS_SYSCALL(open(dp->dev_name, O_RDONLY, 0));
-      leave_priv_setting();
       if (dp->fdesc < 0) {
-        d_printf("ERROR: (disk) can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
+        error("ERROR: (disk) can't open %s for read nor write: %s\n", dp->dev_name, strerror(errno));
         /* In case we DO get more clever, we want to share that code */
-        goto fail;
       } else {
         dp->rdonly = 1;
         d_printf("(disk) can't open %s for read/write. Readonly used.\n", dp->dev_name);
       }
     } else {
       d_printf("ERROR: (disk) can't open %s: %s\n", dp->dev_name, strerror(errno));
-    fail:
-#if 0
-      /* We really should be more clever here */
-      fatalerr = 5;
-#endif
-      return;
     }
   }
   else dp->rdonly = dp->wantrdonly;
@@ -738,7 +735,7 @@ disk_open(struct disk *dp)
       dp->tracks = 0;
       return;
     }
-    error("can't get floppy parameter of %s (%s)\n", dp->dev_name, sys_errlist[errno]);
+    error("can't get floppy parameter of %s (%s)\n", dp->dev_name, strerror(errno));
     fatalerr = 5;
     return;
   }
@@ -794,7 +791,6 @@ disk_close_all(void)
 void
 disk_init(void)
 {
-  PRIV_SAVE_AREA
   struct disk *dp=NULL;
   struct stat stbuf;
 
@@ -806,17 +802,13 @@ disk_init(void)
 
   disks_initiated = 1;  /* disk_init has been called */
   if (config.bootdisk) {
-    enter_priv_on();
     bootdisk.fdesc = open(bootdisk.dev_name,
 			  bootdisk.rdonly ? O_RDONLY : O_RDWR, 0);
-    leave_priv_setting();
     if (bootdisk.fdesc < 0) {
       if (errno == EROFS) {
-        enter_priv_on();
         bootdisk.fdesc = open(bootdisk.dev_name, O_RDONLY, 0);
-        leave_priv_setting();
         if (bootdisk.fdesc < 0) {
-          error("can't open bootdisk %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
+          error("can't open bootdisk %s for read nor write: %s\n", dp->dev_name, strerror(errno));
           leavedos(23);
         } else {
           bootdisk.rdonly = 1;
@@ -855,16 +847,12 @@ disk_init(void)
       continue;
     }
 #endif
-    enter_priv_on();
     dp->fdesc = open(dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
-    leave_priv_setting();
     if (dp->fdesc < 0) {
       if (errno == EROFS || errno == EACCES) {
-        enter_priv_on();
         dp->fdesc = open(dp->dev_name, O_RDONLY, 0);
-        leave_priv_setting();
         if (dp->fdesc < 0) {
-          error("can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
+          error("can't open %s for read nor write: %s\n", dp->dev_name, strerror(errno));
           leavedos(25);
         } else {
           dp->rdonly = 1;
@@ -879,6 +867,12 @@ disk_init(void)
     else dp->rdonly = dp->wantrdonly;
   }
 
+  if (!FDISKS && use_bootdisk) {
+  /* if we don't have any configured floppies, we have to use bootdisk instead */
+    memcpy(&disktab[0], &bootdisk, sizeof(bootdisk));
+    FDISKS++;	/* now we have one */
+  }
+
   /*
    * Open hard disks
    */
@@ -886,23 +880,17 @@ disk_init(void)
     if(dp->type == IMAGE)  {
 	if (dp->dexeflags & DISK_DEXE_RDWR) {
 	  d_printf("IMAGE: dexe, RDWR access allowed for %s\n",dp->dev_name);
-	  enter_priv_on();
 	}
 	else {
 	  d_printf("IMAGE: Using user permissions\n");
-	  enter_priv_off();
 	}
     }
-    else enter_priv_on();
     dp->fdesc = open(dp->type == DIR_TYPE ? "/dev/null" : dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
-    leave_priv_setting();
     if (dp->fdesc < 0) {
       if (errno == EROFS || errno == EACCES) {
-        if (dp->type == IMAGE) enter_priv_off(); else enter_priv_on();
         dp->fdesc = open(dp->dev_name, O_RDONLY, 0);
-        leave_priv_setting();
         if (dp->fdesc < 0) {
-          error("can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
+          error("can't open %s for read nor write: %s\n", dp->dev_name, strerror(errno));
           leavedos(26);
         } else {
           dp->rdonly = 1;
@@ -973,8 +961,7 @@ disk_init(void)
   }
 }
 
-int
-checkdp(struct disk *disk)
+static int checkdp(struct disk *disk)
 {
   if (disk == NULL) {
     d_printf("DISK: null dp\n");
@@ -988,8 +975,7 @@ checkdp(struct disk *disk)
     return 0;
 }
 
-void
-int13(u_char i)
+int int13(void)
 {
   unsigned int disk, head, sect, track, number;
   int res;
@@ -999,10 +985,9 @@ int13(u_char i)
   int checkdp_val;
 
   disk = LO(dx);
-  if (disk < FDISKS) {
-    if (!disk && use_bootdisk)
-      dp = &bootdisk;
-    else
+  if (!disk && use_bootdisk)
+    dp = &bootdisk;
+  else if (disk < FDISKS) {
       dp = &disktab[disk];
     switch (HI(ax)) {
       #define DISKETTE_MOTOR_TIMEOUT (*((unsigned char *)0x440))
@@ -1019,6 +1004,8 @@ int13(u_char i)
   else
     dp = NULL;
 
+  d_printf("INT13: disk=%#x ah=%#x dp=%p\n", LO(dx), HI(ax), dp);
+
   /* this is a bad hack to ensure that the cached blocks aren't.
    * Linux only checks disk change on open()
    */
@@ -1051,6 +1038,10 @@ int13(u_char i)
       track |= (HI(dx) & 0xc0) << 4;
     buffer = SEG_ADR((char *), es, bx);
     number = LO(ax);
+#ifdef X86_EMULATOR
+    e_printf("DISK %d read [h:%d,s:%d,t:%d](%d)->%04x:%04x\n",
+	     disk, head, sect, track, number, REG(es), LWORD(ebx));
+#endif
 
     if (checkdp_val || head >= dp->heads ||
 	sect >= dp->sectors || track >= dp->tracks) {
@@ -1332,7 +1323,7 @@ int13(u_char i)
 	NOCARRY;
       }
       else {
-	error("gettype: no disk %d\n", disk);
+	d_printf("gettype: no disk %d\n", disk);
 	HI(ax) = 0;		/* disk not there */
 	REG(eflags) |= CF;	/* error */
       }
@@ -1401,7 +1392,7 @@ int13(u_char i)
     diskaddr->blocks = 0;
 
     if (checkdp(dp) || track >= dp->tracks) {
-      error("ERROR: Sector not found, AH=0x42!\n");
+      error("Sector not found, AH=0x42!\n");
       d_printf("DISK %d ext read [h:%d,s:%d,t:%d](%d)->%p\n",
 	       disk, head, sect, track, number, (void *) buffer);
       if (dp) {
@@ -1425,7 +1416,7 @@ int13(u_char i)
       break;
     }
     else if (res & 511) {	/* must read multiple of 512 bytes */
-      error("ERROR: sector_corrupt 1, return = %d!\n", res);
+      error("sector_corrupt 1, return = %d!\n", res);
       HI(ax) = DERR_BADSEC;	/* sector corrrupt */
       CARRY;
       break;
@@ -1453,7 +1444,7 @@ int13(u_char i)
     diskaddr->blocks = 0;
 
     if (checkdp(dp) || track >= dp->tracks) {
-      error("ERROR: Sector not found, AH=0x42!\n");
+      error("Sector not found, AH=0x42!\n");
       d_printf("DISK %d ext write [h:%d,s:%d,t:%d](%d)->%p\n",
 	       disk, head, sect, track, number, (void *) buffer);
       if (dp) {
@@ -1490,7 +1481,7 @@ int13(u_char i)
       break;
     }
     else if (res & 511) {	/* must read multiple of 512 bytes */
-      error("ERROR: sector_corrupt 1, return = %d!\n", res);
+      error("sector_corrupt 1, return = %d!\n", res);
       HI(ax) = DERR_BADSEC;	/* sector corrrupt */
       CARRY;
       break;
@@ -1523,7 +1514,7 @@ int13(u_char i)
     params = SEG_ADR((struct ibm_ms_drive_params *), ds, si);
 
     if (checkdp(dp)) {
-      error("ERROR: Invalid drive, AH=0x48!\n");
+      error("Invalid drive, AH=0x48!\n");
       if (dp) {
 	  d_printf("DISK dev %s GEOM %d heads %d sects %d trk\n",
 		   dp->dev_name, dp->heads, dp->sectors, dp->tracks);
@@ -1573,8 +1564,9 @@ int13(u_char i)
 	  LWORD(eax));
     show_regs(__FILE__, __LINE__);
     CARRY;
-    return;
+    break;
   }
+  return 1;
 }
 
 #define FLUSH_DELAY 2
@@ -1585,10 +1577,12 @@ int13(u_char i)
 void
 floppy_tick(void)
 {
-  static int secs = 0;
+  static int ticks = 0;
 
-  if (++secs == config.fastfloppy) {
+  if (++ticks >= config.fastfloppy) {
     disk_close();
-    secs = 0;
+    if (debug_level('d') > 2)
+      d_printf("FLOPPY: flushing after %d ticks\n", ticks);
+    ticks = 0;
   }
 }
