@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -8,10 +8,11 @@
  * Various sundry utilites for dos emulator.
  *
  */
+#include "emu.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
-#include <malloc.h>
 #include <sys/time.h>
 #include <unistd.h>
 #include <sys/types.h>
@@ -19,11 +20,10 @@
 #include <fcntl.h>
 #include <string.h>
 #include <ctype.h>
-#include "Linux/vt.h"
+#include <sys/vt.h>
 #include <sys/ioctl.h>
 #include <errno.h>
 
-#include "emu.h"
 #include "machcompat.h"
 #include "bios.h"
 #include "timers.h"
@@ -43,12 +43,15 @@
  *                                     --Hans 990213
  */
 #define SHOW_TIME	0		/* 0 or 1 */
+#ifdef X86_EMULATOR
+#include "cpu-emu.h"
+#endif
 
 #ifndef INITIAL_LOGBUFSIZE
 #define INITIAL_LOGBUFSIZE      0
 #endif
 #ifndef INITIAL_LOGBUFLIMIT
-#define INITIAL_LOGFILELIMIT	(10*1024*1024)
+#define INITIAL_LOGFILELIMIT	(50*1024*1024)
 #endif
 
 #ifdef CIRCULAR_LOGBUFFER
@@ -137,14 +140,13 @@ char *chrprintable(char c)
   return strprintable(buf);
 }
 
-static int vlog_printf(int flg, const char *fmt, va_list args)
+int vlog_printf(int flg, const char *fmt, va_list args)
 {
   int i;
   static int is_cr = 1;
 
   if (dosdebug_flags & DBGF_INTERCEPT_LOG) {
     /* we give dosdebug a chance to interrupt on given logoutput */
-    extern int vmhp_log_intercept(int flg, const char *fmt, va_list args);
     i = vmhp_log_intercept(flg, fmt, args);
     if ((dosdebug_flags & DBGF_DISABLE_LOG_TO_FILE) || !dbg_fd) return i;
   }
@@ -206,15 +208,24 @@ static int vlog_printf(int flg, const char *fmt, va_list args)
       if (errno==ENOSPC) leavedos(0x4c4c);
     }
     logptr = logbuf;
+#if 1
     if (logfile_limit) {
       log_written += fsz;
       if (log_written > logfile_limit) {
         fflush(dbg_fd);
+#if 1
         ftruncate(fileno(dbg_fd),0);
         fseek(dbg_fd, 0, SEEK_SET);
         log_written = 0;
+#else
+	fclose(dbg_fd);
+	shut_debug = 1;
+	dbg_fd = 0;	/* avoid recursion in leavedos() */
+	leavedos(0);
+#endif
       }
     }
+#endif
   }
 #endif
   return i;
@@ -357,15 +368,13 @@ char *get_proc_string_by_key(char *key)
     if ((p == procbufptr) || (p[-1] == '\n')) {
       p += strlen(key);
       while (*p && isblank(*p)) p++;
-      if (*p == ':') {
-        p++;
-        while (*p && isblank(*p)) p++;
-        proclastpos = p;
-        while (*proclastpos && (*proclastpos != '\n')) proclastpos++;
-        proclastdelim = *proclastpos;
-        *proclastpos = 0;
-        return p;
-      }
+      if (*p == ':') p++;
+      while (*p && isblank(*p)) p++;
+      proclastpos = p;
+      while (*proclastpos && (*proclastpos != '\n')) proclastpos++;
+      proclastdelim = *proclastpos;
+      *proclastpos = 0;
+      return p;
     }
   }
 #if 0
@@ -452,8 +461,7 @@ char *mkdir_under(char *basedir, char *dir, int append_pid)
 	if (dir) s = assemble_path(basedir, dir, append_pid);
 	if (!exists_dir(s)) {
 		if (mkdir(s, S_IRWXU)) {
-			fprintf(stderr, "can't create local %s directory, giving up\n", s);
-			longjmp(NotJEnv, 0x42);
+			fprintf(stderr, "can't create local %s directory\n", s);
 		}
 	}
 	return s;
@@ -516,28 +524,7 @@ void subst_file_ext(char *ptr)
 {
 #define ext_fix(s) { char *r=(s); \
 		     while (*r) { *r=toupper(*r); r++; } }
-    static int subst_bat=0, subst_sys=0;
-    static int needed = -1;
-
-    if (!needed) return;
-
-    if (needed < 0) {
-	needed = 0;
-        if (config.emubat) {
-	    needed++;
-	    subst_bat = 3;	/* This realy is an ugly hack:
-				 * freedos needs 3 fakes,  PCDOS 2 and MSDOS 1
-				 */
-        }
-        if (config.emusys) {
-	    needed++;
-	    subst_sys = 2;
-        }
-        if (config.emuini) {
-	    needed++;
-        }
-	if (!needed) return;
-    }
+    static int subst_sys=2;
 
     /* skip leading drive name and \ */
     if (ptr[1]==':' && ptr[2]=='\\') ptr+=3;
@@ -550,15 +537,8 @@ void subst_file_ext(char *ptr)
 	return;
     }
 
-    if (subst_bat && config.emubat && !strncasecmp(ptr, "AUTOEXEC.BAT", 12)) {
-	ext_fix(config.emubat);
-	sprintf(ptr, "AUTOEXEC.%-3s", config.emubat);
-	d_printf("DISK: Substituted %s for AUTOEXEC.BAT\n", ptr);
-	if (! --subst_bat) needed--;
-	return;
-    } 
-
-    if (subst_sys) {
+    if (subst_sys && config.emusys) {
+        char config_name[6+1+3+1];
 #if 0	/*
 	 * NOTE: as the method used in fatfs.c can't handle multiple
 	 * files to be faked, we can't do it here, because this would
@@ -568,12 +548,15 @@ void subst_file_ext(char *ptr)
 	/* skip the D for DCONFIG.SYS in DR-DOS */
 	if (toupper(ptr[0]) == 'D') ptr++;
 #endif
-	if (config.emusys && !strncasecmp(ptr, "CONFIG.SYS", 10)) {
-	    ext_fix(config.emusys);
-	    sprintf(ptr, "CONFIG.%-3s", config.emusys);
+        ext_fix(config.emusys);
+        sprintf(config_name, "CONFIG.%-3s", config.emusys);
+        if (subst_sys == 1 && strcasecmp(ptr, config_name) &&
+            strcasecmp(ptr, "CONFIG.SYS")) {
+            subst_sys = 0;
+        } else if (!strcasecmp(ptr, "CONFIG.SYS")) {
+            strcpy(ptr, config_name);
 	    d_printf("DISK: Substituted %s for CONFIG.SYS\n", ptr);
-	    if (! --subst_sys) needed--;
-	    return;
+	    subst_sys = 1;
 	}
     } 
 }
@@ -621,15 +604,44 @@ void call_cmd(const char *cmd, int maxargs, const struct cmd_db *cmdtab,
 void sigalarm_onoff(int on)
 {
   static struct itimerval itv_old;
+#ifdef X86_EMULATOR
+  static struct itimerval itv_oldp;
+#endif
   static struct itimerval itv;
-  static int is_off = 0;
+  static volatile int is_off = 0;
   if (on) {
-    if (is_off--) setitimer(TIMER_TIME, &itv_old, NULL);
+    if (is_off--) {
+    	setitimer(ITIMER_REAL, &itv_old, NULL);
+#ifdef X86_EMULATOR
+    	setitimer(ITIMER_PROF, &itv_oldp, NULL);
+#endif
+    }
   }
   else if (!is_off++) {
     itv.it_interval.tv_sec = itv.it_interval.tv_usec = 0;
     itv.it_value = itv.it_interval;
-    setitimer(TIMER_TIME, &itv, &itv_old);
+    setitimer(ITIMER_REAL, &itv, &itv_old);
+#ifdef X86_EMULATOR
+    setitimer(ITIMER_PROF, &itv, &itv_oldp);
+#endif
+  }
+}
+
+void sigalarm_block(int block)
+{
+  static volatile int is_blocked = 0;
+  sigset_t blockset;
+  if (block) {
+    if (!is_blocked++) {
+      sigemptyset(&blockset);
+      sigaddset(&blockset, SIGALRM);
+      sigprocmask(SIG_BLOCK, &blockset, NULL);
+    }
+  }
+  else if (is_blocked--) {
+    sigemptyset(&blockset);
+    sigaddset(&blockset, SIGALRM);
+    sigprocmask(SIG_UNBLOCK, &blockset, NULL);
   }
 }
 
@@ -641,12 +653,13 @@ int is_console(int fd)
    * of an VT. ... simple;-)
    */
   int nextvt;
+  struct vt_mode vtm;
   if( ioctl( fd, VT_OPENQRY, &nextvt ) ) return 0;
+  if( ioctl( fd, VT_GETMODE, &vtm ) ) return 0;
   return 1;
 }
 
 
-#ifdef NEED_LLSEEK_PROTOTYPE
 /*
  * getting tired of needless 'llseek() versus lseek64()' war
  * ... calling the kernel directly  --Hans 2000/03/26
@@ -670,70 +683,47 @@ long long libless_llseek(int fd, long long offset, int origin)
   }
   return result;
 }
-#endif
 
-
-#if (GLIBC_VERSION_CODE >= 2000) && defined(PORTABLE_BINARY)
-/*
- * Under glibc we can't make portable static binaries due to libnss stuff.
- * As we only need to look at /etc/passwd (and which non-libnss system won't
- * have it?) we work around this problem by doing getpwuid() getpwnam()
- * ourselves, if we have no /etc/nsswitch.conf.
- *                                                -- Hans, 981201
- */
-
-static struct passwd *our_getpw(const char *name, uid_t uid)
+/* dynamic readlink, adapted from "info libc" */
+char *readlink_malloc (const char *filename)
 {
-  FILE *f;
-  static struct passwd pw;
-  static char buf[1024];
-  char *p;
+  int size = 50;
+  int nchars;
+  char *buffer;
 
-  f = fopen("/etc/passwd", "r");
-  if (!f) return 0;
-  buf[sizeof(buf)-1] = 0;
-  while (fgets(buf, sizeof(buf)-1, f)) {
-    pw.pw_name = strtok(buf, ":");
-    if (!pw.pw_name) continue;
-    if (name && strcmp(name, pw.pw_name)) continue;
-    pw.pw_passwd = strtok(0, ":");
-    if (!pw.pw_passwd) continue;
-    p = strtok(0, ":");
-    if (!p) continue;
-    pw.pw_uid = (uid_t)strtoul(p,0,0);
-    if (!name && pw.pw_uid != uid) continue;
-    p = strtok(0, ":");
-    if (!p) continue;
-    pw.pw_gid = (gid_t)strtoul(p,0,0);
-    /* we don't need more, hence we skip the rest */
-    fclose(f);
-    return &pw;
-  }
-  fclose(f);
-  return 0;
+  do {
+    size *= 2;
+    nchars = 0;
+    buffer = malloc(size);
+    if (buffer != NULL) {
+      nchars = readlink(filename, buffer, size);
+      if (nchars < 0) {
+        free(buffer);
+        buffer = NULL;
+      }
+    }
+  } while (nchars >= size);
+  if (buffer != NULL)
+    buffer[nchars] = '\0';
+  return buffer;
 }
 
-static int have_libnss()
+char * strupr(char *s)
 {
-  static int have_it = -1;
-  if (have_it >= 0 ) return have_it;
-  have_it = exists_file("/etc/nsswitch.conf") ? 1: 0;
-  return have_it;
+	char *p = s;
+	while (*p) {
+		*p = toupper(*p);
+		p++;
+	}
+	return s;
 }
 
-#undef getpwuid
-#undef getpwnam
-
-struct passwd *our_getpwuid(uid_t uid)
+char * strlower(char *s)
 {
-  if (have_libnss()) return getpwuid(uid);
-  return our_getpw(0, uid);
+	char *p = s;
+	while (*p) {
+		*p = tolower(*p);
+		p++;
+	}
+	return s;
 }
-
-struct passwd *our_getpwnam(const char *name)
-{
-  if (have_libnss()) return getpwnam(name);
-  return our_getpw(name, 0);
-}
-
-#endif /* GLIBC_VERSION_CODE >= 2000 */
