@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -9,6 +9,7 @@
  *
  */
 
+#include "config.h"
 #include <stdio.h>
 #include <termios.h>
 #include <stdlib.h>
@@ -17,13 +18,18 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
+#include <sys/stat.h>
 #include <errno.h>
 #include <string.h>
 #include <sys/mman.h>
 
-#include "config.h"
+#ifdef X_SUPPORT
+#include "../env/video/X.h"
+#endif
+
 #include "emu.h"
 #include "bios.h"
+#include "int.h"
 #include "memory.h"
 #include "video.h"		/* video base address */
 #include "mouse.h"
@@ -38,23 +44,13 @@
 
 #include "keyboard.h"
 
-#include "priv.h"
-
-extern int get_perm ();
-extern int release_perm ();
-
-#ifdef X_SUPPORT
-extern void X_show_mouse_cursor(int yes);
-extern void X_set_mouse_cursor(int display, int x, int y, int x_range, int y_range);
-#endif
+#include "iodev.h"
 
 /* This is included for video mode support. Please DO NOT remove !
  * mousevid.h will become part of VGA emulator package */
 #include "mousevid.h"
 #include "gcursor.h"
 
-void DOSEMUSetupMouse(void);
-
 #define Mouse_INT	(0x33 * 16)
 #ifndef USE_NEW_INT
 /* I have not properly tested this INT74 - JES 96/10/20 */
@@ -90,10 +86,8 @@ void mouse_cursor(int), mouse_pos(void), mouse_setpos(void),
  mouse_detpage(void), mouse_getmaxminvirt(void);
 
 /* mouse movement functions */
-void mouse_reset(int);
-void mouse_move(void), mouse_lb(void), mouse_rb(void);
+static void mouse_reset(int);
 static void mouse_do_cur(void), mouse_update_cursor(void);
-static void mouse_reset_to_current_video_mode(void);
 
 /* graphics cursor */
 void graph_cursor(void), text_cursor(void);
@@ -109,11 +103,11 @@ static int last_mouse_call_read_mickeys = 0;
 
 static int mouse_events = 0;
 static mouse_erase_t mouse_erase;
+static int sent_mouse_esc = FALSE;
 
-#if 0
-mouse_t mice[MAX_MOUSE] ;
-struct mouse_struct mouse;
-#endif
+static mouse_t *mice = &config.mouse;
+/* the 'volatile' is there to cover some bug in gcc -O -g3 */
+static volatile struct mouse_struct mouse;
 
 static inline int mouse_roundx(int x) 
 { 
@@ -177,14 +171,7 @@ mouse_helper(void)
   switch (LO(bx)) {
   case 0:				/* Reset iret for mouse */
     m_printf("MOUSE move iret !\n");
-    SETIVEC(0x33, Mouse_SEG, Mouse_INT);
-#ifndef USE_NEW_INT
-    SETIVEC(0x74, Mouse_SEG, Mouse_INT74);
-#else /* USE_NEW_INT */
-#if 0
-    SETIVEC(0x74, Mouse_SEG, Mouse_ROUTINE_OFF);
-#endif
-#endif /* USE_NEW_INT */
+    mouse_enable_internaldriver();
     break;
   case 1:				/* Select Microsoft Mode */
     m_printf("MOUSE Microsoft Mouse (two buttons) selected.\n");
@@ -207,18 +194,18 @@ mouse_helper(void)
     LO(dx) = mouse.ignorexy;
     break;
   case 4:				/* Set vertical speed */
-    if ((LO(cx) < 1) || (LO(cx) > 255)) {
+    if (LO(cx) < 1) {
       m_printf("MOUSE Vertical speed out of range. ERROR!\n");
       LWORD(eax) = 1;
     } else 
-      mouse.speed_y = LO(cx);
+      mouse.init_speed_y = mouse.speed_y = LO(cx);
     break;
   case 5:				/* Set horizontal speed */
-    if ((LO(cx) < 1) || (LO(cx) > 255)) {
+    if (LO(cx) < 1) {
       m_printf("MOUSE Horizontal speed out of range. ERROR!\n");
       LWORD(eax) = 1;
     } else
-      mouse.speed_x = LO(cx);
+      mouse.init_speed_x = mouse.speed_x = LO(cx);
     break;
   case 6:				/* Ignore horz/vert selection */
     if (LO(cx) == 1) 
@@ -255,10 +242,112 @@ mouse_helper(void)
   }
 }
 
-void
+void mouse_ps2bios(void)
+{        
+  m_printf("PS2MOUSE: Call ax=0x%04x\n", LWORD(eax));
+  if (!mice->intdrv) {
+    REG(eax) = 0x0500;        /* No ps2 mouse device handler */
+    CARRY;
+    return;
+  }
+  
+  switch (REG(eax) &= 0x00FF) {
+  case 0x0000:                    
+    mouse.ps2.state = HI(bx);
+    if (mouse.ps2.state == 0)
+      mice->intdrv = FALSE;
+    else
+      mice->intdrv = TRUE;
+    HI(ax) = 0;
+    NOCARRY;		
+    break;
+  case 0x0001:
+    HI(ax) = 0;
+    LWORD(ebx) = 0xAAAA;    /* we have a ps2 mouse */
+    NOCARRY;
+    break;
+  case 0x0003:
+    if (HI(bx) != 0) {
+      CARRY;
+      HI(ax) = 1;
+    } else {
+      NOCARRY;
+      HI(ax) = 0;
+    }
+    break;
+  case 0x0004:
+    HI(bx) = 0xAA;
+    HI(ax) = 0;
+    NOCARRY;
+    break;
+  case 0x0005:			/* Initialize ps2 mouse */
+    HI(ax) = 0;
+    mouse.ps2.pkg = HI(bx);
+    NOCARRY;
+    break;
+  case 0x0006:
+    switch (HI(bx)) {
+    case 0x00:
+      LO(bx)  = (mouse.rbutton ? 1 : 0);
+      LO(bx) |= (mouse.lbutton ? 4 : 0);
+      LO(bx) |= 0; 	/* scaling 1:1 */
+      LO(bx) |= 0x20;	/* device enabled */
+      LO(bx) |= 0;	/* stream mode */
+      LO(cx) = 0;		/* resolution, one count */
+      LO(dx) = 0;		/* sample rate */
+      HI(ax) = 0;
+      NOCARRY;
+      break;
+    case 0x01:
+      HI(ax) = 0;
+      NOCARRY;
+      break;
+    case 0x02:
+      HI(ax) = 1;
+      CARRY;
+      break;
+    }
+    break;
+#if 0
+  case 0x0007:
+    pushw(ssp, sp, 0x000B);
+    pushw(ssp, sp, 0x0001);
+    pushw(ssp, sp, 0x0001);
+    pushw(ssp, sp, 0x0000);
+    REG(cs) = REG(es);
+    REG(eip) = REG(ebx);
+    HI(ax) = 0;
+    NOCARRY;
+    break;
+#endif
+  default:
+    HI(ax) = 1;
+    g_printf("PS2MOUSE: Unknown call ax=0x%04x\n", LWORD(eax));
+    CARRY;
+  }
+}
+
+int
 mouse_int(void)
 {
+  /* delayed mouse init for xterms; should be done cleaner in 1.3.x */
+  if (mice->type == MOUSE_XTERM) {
+    /* save old highlight mouse tracking */
+    printf("\033[?1001s");
+    /* enable mouse tracking */
+    printf("\033[?9h\033[?1000h\033[?1002h\033[?1003h");	
+    fflush (stdout);
+    m_printf("XTERM MOUSE: Remote terminal mouse tracking enabled\n");
+    sent_mouse_esc = TRUE;
+  }
+  
   m_printf("MOUSEALAN: int 0x%x ebx=%x\n", LWORD(eax), LWORD(ebx));
+  if (!mice->intdrv || (!mouse.enabled && LWORD(eax) != 0x20)) {
+    m_printf("MOUSE: driver disabled, intdrv=%i enable=%i\n",
+      mice->intdrv, mouse.enabled);
+    return 0;
+  }
+
   switch (LWORD(eax)) {
   case 0x00:			/* Mouse Reset/Get Mouse Installed Flag */
     mouse_reset(0);
@@ -505,6 +594,8 @@ mouse_int(void)
 
   if (mouse.cursor_on == 0)
      mouse_update_cursor();
+
+  return 1;
 }
 
 void
@@ -631,14 +722,7 @@ mouse_software_reset(void)
   mouse.cursor_on = -1;      
   mouse.cs=0;
   mouse.ip=0;
-  SETIVEC(0x33, Mouse_SEG, Mouse_INT);
-#ifndef USE_NEW_INT
-  SETIVEC(0x74, Mouse_SEG, Mouse_INT74);
-#else /* USE_NEW_INT */
-#if 0
-  SETIVEC(0x74, Mouse_SEG, Mouse_ROUTINE_OFF);
-#endif
-#endif /* USE_NEW_INT */
+  mouse_enable_internaldriver();
 
   /* Return 0xffff on success, 0x21 on failure */
   LWORD(eax) = 0xffff;  
@@ -788,14 +872,8 @@ mouse_reset_to_current_video_mode(void)
    * then in mouse_reset, as it gets called more often.
    * -- Eric Biederman 29 May 2000
    */
-  if (mice->type != MOUSE_X) {
-    mouse.speed_x = 8;
-    mouse.speed_y = 16;
-  } else {
-    /* default to 1 to 1 scaling for apps tha read mickeys */
-    mouse.speed_x = 8;
-    mouse.speed_y = 8;
-  }
+   mouse.speed_x = mouse.init_speed_x;
+   mouse.speed_y = mouse.init_speed_y;
 
  /*
   * Here we make sure text modes are resolved properly, according to the
@@ -882,10 +960,6 @@ mouse_reset_to_current_video_mode(void)
   mouse.maxx = current_video.width << mouse.xshift;
   mouse.maxy = current_video.height << mouse.yshift;
 
-  /* Find the center of the screen */
-  mouse.x = (mouse.minx + mouse.maxx) / 2;
-  mouse.y = (mouse.miny + mouse.maxy) / 2;
-
   /* force update first time after reset */
   mouse.oldrx = -1;
   
@@ -908,18 +982,10 @@ mouse_reset_to_current_video_mode(void)
 }
 
 
-void 
-mouse_reset(int flag)
+static void mouse_reset(int flag)
 {
   m_printf("MOUSE: reset mouse/installed!\n");
-  SETIVEC(0x33, Mouse_SEG, Mouse_INT);
-#ifndef USE_NEW_INT
-  SETIVEC(0x74, Mouse_SEG, Mouse_INT74);
-#else /* USE_NEW_INT */
-#if 0
-  SETIVEC(0x74, Mouse_SEG, Mouse_ROUTINE_OFF);
-#endif
-#endif /* USE_NEW_INT */
+  if (flag == 0) mouse_enable_internaldriver();
 
   /* Return 0xffff on mouse installed, 0x0000 - no mouse driver installed */
   /* Return number of mouse buttons */
@@ -965,6 +1031,8 @@ mouse_reset(int flag)
   memcpy((void *)mouse.graphcursormask,default_graphcursormask,32);
   mouse.hotx = mouse.hoty = -1;
 
+  mouse.enabled = TRUE;
+
   mouse_do_cur();
 }
 
@@ -1011,7 +1079,6 @@ void
 mouse_setpos(void)
 {
 #if X_SUPPORT
-  extern int grab_active;
   /* WP reads mickeys, and then sets the cursor position to make certain
    * it doesn't loose any mickeys.  This will catch that case, and keeps
    * us from breaking all apps under X that set the mouse position.
@@ -1033,6 +1100,9 @@ mouse_bpressinfo(void)
 {
   m_printf("MOUSE: Button press info\n");
 
+  if (LWORD(ebx) > 2)
+    LWORD(ebx) = 0;
+
   switch(LWORD(ebx)) {
  
   case 0:				/* left button */
@@ -1050,12 +1120,10 @@ mouse_bpressinfo(void)
     break;
 
   case 2:				/* middle button */
-    if (mouse.threebuttons) {
-      LWORD(ebx) = mouse.mpcount;
-      mouse.mpcount = 0;
-      LWORD(ecx) = mouse.mpx;
-      LWORD(edx) = mouse.mpy;
-    }
+    LWORD(ebx) = mouse.mpcount;
+    mouse.mpcount = 0;
+    LWORD(ecx) = mouse.mpx;
+    LWORD(edx) = mouse.mpy;
     break;
   }
   
@@ -1069,6 +1137,9 @@ mouse_brelinfo(void)
 {
   m_printf("MOUSE: Button release info\n");
 
+  if (LWORD(ebx) > 2)
+    LWORD(ebx) = 0;
+
   switch(LWORD(ebx)) {
  
   case 0:				/* left button */
@@ -1085,12 +1156,10 @@ mouse_brelinfo(void)
     LWORD(edx) = mouse.rry;
 
   case 2:				/* middle button */
-    if (mouse.threebuttons) {
-      LWORD(ebx) = mouse.mrcount;
-      mouse.mrcount = 0;
-      LWORD(ecx) = mouse.mrx;
-      LWORD(edx) = mouse.mry;
-    }
+    LWORD(ebx) = mouse.mrcount;
+    mouse.mrcount = 0;
+    LWORD(ecx) = mouse.mrx;
+    LWORD(edx) = mouse.mry;
     break;
   }
   
@@ -1142,7 +1211,11 @@ mouse_set_gcur(void)
   memcpy((void *)mouse.graphcursormask,ptr+16,32);
 
   /* compile it so that it can acutally be drawn. */
-  if (mice->type != MOUSE_X) {
+  if (mice->type != MOUSE_X && mice->type != MOUSE_XTERM 
+#ifdef USE_GPM
+      && mice->type != MOUSE_GPM
+#endif
+      ) {
     define_graphics_cursor((short *)mouse.graphscreenmask,(short *)mouse.graphcursormask);
   }
 }
@@ -1219,7 +1292,7 @@ mouse_disable_internaldriver()
   LWORD(es) = mouse.cs;
   LWORD(ebx) = mouse.ip;
 
-  mice->intdrv = FALSE;
+  mouse.enabled = FALSE;
 
   m_printf("MOUSE: Disable InternalDriver\n");
 }
@@ -1227,20 +1300,96 @@ mouse_disable_internaldriver()
 void
 mouse_enable_internaldriver()
 {
-  mice->intdrv = TRUE;
+  mouse.enabled = TRUE;
   SETIVEC(0x33, Mouse_SEG, Mouse_INT);
-#ifndef USE_NEW_INT
-  SETIVEC(0x74, Mouse_SEG, Mouse_INT74);
-#else /* USE_NEW_INT */
 #if 0
   SETIVEC(0x74, Mouse_SEG, Mouse_ROUTINE_OFF);
 #endif
-#endif /* USE_NEW_INT */
   m_printf("MOUSE: Enable InternalDriver\n");
 }
 
 
-
+#ifdef HAVE_UNICODE_KEYB
+/* Keyboard mouse control :)  */
+void mouse_keyboard(Boolean make, t_keysym key)
+{
+	static struct keyboard_mouse_state {
+		int l, r, u, d;
+		int ul, ur, dl, dr;
+		int lbutton, mbutton, rbutton; 
+	} state;
+	int dx, dy;
+	switch (key) {
+	case KEY_MOUSE_DOWN:
+		state.d = !!make;
+		break;
+	case KEY_MOUSE_LEFT:
+		state.l = !!make;
+		break;
+	case KEY_MOUSE_RIGHT:
+		state.r = !!make;
+		break;
+	case KEY_MOUSE_UP:
+		state.u = !!make;
+		break;
+	case KEY_MOUSE_UP_AND_LEFT:
+		state.ul = make;
+		break;
+	case KEY_MOUSE_UP_AND_RIGHT:
+		state.ur = make;
+		break;
+	case KEY_MOUSE_DOWN_AND_LEFT:
+		state.dl = make;
+		break;
+	case KEY_MOUSE_DOWN_AND_RIGHT:
+		state.dr = make;
+		break;
+	case KEY_MOUSE_BUTTON_LEFT:
+		state.lbutton = make;
+		break;
+	case KEY_MOUSE_BUTTON_RIGHT:
+		state.rbutton = make;
+		break;
+	case KEY_MOUSE_BUTTON_MIDDLE:
+		state.mbutton = make;
+		break;
+	default:
+		m_printf("MOUSE: keyboard_mouse(), key 0x%02x unknown!\n", key);
+		break;
+	}
+	dx = dy = 0;
+	if (state.d) {
+		dy += 1;
+	}
+	if (state.dl) {
+		dy += 1;
+		dx -= 1;
+	}
+	if (state.dr) {
+		dy += 1;
+		dx += 1;
+	}
+	if (state.u) {
+		dy -= 1;
+	}
+	if (state.ul) {
+		dy -= 1;
+		dx -= 1;
+	}
+	if (state.ur) {
+		dy -= 1;
+		dx += 1;
+	}
+	if (state.r) {
+		dx += 1;
+	}
+	if (state.l) {
+		dx -= 1;
+	}
+	mouse_move_relative(dx, dy);
+	mouse_move_buttons(state.lbutton, state.mbutton, state.rbutton);
+}
+#else
 /* XXX - something's wrong here. Shouldn't the first 4 cases move the cursor?
  */
 void mouse_keyboard(int sc)
@@ -1272,6 +1421,7 @@ void mouse_keyboard(int sc)
     m_printf("MOUSE: keyboard_mouse(), sc 0x%02x unknown!\n", sc);
   }
 }
+#endif
 
 
 static void mouse_round_coords(void)
@@ -1466,65 +1616,6 @@ void mouse_move_absolute(int x, int y, int x_range, int y_range)
 	   mouse_move();
 }
 
-void 
-fake_int(void)
-{
-  unsigned char *ssp;
-  unsigned long sp;
-
-  m_printf("MOUSE: fake_int: CS:IP %04x:%04x\n",LWORD(cs),LWORD(eip));
-  ssp = (unsigned char *)(LWORD(ss)<<4);
-  sp = (unsigned long) LWORD(esp);
-
-  pushw(ssp, sp, vflags);
-  pushw(ssp, sp, LWORD(cs));
-  pushw(ssp, sp, LWORD(eip));
-  LWORD(esp) -= 6;
-#if 0
-  do_hard_int(0x74);
-#endif
-  
-}
-
-void 
-fake_call(int cs, int ip)
-{
-  unsigned char *ssp;
-  unsigned long sp;
-
-  ssp = (unsigned char *)(LWORD(ss)<<4);
-  sp = (unsigned long) LWORD(esp);
-
-  m_printf("MOUSE: fake_call() shows S: %04x:%04x, CS:IP %04x:%04x\n",
-	   LWORD(ss), LWORD(esp), cs, ip);
-  pushw(ssp, sp, cs);
-  pushw(ssp, sp, ip);
-  LWORD(esp) -= 4;
-}
-
-void 
-fake_pusha(void)
-{
-  unsigned char *ssp;
-  unsigned long sp;
-
-  ssp = (unsigned char *)(LWORD(ss)<<4);
-  sp = (unsigned long) LWORD(esp);
-
-  pushw(ssp, sp, LWORD(eax));
-  pushw(ssp, sp, LWORD(ecx));
-  pushw(ssp, sp, LWORD(edx));
-  pushw(ssp, sp, LWORD(ebx));
-  pushw(ssp, sp, LWORD(esp));
-  pushw(ssp, sp, LWORD(ebp));
-  pushw(ssp, sp, LWORD(esi));
-  pushw(ssp, sp, LWORD(edi));
-  LWORD(esp) -= 16;
-  pushw(ssp, sp, REG(ds));
-  pushw(ssp, sp, REG(es));
-  LWORD(esp) -= 4;
-}
-
 /*
  * add the event to the current event mask
  */
@@ -1538,17 +1629,18 @@ mouse_delta(int event)
  * call user event handler
  */
 void
-mouse_event()
+do_mouse_irq()
 {
-  if (mouse.mask & mouse_events && (mouse.cs || mouse.ip)) {
-    if(in_dpmi && !in_dpmi_dos_int 
-          && !((mouse.cs == DPMI_SEG) && 
-          ((Bit32u)mouse.ip == (DPMI_OFF+HLT_OFF(DPMI_mouse_callback)))))
-                  fake_pm_int();
+    if(in_dpmi && !in_dpmi_dos_int)
+      fake_pm_int();
 
-    /* push iret frame on _SS:_SP. At F000:20F7 (bios.S) we get an
+    /* push iret frame on _SS:_SP. At F000:2146 (bios.S) we get an
      * iret and return to _CS:_IP */
-    fake_int();
+    fake_int(LWORD(cs), LWORD(eip));
+
+    /* push iret frame. At F000:20F7 (bios.S) we get an
+     * iret and return to F000:2140 for EOI */
+    fake_int(Mouse_SEG, Mouse_ROUTINE_OFF);
 
     /* push all 16-bit regs plus _DS,_ES. At F000:20F4 (bios.S) we find
      * 'pop es; pop ds; popa' */
@@ -1572,22 +1664,24 @@ mouse_event()
 
     REG(ds) = *mouse.csp;	/* put DS in user routine */
 
-    if (in_dpmi && 
-	(REG(cs) == DPMI_SEG) 
-	&& (REG(eip) == (DPMI_OFF+HLT_OFF(DPMI_mouse_callback))))
-	run_pm_mouse();
-    
     m_printf("MOUSE: event %d, x %d ,y %d, mx %d, my %d, b %x\n",
 	     mouse_events, mouse.x, mouse.y, mouse.maxx, mouse.maxy, LWORD(ebx));
     m_printf("MOUSE: "
 	     "should call %04x:%04x (actually %04x:%04x)"
 	     ".........jumping to %04x:%04x\n",
 	     mouse.cs, mouse.ip, *mouse.csp, *mouse.ipp,
-	     LWORD(cs), LWORD(eip));	
-  }
-  mouse_events = 0;
+	     LWORD(cs), LWORD(eip));
 }
 
+void mouse_event(void)
+{
+  if (mouse.mask & mouse_events && (mouse.cs || mouse.ip))
+    do_irq();
+  else
+    m_printf("MOUSE: Skipping irq, mask=0x%x, ev=0x%x, cs=0x%x, ip=0x%x\n",
+      mouse.mask, mouse_events, mouse.cs, mouse.ip);
+  mouse_events = 0;
+}
 
 /* unconditional mouse cursor update */
 static void
@@ -1610,6 +1704,9 @@ mouse_do_cur(void)
     return;
   }
 #endif
+  if (mice->type == MOUSE_XTERM)
+    return;
+
   if (!scr_state.current) 
   	return;
 
@@ -1674,8 +1771,6 @@ text_cursor(void)
 void 
 graph_cursor(void)
 {
-  get_perm();
-  
   erase_graphics_cursor(&mouse_erase);
 
   /* draw_graphics_cursor wants screen coordinates, we have coordinates
@@ -1683,15 +1778,13 @@ graph_cursor(void)
   if (mouse.cursor_on == 0)
 	  draw_graphics_cursor(mouse.rx >> mouse.xshift, mouse.ry >> mouse.yshift,
 		mouse.hotx,mouse.hoty,16,16,&mouse_erase);
-
-  release_perm();
 }
 
 
 void
 mouse_curtick(void)
 {
-  if (mouse.cursor_on != 0 || mice->type == MOUSE_X)
+  if (!mice->intdrv || mouse.cursor_on != 0 || mice->type == MOUSE_X)
     return;
 
   m_printf("MOUSE: curtick x:%d  y:%d\n", mouse.rx, mouse.ry);
@@ -1702,14 +1795,23 @@ mouse_curtick(void)
 }
 
 /* are ip and cs in right order?? */
-void
-mouse_sethandler(void *f, us * cs, us * ip)
+void mouse_sethandler(void)
 {
-  m_printf("MOUSE: sethandler...%p, %p, %p\n",
+  if (mice->intdrv) {
+    /* this is the mouse handler */
+    unsigned char *f = (unsigned char *) (Mouse_ADD+12);
+    u_short *ip = (u_short *) (Mouse_ADD + 8);
+    u_short *cs = (u_short *) (Mouse_ADD + 10);
+    /* tell the mouse driver where we are...exec add, seg, offset */
+    m_printf("MOUSE: sethandler...%p, %p, %p\n",
 	   (void *) f, (void *) cs, (void *) ip);
-  m_printf("...............CS:%04x.....IP:%04x\n", *cs, *ip);
-  mouse.csp = cs;
-  mouse.ipp = ip;
+    m_printf("...............CS:%04x.....IP:%04x\n", *cs, *ip);
+    mouse.csp = cs;
+    mouse.ipp = ip;
+  } else {
+    /* point to the retf#2 immediately after the int 33 entry point. */
+    SETIVEC(0x33, BIOSSEG, INT_OFF(0x33)+2);
+  }
 }
 
 /*
@@ -1721,9 +1823,8 @@ mouse_sethandler(void *f, us * cs, us * ip)
  * DANG_END_FUNCTION
  */
 void
-mouse_init(void)
+dosemu_mouse_init(void)
 {
-  PRIV_SAVE_AREA
   serial_t *sptr=NULL;
   char mouse_ver[]={2,3,4,5,0x14,0x7,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f};
 #if 1 /* BUG CATCHER */
@@ -1754,45 +1855,38 @@ mouse_init(void)
 
 #ifdef X_SUPPORT
   if (config.X) {
-    mice->intdrv = TRUE;
-    mice->type = MOUSE_X;
-    memcpy(p,mouse_ver,sizeof(mouse_ver));
-    m_printf("MOUSE: X Mouse being set\n");
+    if (mice->intdrv) {
+      mice->fd = -1;
+      mice->type = MOUSE_X;
+      memcpy(p,mouse_ver,sizeof(mouse_ver));
+      m_printf("MOUSE: X Mouse being set\n");
+    }
   }
   else 
 #endif
-  
-  if ( mice->type != MOUSE_X ){
+  if (
+#ifdef USE_GPM
+      !Mouse_gpm.init() &&
+#endif
+      !Mouse_xterm.init()) {
     if (mice->intdrv) {
+      struct stat buf;
       m_printf("Opening internal mouse: %s\n", mice->dev);
-      enter_priv_on();  /* The mouse might need special permisions to open (esp r/w). */
-      mice->fd = DOS_SYSCALL(open(mice->dev, O_RDWR | O_NONBLOCK));
-      leave_priv_setting();
-      if (mice->fd == -1) {
- 	mice->intdrv = FALSE;
- 	mice->type = MOUSE_NONE;
- 	mice->add_to_io_select = 0;
+      if (!parent_open_mouse())
  	return;
-      }
-      /* want_sigio causes problems with internal mouse driver */
-      add_to_io_select(mice->fd, mice->add_to_io_select);
-      DOSEMUSetupMouse();
-      memcpy(p,mouse_ver,sizeof(mouse_ver));
-    }
-    else if ((mice->type == MOUSE_PS2) || (mice->type == MOUSE_IMPS2) ||
-	     (mice->type == MOUSE_BUSMOUSE)) {
-      enter_priv_on();
-      mice->fd = open(mice->dev, O_RDWR | O_NONBLOCK);
-      leave_priv_setting();
-      mice->add_to_io_select = 0;
-      if (mice->fd == -1) {
-	  error("Cannot open internal mouse device %s\n",mice->dev);
-	  mice->intdrv = FALSE;
-	  mice->type = MOUSE_NONE;
-	  return;
-      }
-      add_to_io_select(mice->fd, mice->add_to_io_select); 
+
+      fstat(mice->fd, &buf);
+      if (!S_ISFIFO(buf.st_mode) && mice->type != MOUSE_BUSMOUSE && mice->type != MOUSE_PS2)
+        DOSEMUSetupMouse();
+      /* this is only to try to get the initial internal driver two/three
+         button mode state correct; user can override it later. */ 
+      if (mice->type == MOUSE_MICROSOFT || mice->type == MOUSE_MS3BUTTON ||
+        mice->type == MOUSE_BUSMOUSE || mice->type == MOUSE_PS2)
+        mice->has3buttons = FALSE;
+      else
+        mice->has3buttons = TRUE;
       memcpy(p,mouse_ver,sizeof(mouse_ver));
+      iodev_add_device(mice->dev);
     }
     else {
       int x;
@@ -1809,37 +1903,144 @@ mouse_init(void)
       }
     }
   }
-  else {
-    mice->fd = mousepipe;
-    if (mice->fd == -1) {
-      mice->intdrv = FALSE;
-      mice->type = MOUSE_NONE;
-      mice->add_to_io_select = 1;
-      return;
-    }
-    add_to_io_select(mice->fd, mice->add_to_io_select);
-    DOSEMUSetupMouse();
-    memcpy(p,mouse_ver,sizeof(mouse_ver));
+
+  if (mice->type != MOUSE_X) {
+    mouse.init_speed_x = 8;
+    mouse.init_speed_y = 16;
+  } else {
+    /* default to 1 to 1 scaling for apps tha read mickeys */
+    mouse.init_speed_x = 8;
+    mouse.init_speed_y = 8;
   }
   
+  if (mice->intdrv) {
+    pic_seti(PIC_IMOUSE, DOSEMUMouseEvents, 0, do_mouse_irq);
+    pic_unmaski(PIC_IMOUSE);
+  }
+
   /* We set the defaults at the end so that we can test the mouse type */
   mouse_reset(1);		/* Let's set defaults now ! */
+  m_printf("MOUSE: INIT complete\n");
+}
+
+void mouse_post_boot(void)
+{
+  us *ptr;
+        
+  if (!mice->intdrv) return;
+  
+  /* This is needed here to revectoring the interrupt, after dos
+   * has revectored it. --EB 1 Nov 1997 */
+  
+  mouse_enable_internaldriver();
+  
+  /* grab int10 back from video card for mouse */
+  ptr = (us*)((BIOSSEG << 4) +
+              ((long)bios_f000_int10_old - (long)bios_f000));
+  m_printf("ptr is at %p; ptr[0] = %x, ptr[1] = %x\n",ptr,ptr[0],ptr[1]);
+  ptr[0] = IOFF(0x10);
+  ptr[1] = ISEG(0x10);
+  m_printf("after store, ptr[0] = %x, ptr[1] = %x\n",ptr[0],ptr[1]);
+  /* Otherwise this isn't safe */
+  SETIVEC(0x10, INT10_WATCHER_SEG, INT10_WATCHER_OFF);
+}
+
+void mouse_io_callback(void)
+{
+  if (mice->intdrv && mice->fd >= 0) {
+    m_printf("MOUSE: We have data\n");
+    pic_request(PIC_IMOUSE);
+  }
+}
+
+void parent_close_mouse (void)
+{
+  if (mice->intdrv)
+     {
+        if (mice->type == MOUSE_XTERM || mice->type == MOUSE_X
+#ifdef USE_GPM
+	    || mice->type == MOUSE_GPM
+#endif
+	    ) return;
+    
+	if (mice->fd > 0) {
+   	   remove_from_io_select(mice->fd, 1);
+           DOS_SYSCALL(close (mice->fd));
+	}
+    }
+  else
+    child_close_mouse ();
+}
+
+int parent_open_mouse (void)
+{
+  if (mice->intdrv)
+    {
+      struct stat buf;
+      int mode;
+
+      if (mice->type == MOUSE_XTERM || mice->type == MOUSE_X
+#ifdef USE_GPM
+	  || mice->type == MOUSE_GPM
+#endif
+	  ) return 1;
+
+      stat(mice->dev, &buf);
+      mode = O_RDWR | O_NONBLOCK;
+      if (S_ISFIFO(buf.st_mode) || mice->type == MOUSE_BUSMOUSE || mice->type == MOUSE_PS2) {
+	/* no write permission is necessary for FIFO's (eg., gpm) */
+        mode = O_RDONLY | O_NONBLOCK;
+      }
+      mice->fd = -1;
+      /* gpm + non-graphics mode doesn't work */
+      if (!S_ISFIFO(buf.st_mode) || config.vga)
+      {
+        mice->fd = DOS_SYSCALL(open(mice->dev, mode));
+        if (mice->fd == -1) {
+          error("Cannot open internal mouse device %s\n",mice->dev);
+        }
+      }
+      if (mice->fd == -1) {
+ 	mice->intdrv = FALSE;
+ 	mice->type = MOUSE_NONE;
+ 	return 0;
+      }
+      add_to_io_select(mice->fd, 1, mouse_io_callback);
+    }
+  else
+    child_open_mouse ();
+  return 1;
 }
 
 void
-mouse_close(void)
+dosemu_mouse_close(void)
 {
   int result;
 
   if (mice->type == MOUSE_X) return;   
+  if (mice->type == MOUSE_XTERM) {
+    if (sent_mouse_esc) {
+      Mouse_xterm.close();
+      sent_mouse_esc = FALSE;
+    }
+    return;
+  }
+#ifdef USE_GPM
+  if (mice->type == MOUSE_GPM) {
+    Mouse_gpm.close();
+    return;
+  }
+#endif
   
   if (mice->intdrv && mice->fd != -1 ) {
-    m_printf("mouse_close: calling tcsetattr\n");
-    result=tcsetattr(mice->fd, TCSANOW, &mice->oldset);
-    if (result==0)
-       m_printf("mouse_close: tcsetattr ok\n");
-    else
-       m_printf("mouse_close: tcsetattr failed: %s\n",strerror(errno));
+    if (mice->oldset) {
+      m_printf("mouse_close: calling tcsetattr\n");
+      result=tcsetattr(mice->fd, TCSANOW, mice->oldset);
+      if (result==0)
+        m_printf("mouse_close: tcsetattr ok\n");
+      else
+        m_printf("mouse_close: tcsetattr failed: %s\n",strerror(errno));
+    }
     m_printf("mouse_close: closing mouse device, fd=%d\n",mice->fd);
     close(mice->fd);
     m_printf("mouse_close: ok\n");
