@@ -1,6 +1,6 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -36,6 +36,7 @@
  *
  */
 
+#include <stdlib.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
@@ -46,8 +47,7 @@
 #include "emu.h"
 #include "mouse.h"
 
-void mouse_move(void), mouse_lb(void), mouse_rb(void), mouse_mb(void);
-void DOSEMUSetMouseSpeed();
+static void DOSEMUSetMouseSpeed(int old, int new, unsigned cflag);
 
 /*
  * DOSEMUSetupMouse --
@@ -57,9 +57,10 @@ void DOSEMUSetMouseSpeed();
 void
 DOSEMUSetupMouse(void)
 {
-  m_printf("MOUSE: DOSEMUSetupMouse called\n");
-  if ( ! config.usesX ){
-      tcgetattr(mice->fd, &mice->oldset);
+      mouse_t *mice = &config.mouse;
+      m_printf("MOUSE: DOSEMUSetupMouse called\n");
+      mice->oldset = malloc(sizeof(*mice->oldset));
+      tcgetattr(mice->fd, mice->oldset);
       if (mice->type == MOUSE_MOUSEMAN)
         {
           DOSEMUSetMouseSpeed(1200, 1200, mice->flags);
@@ -162,23 +163,12 @@ DOSEMUSetupMouse(void)
         }
 #endif
 #endif
-    /* this is only to try to get the initial internal driver two/three
-    	button mode state correct; user can override it later. */
-    if (mice->type == MOUSE_MICROSOFT || mice->type == MOUSE_MS3BUTTON ||
-    	mice->type == MOUSE_BUSMOUSE ||	mice->type == MOUSE_PS2)
-	mice->has3buttons = FALSE;
-    else
-    	mice->has3buttons = TRUE;
-  }
-  m_printf("MOUSE: INIT complete\n");
 }
  
-void
-DOSEMUMouseProtocol(rBuf, nBytes)
-     unsigned char *rBuf;
-     int nBytes;
+static int DOSEMUMouseProtocol(unsigned char *rBuf, int nBytes)
 {
   int                  i, buttons=0, dx=0, dy=0;
+  mouse_t             *mice = &config.mouse;
   static int           pBufP = 0;
   static unsigned char pBuf[8];
 
@@ -198,7 +188,6 @@ DOSEMUMouseProtocol(rBuf, nBytes)
     {	0xc0,	0x00,	0x00,	0x00,	4	},  /* IMPS/2 mouse */
   };
   
-  if (!config.usesX) {
      for ( i=0; i < nBytes; i++) {
 	/*
 	 * check, if we have a usable data byte
@@ -364,19 +353,18 @@ DOSEMUMouseProtocol(rBuf, nBytes)
 	 */
 	mouse_move_buttons(buttons & 0x04, buttons & 0x02, buttons & 0x01);
 	mouse_move_relative(dx, dy);
-	/*
-	 * check if user events need to be processed
-	 */
-	mouse_event();
+
 	pBufP = 0;
+	return (i + 1);
      }	/* assembly full package */
-  }	/* !config.usesX */
+     return nBytes;
 }
 
-void DOSEMUSetMouseSpeed(int old, int new, unsigned cflag)
+static void DOSEMUSetMouseSpeed(int old, int new, unsigned cflag)
 {
 	struct termios tty;
 	char *c;
+        mouse_t *mice = &config.mouse;
 
         m_printf("MOUSE: set speed %d -> %d\n",old,new);
    
@@ -459,12 +447,39 @@ void DOSEMUMouseEvents(void)
 {
 /* We define a large buffer, because of high overheads with other processes */
 #define MOUSE_BUFFER 1024
-	unsigned char rBuf[MOUSE_BUFFER];
-	int nBytes;
+	static unsigned char rBuf[MOUSE_BUFFER];
+	static int qBeg = 0, qEnd = 0;
+        mouse_t *mice = &config.mouse;
+        
+	int nBytes, nBytesProc;
 
-	nBytes = RPT_SYSCALL(read(mice->fd, (char *)rBuf, sizeof(rBuf)));
-	if (nBytes>0 && !config.usesX) {
-	  m_printf("MOUSE: Read %d bytes\n", nBytes);
-	  DOSEMUMouseProtocol(rBuf, nBytes);
+#ifdef USE_GPM
+	if (mice->type == MOUSE_GPM) {
+	  Mouse_gpm.run();
+	  mouse_event();
+	  return;
+	}
+#endif
+	if (mice->type != MOUSE_X && mice->type != MOUSE_XTERM)
+	  nBytes = RPT_SYSCALL(read(mice->fd, (char *)(rBuf+qEnd),
+	    sizeof(rBuf)-qEnd));
+	else
+	  nBytes = 0;
+	if (nBytes>0)
+	  qEnd += nBytes;
+	if (qBeg < qEnd) {
+	  m_printf("MOUSE: Read %d bytes. %d bytes in queue\n",
+	    nBytes>0 ? nBytes : 0, qEnd-qBeg);
+	  nBytesProc = DOSEMUMouseProtocol(rBuf+qBeg, qEnd-qBeg);
+	  qBeg += nBytesProc;
+	  m_printf("MOUSE: Processed %d bytes. %d bytes still in queue\n",
+	    nBytesProc, qEnd-qBeg);
+	  if (qBeg >= qEnd)
+	    qBeg = qEnd = 0;
+	}
+	mouse_event();
+	if (qBeg < qEnd) {
+	  m_printf("MOUSE: Requesting for the next event\n");
+	  pic_request(PIC_IMOUSE);
 	}
 }
