@@ -1,5 +1,13 @@
+/* 
+ * All modifications in this file to the original code are
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* DANG_BEGIN_MODULE
  * 
+ * REMARK
  * ser_irq.c: Serial interrupt services for DOSEMU
  * Please read the README.serial file in this directory for more info!
  * 
@@ -10,6 +18,7 @@
  * as published by the Free Software Foundation; either version 2 of 
  * the License, or (at your option) any later version.
  *
+ * /REMARK
  * This module is maintained by Mark Rejhon at these Email addresses:
  *      marky@magmacom.com
  *      ag115@freenet.carleton.ca
@@ -17,6 +26,7 @@
  * DANG_END_MODULE
  */
 
+#include <unistd.h>
 #include <stdio.h>
 #include <sys/time.h>
 #include <termios.h>
@@ -24,6 +34,7 @@
 
 #include "config.h"
 #include "emu.h"
+#include "timers.h"
 #include "pic.h"
 #include "serial.h"
 #include "ser_defs.h"
@@ -45,30 +56,22 @@ static int into_irq = 0;
  */
 void serial_timer_update(void)
 {
-  static struct timeval tp;		/* Current timer value */
-  static struct timeval oldtp;		/* Timer value from last call */
-  static long int elapsed;		/* No of 115200ths seconds elapsed */
-  static int i;				/* Loop index */
+  static hitimer_t oldtp = 0;	/* Timer value from last call */
+  hitimer_t tp;			/* Current timer value */
+  unsigned long elapsed;	/* No of 115200ths seconds elapsed */
+  int i;
 
   /* Get system time.  PLEASE DONT CHANGE THIS LINE, unless you can 
    * _guarantee_ that the substitute/stored timer value _is_ up to date 
    * at _this_ instant!  (i.e: vm86s exit time did not not work well)
    */
-  gettimeofday(&tp, NULL);
-
+  tp = GETusTIME(0);
+  if (oldtp==0)	oldtp=tp;
   /* compute the number of 115200ths of seconds since last timer update */
-  elapsed  = (tp.tv_sec - oldtp.tv_sec) * 115200;
-  elapsed += ((tp.tv_usec - oldtp.tv_usec) * 1152) / 10000;
-
-  /* Reset to 0 if the timer had wrapped around back to 0, just in case */
-  if (elapsed < 0) {
-    s_printf("SER: Timer wrapped around back to 0!\n");
-    elapsed = 0;
-  }
+  elapsed = T64DIV((tp-oldtp),8.680555556);
 
   /* Save the old timer values for next time */
-  oldtp.tv_sec  = tp.tv_sec;
-  oldtp.tv_usec = tp.tv_usec;
+  oldtp = tp;
 
   /* Update all the timers */
   for (i = 0; i < config.num_ser; i++) {
@@ -98,12 +101,11 @@ void serial_timer_update(void)
  * and checking if it's time to generate a hardware interrupt (RDI).
  * [num = port]
  */
-void receive_engine(int num)		/* Internal 16550 Receive emulation */ 
+void receive_engine(int num)	/* Internal 16550 Receive emulation */ 
 {
   if (com[num].MCR & UART_MCR_LOOP) return;	/* Return if loopback */
 
-  /* Occasional stack overflows occured when "uart_fill" done inside intr */
-  if (!com[num].int_pend) uart_fill(num);
+  uart_fill(num);
 
   if (com[num].LSR & UART_LSR_DR) {	/* Is data waiting? */
     if (com[num].fifo_enable) {		/* Is it in FIFO mode? */
@@ -133,10 +135,10 @@ void receive_engine(int num)		/* Internal 16550 Receive emulation */
  * of the XMIT FIFO/THR register, and checking if it's time to generate
  * a hardware interrupt (THRI).    [num = port]
  */
-void transmit_engine(int num)	/* Internal 16550 Transmission emulation */
+void transmit_engine(int num) /* Internal 16550 Transmission emulation */
 {
-  static int rtrn;
-  static int control;
+  int rtrn;
+  int control;
 
   #if 0
     /* This is for transmit timer debugging in case it screws up
@@ -158,13 +160,20 @@ void transmit_engine(int num)	/* Internal 16550 Transmission emulation */
     if (!(control & TIOCM_CTS)) return;		/* Return if CTS is low */
   }
   
-  if (com[num].tx_overflow) {		/* Is it in overflow state? */
-    rtrn = RPT_SYSCALL(write(com[num].fd, &com[num].TX, 1));  /* Write port */
-    if (rtrn == 1)                               /* Did it succeed? */
-      com[num].tx_overflow = 0;                  /* Exit overflow state */
-  }
-  else if (com[num].fifo_enable) {	/* Is FIFO enabled? */
+  if (com[num].fifo_enable) {  /* Is FIFO enabled? */
 
+    if (com[num].tx_overflow){
+      if(RPT_SYSCALL(write(com[num].fd,&com[num].tx_buf[com[num].tx_buf_start], 1))!=1){
+        return;  /* write error */
+      } else {
+        /* Squeeze char into FIFO */
+        com[num].tx_buf[com[num].tx_buf_end] = com[num].TX ;
+        /* Update FIFO queue pointers */
+        com[num].tx_buf_end = (com[num].tx_buf_end + 1) % TX_BUFFER_SIZE;
+        com[num].tx_buf_start = (com[num].tx_buf_start + 1) % TX_BUFFER_SIZE;
+        com[num].tx_overflow = 0;                  /* Exit overflow state */
+      }
+    }
     /* Clear as much of the transmit FIFO as possible! */
     while (com[num].tx_buf_bytes > 0) {		/* Any data in fifo? */
       rtrn = RPT_SYSCALL(write(com[num].fd, &com[num].tx_buf[com[num].tx_buf_start], 1));
@@ -182,6 +191,11 @@ void transmit_engine(int num)	/* Internal 16550 Transmission emulation */
     }
   }
   else {					/* Not in FIFO mode */
+    if (com[num].tx_overflow) {        /* Is it in overflow state? */
+      rtrn = RPT_SYSCALL(write(com[num].fd, &com[num].TX, 1));  /* Write port */
+      if (rtrn == 1)                               /* Did it succeed? */
+        com[num].tx_overflow = 0;                  /* Exit overflow state */
+    }
     if (com[num].tx_trigger) { 			/* Is it time to trigger int */
       com[num].tx_trigger = 0;
       com[num].LSRqueued |= UART_LSR_TEMT | UART_LSR_THRE;
@@ -199,7 +213,7 @@ void transmit_engine(int num)	/* Internal 16550 Transmission emulation */
  */
 void modstat_engine(int num)		/* Internal Modem Status processing */ 
 {
-  static int control;
+  int control;
   int newmsr, delta;
 
   /* Return if in loopback mode */
@@ -209,12 +223,16 @@ void modstat_engine(int num)		/* Internal Modem Status processing */
   if (com[num].ms_timer > 0) return;
   com[num].ms_timer += MS_MIN_FREQ;
   
-  ioctl(com[num].fd, TIOCMGET, &control);	/* WARNING: Non re-entrant! */
   
-  newmsr = convert_bit(control, TIOCM_CTS, UART_MSR_CTS) |
-           convert_bit(control, TIOCM_DSR, UART_MSR_DSR) |
-           convert_bit(control, TIOCM_RNG, UART_MSR_RI) |
-           convert_bit(control, TIOCM_CAR, UART_MSR_DCD);
+  if(com[num].pseudo) {
+    newmsr = UART_MSR_CTS | UART_MSR_DSR | UART_MSR_DCD;
+  } else {
+    ioctl(com[num].fd, TIOCMGET, &control);	/* WARNING: Non re-entrant! */
+    newmsr = convert_bit(control, TIOCM_CTS, UART_MSR_CTS) |
+             convert_bit(control, TIOCM_DSR, UART_MSR_DSR) |
+             convert_bit(control, TIOCM_RNG, UART_MSR_RI) |
+             convert_bit(control, TIOCM_CAR, UART_MSR_DCD);
+  }
            
   delta = msr_compute_delta_bits(com[num].MSR, newmsr);
   
@@ -264,7 +282,7 @@ static inline void flag_IIR_linestat(int num)
 }
 
 
-/* This function updates the IIR for Recieve.  [num = port] */
+/* This function updates the IIR for Receive.  [num = port] */
 static inline void flag_IIR_receive(int num)
 {
   /* Update the IIR for RDI status */
@@ -312,8 +330,8 @@ static inline int check_and_update_uart_status(int num)
     com[num].int_condition &= ~LS_INTR;
   }  
 
-  /* Check and updated queued Recieve condition. */
-  /* Safety check, avoid recieve interrupt while DLAB is set high */
+  /* Check and updated queued Receive condition. */
+  /* Safety check, avoid receive interrupt while DLAB is set high */
   if ((com[num].LSRqueued & UART_LSR_DR) && !com[num].DLAB) {
     com[num].rx_timeout = 0;
     com[num].LSR |= UART_LSR_DR;
@@ -367,34 +385,32 @@ static inline int check_and_update_uart_status(int num)
  */
 void serial_int_engine(int num, int int_requested)
 {
-  static u_char tmp;
+  u_char tmp;
   
-  /* Safety code to avoid recieve and transmit while DLAB is set high */
+  /* Safety code to avoid receive and transmit while DLAB is set high */
   if (com[num].DLAB) int_requested &= ~(RX_INTR | TX_INTR);
 
-  /* At this point, we don't much care which function is requested; that  
-   * is taken care of in the serial_int_engine.  However, if interrupts are
-   * disabled, then the Interrupt Identification Register must be set.
-   */
-
-  /* See if a requested interrupt is enabled */
-  if (!com[num].int_pend && com[num].int_enab && com[num].interrupt) {
-    if ((int_requested | com[num].int_condition) & com[num].IER) {
-      if(s3_printf) s_printf("SER%d: Func pic_request intlevel=%d, int_requested=%d\n", 
-                 num, com[num].interrupt, int_requested);
-      /*irq_source_num[com[num].interrupt] = num;*/ /* Permit shared IRQ */
-      com[num].int_pend = 1;			/* Interrupt is now pending */
-      pic_request(com[num].interrupt);		/* Flag PIC for interrupt */
-      return;
-    }
+  if (com[num].tx_trigger) {
+    com[num].tx_timer = 0;
+    transmit_engine(num);
   }
-  /* Else, the below code executes if we aren't going to invoke interrupt */
+  
+  /* reset the served requests */
+  com[num].int_request &= (com[num].int_condition & com[num].IER);
 
-  /* Update the UART status immediately */
   check_and_update_uart_status(num);
+
+  /* and just to be sure... */
+  com[num].int_request &= (com[num].int_condition & com[num].IER);
    
   /* Update the IIR status immediately */
   tmp = (com[num].int_condition & com[num].IER);
+
+  if(s3_printf) s_printf("SER%d: tmp=%d int_cond=%d int_req=%d int=%d\n",
+    num, tmp, com[num].int_condition, com[num].int_request, int_requested);
+  if (int_requested && !(int_requested & com[num].int_condition))
+    s_printf("SER%d: Warning: int_condition is not set for the requested interrupt!\n", num);
+
   if (!tmp)
     flag_IIR_noint(num);		/* No interrupt */
   else if (tmp & LS_INTR)
@@ -405,6 +421,48 @@ void serial_int_engine(int num, int int_requested)
     flag_IIR_transmit(num);		/* Transmit */
   else if (tmp & MS_INTR)
     flag_IIR_modstat(num);		/* Modem Status */
+
+  if (com[num].int_request) {
+    if(int_requested) s_printf("SER%d: Interrupt for another event (%d) is already requested\n",
+	num, com[num].int_request);
+    return;
+  }
+
+  if (!int_requested) {
+    if (!tmp) {
+      /* just an IIR update */
+      return;
+    }
+    else {
+      if(s3_printf) s_printf("SER%d: Requesting serial interrupt upon IIR update\n",
+        num);
+      int_requested = tmp;
+    }
+  }
+
+  /* At this point, we don't much care which function is requested; that  
+   * is taken care of in the serial_int_engine.  However, if interrupts are
+   * disabled, then the Interrupt Identification Register must be set.
+   */
+
+  /* See if a requested interrupt is enabled */
+  if (com[num].int_enab && com[num].interrupt && (int_requested & com[num].IER)) {
+    
+    /* if no int was requested before, request this one */
+    if (!com[num].int_pend) {
+      if(s3_printf) s_printf("SER%d: Func pic_request intlevel=%d, int_requested=%d\n", 
+                 num, com[num].interrupt, int_requested);
+      /*irq_source_num[com[num].interrupt] = num;*/ /* Permit shared IRQ */
+      com[num].int_pend = 1;			/* Interrupt is now pending */
+      pic_request(com[num].interrupt);		/* Flag PIC for interrupt */
+    }
+    else
+      if(s3_printf) s_printf("SER%d: Interrupt already requested (%d)\n", 
+                 num, int_requested);
+  }
+  else
+    if(s3_printf) s_printf("SER%d: Interrupt %d (%d) cannot be requested: enable=%d IER=0x%x\n", 
+        num, com[num].interrupt, int_requested, com[num].int_enab, com[num].IER);
 }  
 
 
@@ -462,28 +520,30 @@ pic_serial_run(void)
              com[num].int_condition);
 
   /* Execute any pre-interrupt code that may be necessary */
-  if (tmp & LS_INTR) {			/* Receiver Line Status Interrupt */
-    flag_IIR_linestat(num);
-    com[num].int_condition &= ~LS_INTR;
-    do_irq();
+  if (!tmp) {
+    flag_IIR_noint(num);		/* No interrupt */
+    com[num].int_request = 0;
   }
-  else if (tmp & RX_INTR) {		/* Received Data Interrupt */
-    flag_IIR_receive(num);
-    com[num].int_condition &= ~RX_INTR;
-    do_irq();
+  else if (tmp & LS_INTR) {
+    flag_IIR_linestat(num);		/* Line Status */
+    com[num].int_request = LS_INTR;
   }
-  else if (tmp & TX_INTR) {		/* Transmit Data Interrupt */
-    flag_IIR_transmit(num);
-    com[num].int_condition &= ~TX_INTR;
-    do_irq();
+  else if (tmp & RX_INTR) {
+    flag_IIR_receive(num);		/* Receive */
+    com[num].int_request = RX_INTR;
   }
-  else if (tmp & MS_INTR) {		/* Modem Status Interrupt */
-    flag_IIR_modstat(num);
-    com[num].int_condition &= ~MS_INTR;
-    do_irq();
+  else if (tmp & TX_INTR) {
+    flag_IIR_transmit(num);		/* Transmit */
+    com[num].int_request = TX_INTR;
   }
+  else if (tmp & MS_INTR) {
+    flag_IIR_modstat(num);		/* Modem Status */
+    com[num].int_request = MS_INTR;
+  }
+
+  if (com[num].int_request)
+    do_irq();
   else {				/* What?  We can't be this far! */
-    flag_IIR_noint(num);
     s_printf("SER%d: Interrupt Error: cancelled serial interrupt!\n",num);  
     /* No interrupt flagged?  Then the interrupt was cancelled sometime
      * after the interrupt was flagged, but before pic_serial_run executed.
@@ -514,6 +574,7 @@ pic_serial_run(void)
    * DANG_FIXTHIS Perhaps this can be modified to limit max chain length?
    */
   into_irq = 0;
+  serial_int_engine(num, 0);              /* Update interrupt status */
   receive_engine(num);
   transmit_engine(num);
   modstat_engine(num);
@@ -541,7 +602,7 @@ pic_serial_run(void)
 void
 serial_run(void)	
 {
-  static int i;
+  int i;
   static int flip = 0;
 
   /* I don't know why, but putting the following code here speeds up serial
@@ -561,6 +622,7 @@ serial_run(void)
    * system if they are called 100x's per second.
    */
   for (i = 0; !into_irq && (i < config.num_ser); i++) {
+    if (com[i].fd < 0) continue;
     receive_engine(i);		/* Receive operations */
     transmit_engine(i);		/* Transmit operations */
     modstat_engine(i);  	/* Modem Status operations */
