@@ -1,6 +1,6 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -33,6 +33,8 @@
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <string.h>
+#include <errno.h>
 
 #include "config.h"
 #include "emu.h"
@@ -75,7 +77,7 @@ static inline void tx_buffer_dump(int num)
 /* This function updates the flow control status depending on buffer condition */
 static inline void flow_control_update(int num)
 {
-  static int control;
+  int control;
   if (com[num].rx_buf_bytes == 0) {			/* buffer empty? */
     control = TIOCM_RTS;
     ioctl(com[num].fd, TIOCMBIS, &control);		/* Raise RTS */
@@ -95,11 +97,11 @@ static inline void flow_control_update(int num)
  */
 static inline void rx_buffer_slide(int num)
 {
-  static int i;
+  int i;
   
   /* Move existing chars in receive buffer to the start of buffer */
   for(i = 0; i < com[num].rx_buf_bytes; i++)
-    com[num].rx_buf[i] = com[num].rx_buf[com[num].rx_buf_start++];
+    com[num].rx_buf[i] = com[num].rx_buf[com[num].rx_buf_start + i];
 
   /* Update start and end pointers in buffer */        
   com[num].rx_buf_start = 0;
@@ -117,7 +119,7 @@ static inline void rx_buffer_slide(int num)
  */
 void uart_fill(int num)
 {
-  static int size;
+  int size = 0;
   
   /* Return if in loopback mode */
   if (com[num].MCR & UART_MCR_LOOP) return;
@@ -164,15 +166,12 @@ void uart_fill(int num)
        * Reset the receive FIFO counter up to a value of 16 (configurable)
        * if we are not within an interrupt.
        */
-      if (!com[num].int_pend) {
-        if (com[num].rx_buf_bytes > com[num].rx_fifo_size)
-          com[num].rx_fifo_bytes = com[num].rx_fifo_size;
-        else
-          com[num].rx_fifo_bytes = com[num].rx_buf_bytes;
-      }
+      com[num].rx_fifo_bytes = com[num].rx_buf_bytes;
+      if (com[num].rx_fifo_bytes > com[num].rx_fifo_size)
+	      com[num].rx_fifo_bytes = com[num].rx_fifo_size;
       
       /* Has it gone above the receive FIFO trigger level? */
-      if (com[num].rx_buf_bytes >= com[num].rx_fifo_trigger) {
+      if (com[num].rx_fifo_bytes >= com[num].rx_fifo_trigger) {
         if(s3_printf) s_printf("SER%d: Func uart_fill requesting RX_INTR\n",num);
         com[num].LSRqueued |= UART_LSR_DR;	/* Update queued LSR */
         serial_int_engine(num, RX_INTR);	/* Update interrupt status */
@@ -181,6 +180,8 @@ void uart_fill(int num)
     }
     /* Else, the following code executes if emulated UART is in 16450 mode. */    
 
+    if(s3_printf) s_printf("SER%d: Got %i bytes, %i in buffer\n",num,
+      size, com[num].rx_buf_bytes);
     com[num].rx_timeout = 0;			/* Reset receive timeout */
     com[num].LSRqueued |= UART_LSR_DR;		/* Update queued LSR */
     if(s3_printf) s_printf("SER%d: Func uart_fill requesting RX_INTR\n",num);
@@ -196,6 +197,8 @@ void uart_fill(int num)
 void uart_clear_fifo(int num, int fifo)
 {
   /* DANG_FIXTHIS Should clearing UART cause THRE int if it's enabled? */
+
+  if(s1_printf) s_printf("SER%d: Clear FIFO.\n",num);
   
   /* Clear the receive FIFO */
   if (fifo & UART_FCR_CLEAR_RCVR) {
@@ -244,7 +247,7 @@ void ser_termios(int num)
 
   /* Return if not a tty */
   if (tcgetattr(com[num].fd, &com[num].newset) == -1) {
-    if(s1_printf) s_printf("SER%d: Line Control: NOT A TTY.\n",num);
+    if(s1_printf) s_printf("SER%d: Line Control: NOT A TTY (%s).\n",num,strerror(errno));
     return;
   }
   s_printf("SER%d: LCR = 0x%x, ",num,com[num].LCR);
@@ -397,7 +400,7 @@ void ser_termios(int num)
   /* The following does the actual system calls to set the line parameters */
   cfsetispeed(&com[num].newset, baud);
   cfsetospeed(&com[num].newset, baud);
-  tcsetattr(com[num].fd, TCSANOW, &com[num].newset);
+  tcsetattr(com[num].fd, TCSADRAIN, &com[num].newset);
 
 #if 0
   /* Many mouse drivers require this, they detect for Framing Errors
@@ -426,7 +429,7 @@ void ser_termios(int num)
  */
 static int get_rx(int num)
 {
-  static int val;
+  int val;
   com[num].rx_timeout = TIMEOUT_RX;		/* Reset timeout counter */
 
   /* If the Received-Data-Ready bit is clear then return a 0
@@ -471,7 +474,8 @@ static int get_rx(int num)
         com[num].int_condition &= ~RX_INTR;	/* Clear receive condition */
 
         /* DANG_FIXTHIS Is this safe to put this here? */
-        serial_int_engine(num, 0);		/* Update interrupt status */
+	/* I don't think so  - SS */
+//        serial_int_engine(num, 0);		/* Update interrupt status */
       }
       else {		
         /* No, the FIFO didn't drop below trigger, but clear timeout bit */
@@ -566,7 +570,7 @@ get_msr(int num)
 static int
 get_lsr(int num)
 {
-  static int val;
+  int val;
 
   val = com[num].LSR;			/* Save old LSR value */
   com[num].int_condition &= ~LS_INTR;	/* RLSI int condition satisfied */
@@ -592,7 +596,7 @@ get_lsr(int num)
  */
 static void put_tx(int num, int val)
 {
-  static int rtrn;
+  int rtrn;
   
   /* Update the transmit timer */
   com[num].tx_timer += com[num].tx_char_time;
@@ -696,10 +700,13 @@ static void put_tx(int num, int val)
     rtrn = RPT_SYSCALL(write(com[num].fd, &val, 1));   /* Attempt char xmit */
     if (rtrn != 1) 				/* Did transmit fail? */
       com[num].tx_overflow = 1; 		/* Set overflow flag */
+    else tcdrain(com[num].fd);
   }
   if (!com[num].fifo_enable ||
        com[num].tx_buf_bytes >= (TX_BUFFER_SIZE-1)) 	/* Is FIFO full? */
       com[num].LSR &= ~UART_LSR_THRE;		/* THR full */
+
+  transmit_engine(num);
 }
 
 
@@ -764,7 +771,10 @@ put_fcr(int num, int val)
 static void
 put_lcr(int num, int val)
 {
-  com[num].LCR = val;			/* Set new LCR value */
+  int changed = com[num].LCR ^ val;    /* bitmask of changed bits */
+  
+  com[num].LCR = val;                  /* Set new LCR value */
+  
   if (val & UART_LCR_DLAB) {		/* Is Baudrate Divisor Latch set? */
     s_printf("SER%d: LCR = 0x%x, DLAB high.\n", num, val);
     com[num].DLAB = 1;			/* Baudrate Divisor Latch flag */
@@ -776,9 +786,22 @@ put_lcr(int num, int val)
      */
   }
   else {
+    s_printf("SER%d: LCR = 0x%x, DLAB low.\n", num, val);
     com[num].DLAB = 0;			/* Baudrate Divisor Latch flag */
     ser_termios(num);			/* Sets new line settings */
   }
+
+  if (changed & UART_LCR_SBC) {
+    /* there is change of break state */
+    if (val & UART_LCR_SBC) {
+      s_printf("SER%d: Setting BREAK state.\n", num);
+      tcdrain(com[num].fd);
+      ioctl(com[num].fd, TIOCSBRK);
+    } else {
+      s_printf("SER%d: Clearing BREAK state.\n", num);
+      ioctl(com[num].fd, TIOCCBRK);
+    }
+  }
 }
 
 
@@ -788,9 +811,9 @@ put_lcr(int num, int val)
 static void
 put_mcr(int num, int val)
 {
-  static int newmsr, delta;
-  static int changed;
-  static int control;
+  int newmsr, delta;
+  int changed;
+  int control;
   changed = com[num].MCR ^ val;			/* Bitmask of changed bits */
   com[num].MCR = val & UART_MCR_VALID;		/* Set valid bits for MCR */
 
@@ -828,7 +851,11 @@ put_mcr(int num, int val)
     if (changed & UART_MCR_LOOP) uart_clear_fifo(num,UART_FCR_CLEAR_CMD);
 
     /* Set interrupt enable flag according to OUT2 bit in MCR */
-    com[num].int_enab = (val & UART_MCR_OUT2);
+    com[num].int_enab = (val & UART_MCR_OUT2) ? 1 : 0;
+    if (com[num].int_enab) {
+      if(s3_printf) s_printf("SER%d: Update interrupt status after MCR update\n",num);
+      serial_int_engine(num, 0);	/* Update interrupt status */
+    }
 
     /* Force RTS & DTR reinitialization if the loopback state has changed */
     if (UART_MCR_LOOP) changed |= UART_MCR_RTS | UART_MCR_DTR;
@@ -933,7 +960,16 @@ put_msr(int num, int val)
 int
 do_serial_in(int num, ioport_t address)
 {
-  static int val;
+  int val;
+
+  /* delayed open happens here */
+  if (com[num].fd == -1) {
+    if (ser_open(num) >= 0)
+      ser_set_params(num);
+  }
+  if (com[num].fd < 0)
+    return 0;
+
   switch (address - com[num].base_port) {
   case UART_RX:		/* Read from Received Byte Register */	
 /*case UART_DLL:*/      /* or Read from Baudrate Divisor Latch LSB */
@@ -1016,6 +1052,15 @@ do_serial_in(int num, ioport_t address)
 int
 do_serial_out(int num, ioport_t address, int val)
 {
+
+  /* delayed open happens here */
+  if (com[num].fd == -1) {
+    if (ser_open(num) >= 0)
+      ser_set_params(num);
+  }
+  if (com[num].fd < 0)
+    return 0;
+
   switch (address - com[num].base_port) {
   case UART_TX:		/* Write to Transmit Holding Register */
 /*case UART_DLL:*/	/* or write to Baudrate Divisor Latch LSB */
@@ -1046,7 +1091,7 @@ do_serial_out(int num, ioport_t address, int val)
         com[num].tx_trigger = 1;
       }
       com[num].IER = (val & 0xF);	/* Write to IER */
-      if(s1_printf) s_printf("SER%d: IER = 0x%x\n", num, val);
+      if(s1_printf) s_printf("SER%d: Write IER = 0x%x\n", num, val);
       serial_int_engine(num, 0);		/* Update interrupt status */
     }
     break;
