@@ -1,9 +1,10 @@
 /* 
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
 
+#include "config.h"
 #include <X11/X.h>
 #include <X11/Xlib.h>
 #include "speaker.h"
@@ -17,14 +18,20 @@ void X_speaker_on(void *gp, unsigned ms, unsigned short period)
 	XKeyboardControl new_state;
 	Display *display = gp;
 
+#if 0
 	/* Turn off the previos sound, I hope */
 	new_state.bell_pitch = 0;  
 	new_state.bell_duration = 0;
 	XChangeKeyboardControl(display, KBBellDuration | KBBellPitch, &new_state);
 	XBell(display, -100);
+#endif
 
-        /* Make the sound I want */
-	new_state.bell_pitch = speaker_period_to_Hz(period);  
+	/* Make the sound I want */
+	new_state.bell_pitch = speaker_period_to_Hz(period);
+	/* Prohibit a crashing X server by not allowing to go higher than
+	 * 32767 kHz. Since the X server interpretes the pitch as a short int
+	 */
+	if (new_state.bell_pitch >= 0x8000) new_state.bell_pitch = 0x7fff;
 	new_state.bell_duration = ms;
 	XChangeKeyboardControl(display, KBBellDuration | KBBellPitch, &new_state);
 	
