@@ -1,3 +1,9 @@
+/* 
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* this is for the DPMI support */
 #ifndef DPMI_H
 #define DPMI_H
@@ -5,7 +11,7 @@
 #define DPMI_VERSION   		0x00	/* major version 0 */
 #define DPMI_DRIVER_VERSION	0x5a	/* minor version 0.90 */
 
-#define DPMI_MAX_CLIENTS	8	/* maximal number of clients */
+#define DPMI_MAX_CLIENTS	32	/* maximal number of clients */
 
 #define DPMI_page_size		4096	/* 4096 bytes per page */
 
@@ -18,56 +24,43 @@
 
 #define DPMI_private_paragraphs	((DPMI_max_rec_rm_func * DPMI_rm_stack_size)>>4)
 					/* private data for DPMI server */
-
-#ifdef __linux__
-#define UCODESEL 0x23
-#define UDATASEL 0x2b
-#endif
-#ifdef __NetBSD__
-#define	UCODESEL GSEL(GUCODE_SEL, SEL_UPL)
-#define	UDATASEL GSEL(GUDATA_SEL, SEL_UPL)
-#endif
-
-EXTERN int in_dpmi INIT(0);        /* Set to 1 when running under DPMI */
 #define current_client (in_dpmi-1)
-EXTERN int in_win31 INIT(0);       /* Set to 1 when running Windows 3.1 */
-EXTERN int dpmi_eflags INIT(0);    /* used for virtuell interruptflag and pending interrupts */
-EXTERN int in_dpmi_dos_int INIT(0);
-EXTERN int in_dpmi_pm_int INIT(0);
-EXTERN int dpmi_mhp_TF INIT(0);
-EXTERN unsigned char dpmi_mhp_intxxtab[256] INIT({0});
+#define DPMI_CLIENT (DPMIclient[current_client])
+#define PREV_DPMI_CLIENT (DPMIclient[current_client-1])
 
-void dpmi_get_entry_point();
+/* Aargh!! Is this the only way we have to know if a signal interrupted
+ * us in DPMI server or client code? */
+#ifdef __linux__
+#define UCODESEL ucodesel
+#define UDATASEL udatasel
+#endif
+
+/* DANG_BEGIN_REMARK
+ * Handling of the virtual interrupt flag is still not correct and there
+ * are many open questions since DPMI specifications are unclear in this
+ * point.
+ * An example: If IF=1 in protected mode and real mode code is called
+ * which is disabling interrupts via cli and returning to protected
+ * mode, is IF then still one or zero?
+ * I guess I have to think a lot about this and to write a small dpmi
+ * client running under a commercial dpmi server :-).
+ * DANG_END_REMARK
+ */
+
+#define dpmi_cli() 	({ dpmi_eflags &= ~IF; pic_cli(); })
+
+#define dpmi_sti() 	({ dpmi_eflags |= IF; is_cli = 0; pic_sti(); })
 
 #ifdef __linux__
-void dpmi_fault(struct sigcontext_struct *);
-#endif
-#ifdef __NetBSD__
-void dpmi_fault(struct sigcontext *, int);
-#endif
-void dpmi_realmode_hlt(unsigned char *);
-void run_pm_int(int);
-void run_pm_mouse();
-void fake_pm_int(void);
-
-#ifdef __linux__
-int dpmi_mhp_regs(void);
-void dpmi_mhp_getcseip(unsigned int *seg, unsigned int *off);
-void dpmi_mhp_getssesp(unsigned int *seg, unsigned int *off);
-int dpmi_mhp_get_selector_size(int sel);
-int dpmi_mhp_getcsdefault(void);
-int dpmi_mhp_setTF(int on);
-void dpmi_mhp_GetDescriptor(unsigned short selector, unsigned long *lp);
-int dpmi_mhp_getselbase(unsigned short selector);
-unsigned long dpmi_mhp_getreg(int regnum);
-void dpmi_mhp_setreg(int regnum, unsigned long val);
+int modify_ldt(int func, void *ptr, unsigned long bytecount);
+#define LDT_WRITE 0x11
 #endif
 
 /* this is used like: SEL_ADR(_ss, _esp) */
 #define SEL_ADR(seg, reg) \
 ({ unsigned long __res; \
   if (!((seg) & 0x0004)) { \
-    /* GTD */ \
+    /* GDT */ \
     __res = (unsigned long) reg; \
   } else { \
     /* LDT */ \
@@ -99,10 +92,6 @@ typedef struct segment_descriptor_s
     unsigned int	used;		/* Segment in use by client # */
 } SEGDESC;
 
-#define MAX_SELECTORS	8192
-
-extern SEGDESC Segments[];
-
 struct RealModeCallStructure {
   unsigned long edi;
   unsigned long esi;
@@ -136,9 +125,73 @@ typedef struct dpmi_pm_block_stuct {
   struct   dpmi_pm_block_stuct *next;
   unsigned long handle;
   unsigned long size;
-  void     *base;
+  char     *base;
 } dpmi_pm_block;
 
+struct DPMIclient_struct {
+  struct sigcontext_struct stack_frame;
+  int is_32;
+  dpmi_pm_block *pm_block_root;
+  /* for real mode call back, DPMI function 0x303 0x304 */
+  RealModeCallBack realModeCallBack[0x10];
+  INTDESC Interrupt_Table[0x100];
+  INTDESC Exception_Table[0x20];
+  unsigned short LDT_ALIAS;
+  unsigned short PMSTACK_SEL;	/* protected mode stack selector */
+  unsigned short DPMI_SEL;
+  /* used when passing a DTA higher than 1MB */
+  unsigned short USER_DTA_SEL;
+  unsigned long USER_DTA_OFF;
+  unsigned short USER_PSP_SEL;
+  void *user_FCB;
+  unsigned short CURRENT_PSP;
+  unsigned short CURRENT_ENV_SEL;
+  unsigned short PARENT_PSP;
+  unsigned short PARENT_ENV_SEL;
+};
+extern struct DPMIclient_struct DPMIclient[DPMI_MAX_CLIENTS];
+
+EXTERN volatile int in_dpmi INIT(0);/* Set to 1 when running under DPMI */
+EXTERN volatile int in_win31 INIT(0); /* Set to 1 when running Windows 3.1 */
+EXTERN volatile int in_dpmi_dos_int INIT(0);
+EXTERN volatile int dpmi_mhp_TF INIT(0);
+EXTERN unsigned char dpmi_mhp_intxxtab[256] INIT({0});
+EXTERN volatile int is_cli INIT(0);
+
+extern unsigned short DPMI_private_data_segment;
+extern unsigned long dpmi_free_memory; /* how many bytes memory client */
+				       /* can allocate */
+extern unsigned long pm_block_handle_used;       /* tracking handle */
+extern SEGDESC Segments[];
+extern struct sigcontext_struct _emu_stack_frame;
+/* used to store the dpmi client registers */
+extern RealModeCallBack mouseCallBack; /* user\'s mouse routine */
+extern char *ldt_buffer;
+
+void dpmi_get_entry_point(void);
+void indirect_dpmi_switch(struct sigcontext_struct *);
+#ifdef __linux__
+void dpmi_fault(struct sigcontext_struct *);
+#endif
+void dpmi_realmode_hlt(unsigned char *);
+void run_pm_int(int);
+void fake_pm_int(void);
+
+#ifdef __linux__
+int dpmi_mhp_regs(void);
+void dpmi_mhp_getcseip(unsigned int *seg, unsigned int *off);
+void dpmi_mhp_getssesp(unsigned int *seg, unsigned int *off);
+int dpmi_mhp_get_selector_size(int sel);
+int dpmi_mhp_getcsdefault(void);
+int dpmi_mhp_setTF(int on);
+void dpmi_mhp_GetDescriptor(unsigned short selector, unsigned long *lp);
+int dpmi_mhp_getselbase(unsigned short selector);
+unsigned long dpmi_mhp_getreg(int regnum);
+void dpmi_mhp_setreg(int regnum, unsigned long val);
+void dpmi_mhp_modify_eip(int delta);
+#endif
+
+void add_cli_to_blacklist(void);
 dpmi_pm_block* DPMImalloc(unsigned long size);
 dpmi_pm_block* DPMImallocFixed(unsigned long base, unsigned long size);
 int DPMIfree(unsigned long handle);
@@ -149,54 +202,24 @@ dpmi_pm_block *lookup_pm_block(unsigned long h);
 int
 DPMIMapConventionalMemory(dpmi_pm_block *block, unsigned long offset,
 			  unsigned long low_addr, unsigned long cnt);
+unsigned long dpmi_GetSegmentBaseAddress(unsigned short selector);
+unsigned long GetSegmentBaseAddress(unsigned short);
+unsigned long GetSegmentLimit(unsigned short);
+void CheckSelectors(struct sigcontext_struct *scp);
+int ValidSelector(unsigned short selector);
+int ValidAndUsedSelector(unsigned short selector);
 
-#define DPMI_show_state \
-    D_printf("eip: 0x%08lx  esp: 0x%08lx  eflags: 0x%08lx\n" \
-	     "trapno: 0x%02lx  errorcode: 0x%08lx  cr2: 0x%08lx\n" \
-	     "cs: 0x%04x  ds: 0x%04x  es: 0x%04x  ss: 0x%04x  fs: 0x%04x  gs: 0x%04x\n", \
-	     _eip, _esp, _eflags, _trapno, _err, _cr2, _cs, _ds, _es, _ss, _fs, _gs); \
-    D_printf("EAX: %08lx  EBX: %08lx  ECX: %08lx  EDX: %08lx\n", \
-	     _eax, _ebx, _ecx, _edx); \
-    D_printf("ESI: %08lx  EDI: %08lx  EBP: %08lx\n", \
-	     _esi, _edi, _ebp); \
-    /* display the 10 bytes before and after CS:EIP.  the -> points \
-     * to the byte at address CS:EIP \
-     */ \
-    if (!((_cs) & 0x0004)) { \
-      /* GTD */ \
-      csp2 = (unsigned char *) _eip - 10; \
-    } \
-    else { \
-      /* LDT */ \
-      csp2 = (unsigned char *) (GetSegmentBaseAddress(_cs) + _eip) - 10; \
-    } \
-    D_printf("OPS  : "); \
-    for (i = 0; i < 10; i++) \
-      D_printf("%02x ", *csp2++); \
-    D_printf("-> "); \
-    for (i = 0; i < 10; i++) \
-      D_printf("%02x ", *csp2++); \
-    D_printf("\n"); \
-    if (!((_ss) & 0x0004)) { \
-      /* GTD */ \
-      ssp2 = (unsigned char *) _esp - 10; \
-    } \
-    else { \
-      /* LDT */ \
-      if (Segments[_ss>>3].is_32) \
-	ssp2 = (unsigned char *) (GetSegmentBaseAddress(_ss) + _esp ) - 10; \
-      else \
-	ssp2 = (unsigned char *) (GetSegmentBaseAddress(_ss) + _LWORD(esp) ) - 10; \
-    } \
-    D_printf("STACK: "); \
-    for (i = 0; i < 10; i++) \
-      D_printf("%02x ", *ssp2++); \
-    D_printf("-> "); \
-    for (i = 0; i < 10; i++) \
-      D_printf("%02x ", *ssp2++); \
-    D_printf("\n");
-
+extern char *DPMI_show_state(struct sigcontext_struct *scp);
 extern void dpmi_sigio(struct sigcontext_struct *scp);
 extern void run_dpmi(void);
 
+extern int ConvertSegmentToDescriptor(unsigned short segment);
+extern int SetSegmentBaseAddress(unsigned short selector,
+					unsigned long baseaddr);
+extern int SetSegmentLimit(unsigned short, unsigned int);
+extern void save_pm_regs(struct sigcontext_struct *);
+extern void restore_pm_regs(struct sigcontext_struct *);
+extern unsigned short AllocateDescriptors(int);
+extern int FreeDescriptor(unsigned short selector);
+
 #endif /* DPMI_H */
