@@ -1,3 +1,10 @@
+/* 
+ * All modifications in this file to the original code are
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* 
    Unix SMB/Netbios implementation.
    Version 1.9.
@@ -18,22 +25,27 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
+/*
+Modified by O.V.Zhirov, July 1998
+
+*/
+
+
+
 #ifdef DOSEMU
 #include "mangle.h"
+#include "mfs.h"
+#include "dos2linux.h"
+#include "emu.h"
 #else
 #include "includes.h"
 #include "loadparm.h"
 #endif
 
-extern int DEBUGLEVEL;
-extern int case_default;
-extern BOOL case_mangle;
-
-
 /****************************************************************************
 provide a checksum on a string
 ****************************************************************************/
-int str_checksum(char *s)
+static int str_checksum(char *s)
 {
   int res = 0;
   int c;
@@ -48,37 +60,114 @@ int str_checksum(char *s)
 }
 
 /****************************************************************************
-return True if a name is a special msdos reserved name
+check if a name is a special msdos reserved name:
+the name is either a full Unix name or an 8 character candidate
 ****************************************************************************/
-BOOL is_reserved_msdos(char *fname)
+unsigned long is_dos_device(const char *fname)
 {
-  char upperFname[13];
   char *p;
+  unsigned char *dev;
+  unsigned long devfar;
+  int i;
 
-  StrnCpy (upperFname, fname, 12);
+  /*
+   * LPTx e.t.c. is reserved no matter the path (e.g. .\LPT1 _is_ reserved),
+   * but mfs.c sends an entire path.
+   *             -- Rob Clark <rclark@list-clark.com> 2000/05/09
+   */
+  if (fname[0] == '/')
+  {
+    d_printf("MFS: is_msdos_device %s\n", fname);
+    p = strrchr(fname,'/');
+    if (p) fname = p+1;	/* strip the directory part */
+  }
+  else
+  {
+    d_printf("MFS: is_msdos_device %.8s\n", fname);
+  }
 
-  /* lpt1.txt and con.txt etc are also illegal */
-  p=strchr(upperFname,'.');
-  if (p)
-   *p='\0';
-  strupper (upperFname);
-  if ((strcmp(upperFname,"CLOCK$") == 0) ||
-    (strcmp(upperFname,"CON") == 0) ||
-    (strcmp(upperFname,"AUX") == 0) ||
-    (strcmp(upperFname,"COM1") == 0) ||
-    (strcmp(upperFname,"COM2") == 0) ||
-    (strcmp(upperFname,"COM3") == 0) ||
-    (strcmp(upperFname,"COM4") == 0) ||
-    (strcmp(upperFname,"LPT1") == 0) ||
-    (strcmp(upperFname,"LPT2") == 0) ||
-    (strcmp(upperFname,"LPT3") == 0) ||
-    (strcmp(upperFname,"NUL") == 0) ||
-    (strcmp(upperFname,"PRN") == 0))
-      return (True) ;
+  /*
+   * The below (now #ifdef'ed out) code strips the extension off for checking.
+   * But the assumption, that file names with a reserved basename remain
+   * reserved even if they have an extension, seems not to be true.
+   * For example CLIPPER-programs _are_ using file names such as
+   * 'LPTx.PRN' for internal spooling purposes.
+   *             -- Bernd Schueler b.schueler@gmx.de 2001/02/23
+   * well, as far as I know the extension can be stripped off;
+   * most likely the bug was somewhere else - we must not strip
+   * the extension off for further processing Bart 2002/08/12
+   */
 
-  return (False);
+  /* walk the chain of DOS devices; see also FreeDOS kernel code */
+  dev = &lol[lol_nuldev_off];
+  devfar = (((unsigned long)dev - 0x26) << 12) | 0x26;
+  do
+  {
+    for (i = 0; i < 8; i++)
+    {
+      char c1 = fname[i];
+      if (c1 == '.' || c1 == '\0')
+      {
+        /* check if remainder of device name consists of spaces or nulls */
+        for (; i < 8; i++)
+        {
+          char c2 = dev[0xa + i];
+          if (c2 != ' ' && c2 != '\0')
+            break;
+        }
+        break;
+      }
+      if (toupperDOS(c1) != toupperDOS(dev[0xa + i]))
+        break;
+    }
+    if (i == 8)
+      return devfar;
+    if (dev[0] == 0xff || dev[1] == 0xff)
+      return 0;
+    memcpy(&devfar, dev, 4);
+    dev = (char *)((((dev[3] << 8) | dev[2]) << 4) + ((dev[1] << 8) | dev[0]));
+  } while (1);
 }
 
+/* input: fpath = unix path
+          current_drive = drive for DOS path
+          alias=1: mangle, alias=0: don't mangle
+   output: dest = DOS path
+*/
+void make_unmake_dos_mangled_path(char *dest, char *fpath,
+					 int current_drive, int alias)
+{
+	char *src;
+	*dest++ = current_drive + 'A';
+	*dest++ = ':';
+	*dest = '\\';
+	src = fpath + strlen(drives[current_drive].root);
+	if (*src == '/') src++;
+	while (src != NULL && *src != '\0') {
+		char *src2 = strchr(src, '/');
+		if (src2 == src) break;
+		if (src - 1 > fpath)
+			src[-1] = '\0';
+		if (src2 != NULL)
+			*src2++ = '\0';
+		dest++;
+		d_printf("LFN: src=%s len=%d\n", src, strlen(src));
+		if (!strcmp(src, "..") || !strcmp(src, ".")) {
+			strcpy(dest, src);
+		} else {
+			if (alias) {
+				name_convert(dest, src, MANGLE, NULL);
+			}
+		}
+		dest += strlen(dest);
+		*dest = '\\';
+		if (src - 1 > fpath)
+			src[-1] = '/';
+		src = src2;
+	}
+	if (dest[-1] == ':') dest++;
+	*dest = '\0';
+}
 
 BOOL valid_dos_char[256];
 
@@ -87,7 +176,7 @@ BOOL valid_dos_char[256];
 /****************************************************************************
 initialise the valid dos char array
 ****************************************************************************/
-void valid_initialise(void)
+static void valid_initialise(void)
 {
   static BOOL initialised = False;    
   int i;
@@ -95,31 +184,8 @@ void valid_initialise(void)
   if (initialised)
     return;
 
-  for (i=0;i<128;i++)
-    valid_dos_char[i] = (isalnum(i) || strchr("._^$~!#%&-{}()@'`",i));
-  for (i=128;i<256;i++) 
-    valid_dos_char[i] = False;	
-  
-  /*    When the special char exists both cases, windows uses
-	the uppercase code for DOS filename */
-  
-  /* any code page, low and uppercase, valid for all DOS PC */
-  valid_dos_char[142] = True;     /* A trema      */
-  valid_dos_char[143] = True;     /* A o          */
-  valid_dos_char[144] = True;     /* E '          */
-  valid_dos_char[146] = True;     /* AE           */
-  valid_dos_char[153] = True;     /* O trema      */
-  valid_dos_char[154] = True;     /* U trema      */
-  valid_dos_char[165] = True;     /* N tilda      */
-  
-  /* any code page, valid for all PC */
-  valid_dos_char[128] = True;     /* C cedille    */
-  valid_dos_char[156] = True;     /* Pound        */
-  
-  /* code page 850 (most common)*/
-  valid_dos_char[183] = True;     /* A `     (WIN)*/
-  valid_dos_char[157] = True;     /* Phi     (WIN)*/
-  valid_dos_char[212] = True;     /* E`      (WIN)*/
+  for (i=0;i<256;i++)
+    valid_dos_char[i] = is_valid_DOS_char(i);
 
 #ifdef DOSEMU
   valid_dos_char['?'] = True;
@@ -131,7 +197,7 @@ void valid_initialise(void)
 /****************************************************************************
 return True if a name is in 8.3 dos format
 ****************************************************************************/
-BOOL is_8_3(char *fname)
+static BOOL is_8_3(char *fname)
 {
   int len = strlen(fname);
   char *dot_pos = strchr(fname,'.');
@@ -139,25 +205,10 @@ BOOL is_8_3(char *fname)
 
   DEBUG(5,("checking %s for 8.3\n",fname));
 
-  if (case_mangle)
-    switch (case_default)
-      {
-      case CASE_LOWER:
-	if (strhasupper(fname)) return(False);
-	break;
-      case CASE_UPPER:
-	if (strhaslower(fname)) return(False);
-	break;
-      }
-
   /* can't be longer than 12 chars */
   if (len == 0 || len > 12)
     return(False);
 
-  /* can't be an MS-DOS Special file such as lpt1 or even lpt1.txt */
-  if (is_reserved_msdos(fname))
-    return(False);
-
   /* can't contain invalid dos chars */
   /* Windows use the ANSI charset.
      But filenames are translated in the PC charset.
@@ -244,7 +295,7 @@ int mangled_stack_len = 0;
 /****************************************************************************
 create the mangled stack
 ****************************************************************************/
-void create_mangled_stack(int size)
+static void create_mangled_stack(int size)
 {
   if (mangled_stack)
     {
@@ -260,7 +311,7 @@ void create_mangled_stack(int size)
 /****************************************************************************
 push a mangled name onto the stack
 ****************************************************************************/
-void push_mangled_name(char *s)
+static void push_mangled_name(char *s)
 {
   int i;
   char *p;
@@ -283,7 +334,7 @@ void push_mangled_name(char *s)
 	      sizeof(fstring)*MIN(mangled_stack_len,mangled_stack_size-1));
   strcpy(mangled_stack[0],s);
   p = strrchr(mangled_stack[0],'.');
-  if (p && (!strhasupper(p+1)) && (strlen(p+1) < 4))
+  if (p && (!strhasupperDOS(p+1)) && (strlen(p+1) < 4))
     *p = 0;
   mangled_stack_len = MIN(mangled_stack_size,mangled_stack_len+1);
 }
@@ -305,7 +356,7 @@ BOOL check_mangled_stack(char *s, char *MangledMap)
     {
       check_extension = True;
       StrnCpy(extension,p,4);
-      strlower(extension); /* XXXXXXX */
+      strlowerDOS(extension); /* XXXXXXX */
     }
 
   for (i=0;i<mangled_stack_len;i++)
@@ -353,7 +404,7 @@ BOOL is_mangled(char *s)
   char *m = strchr(s,magic_char);
   if (!m) return(False);
 
-  /* we use two base 36 chars efore the extension */
+  /* we use two base 36 chars before the extension */
   if (m[1] == '.' || m[1] == 0 ||
       m[2] == '.' || m[2] == 0 ||
       (m[3] != '.' && m[3] != 0))
@@ -408,14 +459,14 @@ void mangle_name_83(char *s, char *MangledMap)
     }
       
 
-  strupper(s);
+  strupperDOS(s);
 
   DEBUG(5,("Mangling name %s to ",s));
 
   if (p)
     {
       if (p == s)
-	strcpy(extension,"___");
+	strcpy(extension,"___");                   /* ??? */
       else
 	{
 	  *p++ = 0;
@@ -616,7 +667,7 @@ BOOL do_fwd_mangled_map(char *s, char *MangledMap)
       *np++ = '\0';             /* NULL terminate it. */
       DEBUG(5,("End of second in pair '%s'\n", end));
       strcpy(s, new_string);    /* Substitute with the new name. */
-      strupper(s);
+      strupperDOS(s);
       DEBUG(5,("s is now '%s'\n", s));
       return True;              /* DONE! */
     } else {
@@ -628,6 +679,8 @@ BOOL do_fwd_mangled_map(char *s, char *MangledMap)
   return False;
 }
 
+/******************************************************/
+
 
 #ifndef DOSEMU
 static BOOL do_rev_mangled_map(char *s, char *MangledMap)
@@ -839,7 +892,7 @@ void unix_convert_83(char *s,char *home,BOOL mangle, char *MangledMap)
       else
 	strcpy(name,s);
 
-      strupper(name);
+      strupperDOS(name);
       /* now search through the directory for a matching name */
       mangle_search_83(name,home,directory, MangledMap);
       string_replace(name,magic_char,1);
