@@ -1,3 +1,10 @@
+/* 
+ * All modifications in this file to the original code are
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
+ *
+ * for details see file COPYING in the DOSEMU distribution
+ */
+
 /* modified for dosemu */
 /*
  * Copyright (c) 1991 Carnegie Mellon University
@@ -26,13 +33,10 @@
 /*
  * DANG_BEGIN_MODULE
  *
+ * REMARK
  * This provides the EMM Memory Management for DOSEMU. It was originally part
  * of the Mach Dos Emulator. 
  *
- * Recent work in this area has involved a patch to the Kernel. If this is used
- * and the DEFINE MMAP_EMS line used, a faster form of EMS memory support is
- * included, using the /proc filesystem.
- * 
  * In contrast to some of the comments (Yes, _I_ know the adage about that...)
  * we appear to be supporting EMS 4.0, not 3.2.  The following EMS 4.0
  * functions are not supported (yet): 0x4f (partial page map functions),
@@ -40,85 +44,33 @@
  * support is missing, and raw page size is 16k (instead of 4k).  Other than
  * that, EMS 4.0 support appears complete.
  *
+ * /REMARK
  * DANG_END_MODULE
  *
  * Purpose:
  *	Mach EMM Memory Manager
  *
- * HISTORY: (DANG_BEGIN_CHANGELOG
- * $Log: emm.c,v $
- *
- * DANG_END_CHANGELOG)
  */
 
-#if defined(__linux__) || defined(__NetBSD__)
-
 #include <unistd.h>
 #include <string.h>
-#include <sys/errno.h>
-#include <sys/mman.h>
-#include <malloc.h>
+#include <errno.h>
+#include <stdlib.h>
 #include <fcntl.h>
 #include <sys/types.h>
 #include <sys/utsname.h>
-/*#ifdef __NetBSD__*/
 #define new_utsname utsname
-/*#endif*/
 
 #include "emu.h"
 #include "memory.h"
 #include "machcompat.h"
+#include "mapping.h"
 
 static inline boolean_t unmap_page(int);
 
-#else
-#include "base.h"
-#include "bios.h"
-#endif
-
 #include <sys/file.h>
 #include <sys/ioctl.h>
-
-#if defined(__NetBSD__) || defined(__linux__)
-#define __freeunices__
-#endif
-
-#ifdef __linux__  /* we try to avoid syscalls over lib */
-
-/* NOTE: Do not optimize higher then  -O2, else GCC will optimize away what we
-         expect to be on the stack */
-caddr_t libless_mmap(caddr_t addr, size_t len,
-		int prot, int flags, int fd, off_t off) {
-	int __res;
-	__asm__ __volatile__("int $0x80\n"
-	:"=a" (__res):"a" ((int)90), "b" ((int)&addr));
-	if (((unsigned)__res) > ((unsigned)-4096)) {
-		errno = -__res;
-		__res=-1;
-	}
-	else errno = 0;
-	return (caddr_t)__res;
-}
-#undef mmap
-#define mmap libless_mmap
-
-static int libless_munmap(caddr_t addr, size_t len)
-{
-	int __res;
-	__asm__ __volatile__("int $0x80\n"
-	:"=a" (__res):"a" ((int)91), "b" ((int)addr), "c" ((int)len));
-	if (__res < 0) {
-		errno = -__res;
-		__res=-1;
-	}
-	else errno =0;
-	return __res;
-}
-#undef munmap
-#define munmap libless_munmap
-
-#endif
-
+#include "priv.h"
 
 
 /*****************************************************************************/
@@ -192,6 +144,9 @@ static int libless_munmap(caddr_t addr, size_t len)
 #define	MAX_HANDLES	255	/* must fit in a byte */
 /* this is in EMS pages, which MAX_EMS (defined in Makefile) is in K */
 #define MAX_EMM		(config.ems_size / 16)
+#ifndef PAGE_SIZE
+#define PAGE_SIZE	4096
+#endif
 #define	EMM_PAGE_SIZE	(16*1024)
 #define EMM_MAX_PHYS	4
 #define NULL_HANDLE	-1
@@ -216,7 +171,7 @@ static int libless_munmap(caddr_t addr, size_t len)
 #define EMM_FEAT_NOSUP	0x91
 #define EMM_MOVE_OVLAP	0x92
 #define EMM_MOVE_OVLAPI	0x97
-#define EMM_NOT_FOUND	0xa1
+#define EMM_NOT_FOUND	0xa0  /* 971120 <ki@kretz.co.at> acc to R.Brown's int list */
 
 #define EMM_ERROR -1
 static u_char emm_error;
@@ -264,18 +219,11 @@ static u_short os_allow=1;
 
 /* FIXME -- inline function */
 #define CLEAR_HANDLE_NAME(nameptr) \
-	bzero((nameptr), 9);
+	memset((nameptr), 0, 9);
 
 #define SET_HANDLE_NAME(nameptr, name) \
 	{ memmove((nameptr), (name), 8); nameptr[8]=0; }
 
-#if 0
-#define CHECK_HANDLE(handle) \
-  if ((handle) < 0 || (handle) > MAX_HANDLES) { \
-    SETHIGH(&(state->eax), EMM_INV_HAN); \
-    return; \
-  }
-#else
 #define CHECK_HANDLE(handle) \
   if ((handle < 0) || (handle > MAX_HANDLES) || \
       (handle_info[handle].active == 0)) { \
@@ -284,7 +232,6 @@ static u_short os_allow=1;
     SETHIGH(&(state->eax), EMM_INV_HAN); \
     return; \
   }
-#endif
 
 /* this will have to change...0 counts are allowed */
 #define HANDLE_ALLOCATED(handle) \
@@ -301,16 +248,11 @@ static u_short os_allow=1;
 #define PHYS_PAGE_ADDR(i) \
   (PHYS_PAGE_SEGADDR(i) << 4)
 
-#ifdef __freeunices__
 #define E_Stub(arg1, s, a...)   E_printf("EMS: "s, ##a)
 #define Kdebug0(args)		E_Stub args
 #define Kdebug1(args)		E_Stub args
 #define Kdebug2(args)		E_Stub args
 
-static int     selfmem_fd = -1;
-static boolean ems_mmap = 0;
-#endif
-
 void
 ems_helper(void) {
   u_char *rhptr;		/* request header pointer */
@@ -355,43 +297,11 @@ ems_init(void)
 {
   int sh_base;
   int j;
-  struct new_utsname unames;
-
-#ifdef __freeunices__
 
   if (!config.ems_size)
     return;
 
-#endif
-
-  /* we need release 1.1.43 or higher to use the mmap stuff */
-#ifdef __linux__
-#define MIN_RELEASE "1.1.43"
-#define MEMFILE "/proc/self/mem"
-#endif
-#ifdef __NetBSD__
-#undef MIN_RELEASE /* processes have no permissions to open their own mem? */
-#define MEMFILE "/proc/curproc/mem"
-#endif
-  uname(&unames);
-  E_printf("EMS: coming up on release %s\n", unames.release);
-#ifdef MIN_RELEASE
-  if (strcmp(unames.release, MIN_RELEASE) >= 0) {
-    E_printf("EMS: using " MEMFILE " mmaping (yeah!!!)\n");
-    ems_mmap = 1;
-  }
-#endif
-
-  if (ems_mmap) {
-     E_printf("EMS: opening " MEMFILE "\n");
-
-     selfmem_fd = open(MEMFILE, O_RDWR);
-     if (selfmem_fd < 0) {
-       error("EMS: cannot open " MEMFILE ": %s\n",strerror(errno));
-       ems_mmap = 0;
-     }
-  }
-
+  open_mapping(MAPPING_EMS);
   E_printf("EMS: initializing memory\n");
 
   for (sh_base = 0; sh_base < EMM_MAX_PHYS; sh_base++) {
@@ -404,13 +314,6 @@ ems_init(void)
   }
 
   /* should set up OS handle here */
-#if 0
-  /* Norton sysinfo reacts badly to having a system handle with a size... */
-  handle_info[OS_HANDLE].numpages = OS_PAGES;
-  handle_info[OS_HANDLE].object = (mach_port_t) OS_PORT;
-  handle_info[OS_HANDLE].objsize = OS_SIZE;
-  handle_info[OS_HANDLE].active = 1;
-#else
   handle_info[OS_HANDLE].numpages = 0;
   handle_info[OS_HANDLE].object = 0;
   handle_info[OS_HANDLE].objsize = 0;
@@ -418,7 +321,6 @@ ems_init(void)
   for (j = 0; j < EMM_MAX_PHYS; j++) {
     handle_info[OS_HANDLE].saved_mappings_handle[j] = NULL_PAGE;
   }
-#endif
 
   handle_total++;
   SET_HANDLE_NAME(handle_info[OS_HANDLE].name, "SYSTEM  ");
@@ -427,33 +329,34 @@ ems_init(void)
   memcheck_reserve('E', EMM_BASE_ADDRESS, EMM_MAX_PHYS * EMM_PAGE_SIZE);
 }
 
-#ifdef __freeunices__
 static mach_port_t
 new_memory_object(size_t bytes)
 {
-  /* NOTE: bytes has to be a multiple of PAGE-SIZE,
-   * else we go into trouble with mmap()-ing under Linux >= 1.3.78
-   * We use valloc(), because this is the correct way to get page-aligned
-   * chunks of memory ( malloc() of libc >5.3 won't do the job ).
-   */
-  mach_port_t addr = (mach_port_t) valloc(bytes);
+  mach_port_t addr;
 
+  addr = alloc_mapping(MAPPING_EMS, bytes, 0);
+  if (!addr) return 0;
   E_printf("EMS: allocating 0x%08x bytes @ %p\n", bytes, (void *) addr);
   return (addr);		/* allocate on a PAGE boundary */
 }
 
 static inline void
-destroy_memory_object(mach_port_t object)
+destroy_memory_object(mach_port_t object, int length)
 {
   E_printf("EMS: destroyed EMS object @ %p\n", (void *) object);
-  free(object);
+  free_mapping(MAPPING_EMS, object, length);
 }
 
-#endif /* __freeunices__ */
+static inline mach_port_t realloc_memory_object(mach_port_t object, size_t oldsize, size_t bytes)
+{
+  void *addr = (mach_port_t)realloc_mapping(MAPPING_EMS, (void *)object,
+  	oldsize, bytes);
+  if ((int)addr == -1) return 0;
+  return (mach_port_t)addr;
+}
 
 static int
-allocate_handle(pages_needed)
-     int pages_needed;
+allocate_handle(int pages_needed)
 {
   int i, j;
   mach_port_t obj;
@@ -498,8 +401,7 @@ allocate_handle(pages_needed)
 }
 
 static boolean_t
-deallocate_handle(handle)
-     int handle;
+deallocate_handle(int handle)
 {
   int numpages, i;
   mach_port_t object;
@@ -514,7 +416,7 @@ deallocate_handle(handle)
   }
   numpages = handle_info[handle].numpages;
   object = handle_info[handle].object;
-  destroy_memory_object(object);
+  destroy_memory_object(object,numpages*EMM_PAGE_SIZE);
   handle_info[handle].numpages = 0;
   handle_info[handle].active = 0;
   handle_info[handle].object = MACH_PORT_NULL;
@@ -524,42 +426,21 @@ deallocate_handle(handle)
   return (TRUE);
 }
 
-#define TOUCH_EMM_PAGE(address) ({ \
-	int i; \
-	volatile char *p= (volatile char *)address; \
-	for (i=EMM_PAGE_SIZE-4096; i>=0 ; i-=4096) *(p+i) /* |= *(p+i) */; \
-})
-
-
 static void _do_map_page(caddr_t base, caddr_t logical)
 {
-  if (ems_mmap) {
-    /* Touch all pages before mmap()-ing,
-     * else Linux >= 1.3.78 will return -EINVAL. (Hans, 96/04/16)
-     */
-    TOUCH_EMM_PAGE(logical);
-    E_printf("EMS: mmap()ing from 0x%x to 0x%x\n", (int)base, (int)logical);
-    
-    if ((caddr_t)base != mmap(base,
+  E_printf("EMS: mmap()ing from 0x%x to 0x%x\n", (int)base, (int)logical);
+  
+  if ((caddr_t)base != mmap_mapping(MAPPING_EMS|MAPPING_ALIAS, base,
 			      EMM_PAGE_SIZE,
 			      PROT_READ | PROT_WRITE | PROT_EXEC,
-			      MAP_SHARED | MAP_FIXED,
-			      selfmem_fd,
-			      (off_t) logical))
-    {
-        E_printf("EMS: mmap() failed: %s\n",strerror(errno));
-        leavedos(2);
-    }
-  } else {
-    E_printf("EMS: COPY-MAP l=0x%x, bs=0x%x, d=0x%02x, s=0x%02x\n", 
-	     (int)logical, (int)base, *(u_short *)base, *(u_short *) logical);
-    memmove((u_char *) base, (u_char *) logical, EMM_PAGE_SIZE);
+			      (void *) logical)) {
+    E_printf("EMS: mmap() failed: %s\n",strerror(errno));
+    leavedos(2);
   }
 }
 
 static boolean_t
-__map_page(physical_page)
-     int physical_page;
+__map_page(int physical_page)
 {
   int handle;
   caddr_t logical, base;
@@ -570,8 +451,6 @@ __map_page(physical_page)
   if (handle == NULL_HANDLE)
     return (FALSE);
 
-#ifdef __freeunices__
-
   E_printf("EMS: map()ing physical page 0x%01x, handle=%d, logical page 0x%x\n", 
            physical_page,handle,emm_map[physical_page].logical_page);
 
@@ -579,19 +458,11 @@ __map_page(physical_page)
   logical = (caddr_t) handle_info[handle].object + emm_map[physical_page].logical_page * EMM_PAGE_SIZE;
 
   _do_map_page(base, logical);
-
-#else /* __freeunices__ */
-  MACH_CALL((vm_allocate(mach_task_self(),
-			 EMM_BASE_ADDRESS + (physical_page * EMM_PAGE_SIZE),
-			   EMM_PAGE_SIZE)), "unmap");
-#endif /* __freeunices__ */
-
   return (TRUE);
 }
 
 static boolean_t
-__unmap_page(physical_page)
-     int physical_page;
+__unmap_page(int physical_page)
 {
   int handle;
   caddr_t logical, base;
@@ -602,38 +473,21 @@ __unmap_page(physical_page)
   if (handle == NULL_HANDLE)
     return (FALSE);
 
-#ifdef __freeunices__
-
   E_printf("EMS: unmap()ing physical page 0x%01x, handle=%d, logical page 0x%x\n", 
            physical_page,handle,emm_map[physical_page].logical_page);
 
   base = (caddr_t) EMM_BASE_ADDRESS + (physical_page * EMM_PAGE_SIZE);
   logical = (caddr_t) handle_info[handle].object + emm_map[physical_page].logical_page * EMM_PAGE_SIZE;
 
-  if (ems_mmap) {
-    /* XXX - add error checking here */
-    munmap(base, EMM_PAGE_SIZE);
-    mmap(base, EMM_PAGE_SIZE, PROT_EXEC | PROT_READ | PROT_WRITE,
-	 MAP_PRIVATE | MAP_FIXED | MAP_ANON, -1, 0);
-  } else {
-
-    E_printf("EMS: COPY-UNMAP l=0x%08x, bs=0x%04x, d=0x%02x, s=0x%02x\n",
-	     (int)logical, (int)base, *(u_short *)logical, * (u_short *) base);
-    memmove((u_char *) logical, (u_char *) base, EMM_PAGE_SIZE);
-  }
-
-#else /* __freeunices__ */
-  MACH_CALL((vm_deallocate(mach_task_self(),
-			 EMM_BASE_ADDRESS + (physical_page * EMM_PAGE_SIZE),
-			   EMM_PAGE_SIZE)), "unmap");
-#endif /* __freeunices__ */
-
+  munmap_mapping(MAPPING_EMS, base, EMM_PAGE_SIZE);
+  mapscratch_mapping(MAPPING_EMS, base, EMM_PAGE_SIZE,
+	PROT_EXEC | PROT_READ | PROT_WRITE);
+  	
   return (TRUE);
 }
 
 static inline boolean_t
-unmap_page(physical_page)
-     int physical_page;
+unmap_page(int physical_page)
 {
    E_printf("EMS: unmap_page(%d)\n",physical_page);
 
@@ -647,18 +501,14 @@ unmap_page(physical_page)
 }
 
 static inline boolean_t
-reunmap_page(physical_page)
-     int physical_page;
+reunmap_page(int physical_page)
 {
    E_printf("EMS: reunmap_page(%d)\n",physical_page);
    return __unmap_page(physical_page);
 }
 
 static boolean_t
-map_page(handle, physical_page, logical_page)
-     int handle;
-     int physical_page;
-     int logical_page;
+map_page(int handle, int physical_page, int logical_page)
 { 
   caddr_t base, logical;
 
@@ -673,17 +523,6 @@ map_page(handle, physical_page, logical_page)
   if (handle_info[handle].numpages <= logical_page)
     return (FALSE);
 
-#if 0
-/* don't use this for the sake of remap_page() */
-
-	if ((emm_map[physical_page].handle == handle) &&
-	    (emm_map[physical_page].logical_page == logical_page)) {
-		return(TRUE);
-	}
-#endif
-
-#ifdef __freeunices__
-
 #ifdef SYNC_ALOT
   sync();
 #endif
@@ -696,40 +535,28 @@ map_page(handle, physical_page, logical_page)
    
   _do_map_page(base, logical);
 
-#else /* __freeunices__ */
-  memory_object_remap(
-		       handle_info[handle].object,
-		       logical_page * EMM_PAGE_SIZE,
-		       EMM_BASE_ADDRESS + (physical_page * EMM_PAGE_SIZE),
-		       EMM_PAGE_SIZE
-    );
-#endif /* __freeunices__ */  
-
   emm_map[physical_page].handle = handle;
   emm_map[physical_page].logical_page = logical_page;
   return (TRUE);
 }
 
 static inline boolean_t
-remap_page(physical_page)
-     int physical_page;
+remap_page(int physical_page)
 {
-  E_printf("EMS: remaping physical page 0x%01x\n", physical_page);
+  E_printf("EMS: remapping physical page 0x%01x\n", physical_page);
 
   return __map_page(physical_page);
 }
 
 
 static inline int
-handle_pages(handle)
-     int handle;
+handle_pages(int handle)
 {
   return (handle_info[handle].numpages);
 }
 
 static int
-save_handle_state(handle)
-     int handle;
+save_handle_state(int handle)
 {
   int i;
 
@@ -739,8 +566,6 @@ save_handle_state(handle)
 	emm_map[i].logical_page;
       handle_info[handle].saved_mappings_handle[i] =
 	emm_map[i].handle;
-      if (!ems_mmap)
-	reunmap_page(i);
     }
     else
       handle_info[handle].saved_mappings_logical[i] =
@@ -750,8 +575,7 @@ save_handle_state(handle)
 }
 
 static int
-restore_handle_state(handle)
-     int handle;
+restore_handle_state(int handle)
 {
   int i;
 
@@ -773,25 +597,6 @@ restore_handle_state(handle)
   return 0;
 }
 
-static void
-test_handle(handle, numpages)
-     int handle, numpages;
-{
-  int i;
-
-  for (i = 0; i < numpages; i++) {
-    map_page(handle, 0, i);
-    *(int *) EMM_BASE_ADDRESS = i;
-  }
-  for (i = 0; i < numpages; i++) {
-    map_page(handle, 0, i);
-    if ((*(int *) EMM_BASE_ADDRESS) != i) {
-      fprintf(stderr, "bios_emm: test_handle, Mapping failure: %d, %d\n",
-	      i, (*(int *) EMM_BASE_ADDRESS));
-    }
-  }
-}
-
 static int
 do_map_unmap(state_t * state, int handle, int physical_page, int logical_page)
 {
@@ -837,25 +642,11 @@ SEG_TO_PHYS(int segaddr)
 {
   E_printf("SEG_TO_PHYS: segment: %x\n", segaddr);
 
-#if 1
   segaddr-=EMM_SEGMENT;
   if ((segaddr<0) || (segaddr >= EMM_MAX_PHYS*EMM_PAGE_SIZE/16) || ((segaddr & 0x3ff) != 0))
      return -1;
   else
      return (segaddr)/(EMM_PAGE_SIZE/16);
-
-#else
-  if (segaddr == EMM_SEGMENT)
-    return 0;
-  else if (segaddr == (EMM_SEGMENT + 0x400))
-    return 1;
-  else if (segaddr == (EMM_SEGMENT + 0x800))
-    return 2;
-  else if (segaddr == (EMM_SEGMENT + 0xc00))
-    return 3;
-  else
-    return (-1);
-#endif
 }
 
 /* EMS 4.0 functions start here */
@@ -978,16 +769,54 @@ reallocate_pages(state_t * state)
      if (emm_map[i].handle == handle)
         reunmap_page(i);
 
-  obj = realloc(handle_info[handle].object, newcount * EMM_PAGE_SIZE);
+  Kdebug0((dbg_fd, "want reallocate_pages handle %d num %d called object=%p\n",
+	   handle, newcount, handle_info[handle].object));
 
-  if (obj==NULL) {
+  if (newcount && handle_info[handle].object) {
+    /*
+     * No special handling required, we safely can realloc() or mremap()
+     */
+    obj = realloc_memory_object(handle_info[handle].object, handle_info[handle].numpages*EMM_PAGE_SIZE ,newcount * EMM_PAGE_SIZE);
+    if (obj==NULL) {
      Kdebug0((dbg_fd, "failed to realloc pages!\n"));
      SETHIGH(&(state->eax), EMM_OUT_OF_LOG);
-  }
-  else {
+    }
+    else {
      handle_info[handle].object = obj;
      handle_info[handle].numpages = newcount;
      SETHIGH(&(state->eax), EMM_NO_ERR);
+    }
+  }
+  else {
+    obj = 0;
+    if (handle_info[handle].object) {
+      /*
+       * Here we come, when reallocation would lead to a ZERO sized block.
+       * This would be bad for mremap(), because this is equal to munmap()
+       * We destroy the objekt instead
+       */
+      destroy_memory_object(handle_info[handle].object,handle_info[handle].numpages*EMM_PAGE_SIZE);
+      handle_info[handle].object = 0;
+      handle_info[handle].numpages = 0;
+      SETHIGH(&(state->eax), EMM_NO_ERR);
+    }
+    else {
+      /*
+       * Here we come, when a reallocation for an above destroyed object
+       * is requested.
+       * We simply create a new one.
+       */
+      if (newcount) obj= new_memory_object(newcount*EMM_PAGE_SIZE);
+      if (obj==NULL) {
+        Kdebug0((dbg_fd, "failed to realloc pages!\n"));
+        SETHIGH(&(state->eax), EMM_OUT_OF_LOG);
+      }
+      else {
+        handle_info[handle].object = obj;
+        handle_info[handle].numpages = newcount;
+        SETHIGH(&(state->eax), EMM_NO_ERR);
+      }
+    }
   }
   
   Kdebug0((dbg_fd, "reallocate_pages handle %d num %d called object=%p\n",
@@ -997,6 +826,12 @@ reallocate_pages(state_t * state)
 
   for (i = 0; i < EMM_MAX_PHYS; i++) {
     if (emm_map[i].handle == handle) {
+       /*
+        * NOTE: In case of the above critical case (newcount==0)
+        *       We _always_ have emm_map[i].logical_page >= newcount
+        *       because even NULL_PAGE is > 0
+        *       Hence we need no special treatment here
+        */
        if (emm_map[i].logical_page >= newcount) {
           emm_map[i].handle = NULL_HANDLE;
           emm_map[i].logical_page = NULL_PAGE;
@@ -1117,9 +952,10 @@ handle_dir(state_t * state)
 
   case SEARCH_NAMED:{
 	int handle;
+	char xxx[9]; 
 	u_char *array = (u_char *) Addr(state, ds, esi);
-
-	Kdebug0((dbg_fd, "SEARCH_NAMED function called\n"));
+	strncpy(xxx,array,8); xxx[8]=0;
+	Kdebug0((dbg_fd, "SEARCH_NAMED '%s' function called\n",xxx));
 
 	for (handle = 0; handle < MAX_HANDLES; handle++) {
 	  if (!HANDLE_ALLOCATED(handle))
@@ -1139,7 +975,7 @@ handle_dir(state_t * state)
   case GET_TOTAL:{
 	Kdebug0((dbg_fd, "GET_TOTAL %d\n", MAX_HANDLES));
 	SETWORD(&(state->ebx), MAX_HANDLES);
-	SETHIGH(&(state->ebx), EMM_NO_ERR);
+	SETHIGH(&(state->eax), EMM_NO_ERR);
 	break;
       }
 
@@ -1201,19 +1037,8 @@ static inline void
 load_move_mem(u_char * mem, struct mem_move_struct *mem_move)
 {
 
-#if 0
-  int seg;
-  int off;
-
-  seg = *(u_short *) mem;
-  mem += 2;
-  off = *(u_short *) mem;
-  mem += 2;
-  mem_move->size = seg * 16 + off;
-#else
   mem_move->size = *(int *) mem;
   mem += 4;
-#endif
 
   mem_move->source_type = *mem++;
   mem_move->source_handle = *(u_short *) mem;
@@ -1236,12 +1061,6 @@ move_memory_region(state_t * state)
 {
   struct mem_move_struct *mem_move=NULL;
   caddr_t dest, source, mem;
-  u_char i;
-
-  if (!ems_mmap) {
-    for (i = 0; i < EMM_MAX_PHYS; i++)
-      reunmap_page(i);
-  }
 
   (u_char *) mem = (u_char *) Addr(state, ds, esi);
   load_move_mem(mem, mem_move);
@@ -1294,12 +1113,6 @@ move_memory_region(state_t * state)
   E_printf("EMS: Move Memory Region from 0x%x -> 0x%x\n", (int)source, (int)dest);
   memmove((u_char *) dest, (u_char *) source, mem_move->size);
 
-  if (!ems_mmap) {
-    for (i = 0; i < EMM_MAX_PHYS; i++)
-      remap_page(i);
-    E_printf("EMS: remap_page\n");
-  }
-
   if (source < dest) {
     if (source + mem_move->size >= dest) {
       E_printf("EMS: Source overlaps Dest\n");
@@ -1320,12 +1133,6 @@ exchange_memory_region(state_t * state)
 {
   struct mem_move_struct *mem_move=NULL;
   caddr_t dest, source, mem, tmp;
-  u_char i;
-
-  if (!ems_mmap) {
-    for (i = 0; i < EMM_MAX_PHYS; i++)
-      reunmap_page(i);
-  }
 
   (u_char *) mem = (u_char *) Addr(state, ds, esi);
   load_move_mem(mem, mem_move);
@@ -1367,11 +1174,6 @@ exchange_memory_region(state_t * state)
   memmove(dest, tmp, mem_move->size);
   free(tmp);
 
-  if (!ems_mmap) {
-    for (i = 0; i < EMM_MAX_PHYS; i++)
-      remap_page(i);
-  }
-
   return (0);
 }
 
@@ -1420,11 +1222,7 @@ get_ems_hardinfo(state_t * state)
       case GET_ARRAY:{
         u_short *ptr = (u_short *) Addr(state, es, edi);
 
-#if 0
-        *ptr = (4 * 1024) / 16;
-#else
         *ptr = (16 * 1024) / 16;
-#endif
         ptr++;			/* raw page size in paragraphs, 4K */
         *ptr = 0;
         ptr++;			/* # of alternate register sets */
@@ -2058,11 +1856,9 @@ ems_fn(state)
     }
     break;
 
-#if 1
   case ALTERNATE_MAP_REGISTER: /* 0x5B */
       alternate_map_register(state); 
     break;     
-#endif
 
   case OS_SET_FUNCTION: /* 0x5D */
       os_set_function(state); 
