@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -9,25 +9,16 @@
 /* Routines to modify INI files */
 /* by David V. Mortenson */
 
+#include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
-#if 0
-#define __USE_GNU
-#define __USE_SVID
-#endif
 #include <string.h>
 #include <ctype.h>
 
 #define FALSE  0
 #define TRUE   1
 
-#if 0
-extern void GetValueFromIniFile (char *, char *, char *, char *);
-extern void GetValueFromIniFilePath (char *, char *, char *, char *);
-extern void SetValueInIniFile (char *, char *, char *, char *);
-#else
 #include "inifile.h"
-#endif
 
 /****************************************
  * An ini file should have the extension .INI, and is formatted
