@@ -1,6 +1,6 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -15,7 +15,6 @@
  *	(c) 1994 Alan Cox	iiitac@pyr.swan.ac.uk	GW4PTS@GB7SWN
  */
 #include <stdio.h>
-#include <features.h>
 #include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
@@ -26,16 +25,14 @@
 #include <sys/stat.h>
 #include <net/if.h>
 #include <netinet/in.h>
-#if GLIBC_VERSION_CODE >= 2000
-  #include <netinet/if_ether.h>
-#else
-  #include "Linux/if_ether.h"
-#endif
+#include "Linux/if_tun.h"
+#include <netinet/if_ether.h>
 
 #include "emu.h"
 #include "priv.h"
 #include "libpacket.h"
 #include "dosnet.h"
+#include "pktdrvr.h"
 
 static int GetDosnetID(void);
 
@@ -80,7 +77,8 @@ OpenNetworkType(unsigned short netid)
 	else
 		proto = htons(netid);
 
-	if (!config.secure) enter_priv_on();
+	enter_priv_on();
+
 #ifdef AF_PACKET
 	if (running_kversion >= 2001000)
 		s = socket(AF_PACKET, SOCK_PACKET, proto);
@@ -89,7 +87,7 @@ OpenNetworkType(unsigned short netid)
 #else
 	s = socket(AF_INET, SOCK_PACKET, proto);
 #endif
-	if (!config.secure) leave_priv_setting();
+	leave_priv_setting();
 	if (s < 0) {
 		if (errno == EPERM) warn("Must be root for virtual TCP/IP\n");
 		return -1;
@@ -100,11 +98,11 @@ OpenNetworkType(unsigned short netid)
 
 /* Return a socket opened in broadcast mode (only used for virtual net) */
 int
-OpenBroadcastNetworkType()
+OpenBroadcastNetworkType(void)
 {
 	PRIV_SAVE_AREA
 	int s;
-	if (!config.secure) enter_priv_on();
+	enter_priv_on();
 #ifdef AF_PACKET
 	if (running_kversion >= 2001000)
 		s = socket(AF_PACKET, SOCK_PACKET, 
@@ -114,7 +112,7 @@ OpenBroadcastNetworkType()
 #else
 	s = socket(AF_INET, SOCK_PACKET, htons(DOSNET_BROADCAST_TYPE));  
 #endif
-	if (!config.secure) leave_priv_setting();
+	leave_priv_setting();
 	if (s < 0) {
 		pd_printf("OpenBroadcast: could not open socket\n");
 		if (errno == EPERM) warn("Must be root for virtual TCP/IP\n");
@@ -151,6 +149,10 @@ WriteToNetwork(int sock, const char *device, const char *data, int len)
 {
 	struct sockaddr sa;
 
+	if (config.vnet == VNET_TYPE_TAP) {
+	  return write(sock, data, len);
+	}
+
 #ifdef AF_PACKET
 	if (running_kversion >= 2001000)
 		sa.sa_family = AF_PACKET;
@@ -186,6 +188,10 @@ ReadFromNetwork(int sock, char *device, char *data, int len)
 	int sz = sizeof(sa);
 	int error;
 
+	if (config.vnet == VNET_TYPE_TAP) {
+		return read(sock, data, len);
+	}
+
 	error = recvfrom(sock, data, len, 0, &sa, &sz);
 
 	if (error == -1)
@@ -274,3 +280,34 @@ GetDeviceMTU(char *device)
 		return -1;
 	return req.ifr_mtu;
 }
+
+
+int tun_alloc(char *dev)
+{
+      struct ifreq ifr;
+      int fd, err;
+
+      if( (fd = open("/dev/net/tun", O_RDWR)) < 0 )
+         return -1;
+
+      memset(&ifr, 0, sizeof(ifr));
+
+      /* Flags: IFF_TUN   - TUN device (no Ethernet headers)
+       *        IFF_TAP   - TAP device
+       *
+       *        IFF_NO_PI - Do not provide packet information
+       */
+      ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
+      if( *dev )
+         strncpy(ifr.ifr_name, dev, IFNAMSIZ);
+
+      if( (err = ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0 ){
+         close(fd);
+         return err;
+      }
+      strcpy(dev, ifr.ifr_name);
+
+      DosnetID = GetDosnetID();
+
+      return fd;
+}
