@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -40,8 +40,16 @@
 #include "dpmi.h"
 #include "priv.h"
 
-extern void xms_control(void);
-
+#ifdef X86_EMULATOR
+#include "simx86/syncpu.h"
+#include "cpu-emu.h"
+#define CRs	TheCPU.cr
+#define DRs	TheCPU.dr
+#define TRs	TheCPU.tr
+#else
+/* Extra CPU registers. Note that GDTR,LDTR,IDTR are internal
+ * to the cpuemu and are defined in emu-ldt.c:
+ */
 unsigned long CRs[5] =
 {
 	0x00000013,	/* valid bits: 0xe005003f */
@@ -86,6 +94,13 @@ unsigned long TRs[2] =
 	0x00000000,
 	0x00000000
 };
+#endif
+
+/* DOSEMU code selector and data selector. Initialized at startup.
+ * These used to be 0x23 and 0x2b, but change to 0x73 and 0x7b
+ * in Linux kernel 2.6
+ */
+unsigned short ucodesel, udatasel;
 
 /* 
  * DANG_BEGIN_FUNCTION cpu_trap_0f
@@ -113,6 +128,12 @@ int cpu_trap_0f (unsigned char *csp, struct sigcontext_struct *scp)
 	else if (csp[1] == 0x31) {
 		/* ref: Van Gilluwe, "The Undocumented PC". The program
 		 * 'cpurdtsc.exe' traps here */
+#ifdef X86_EMULATOR
+		if (config.cpuemu>1) {
+		  REG(eax) = (unsigned long)TheCPU.EMUtime;
+		  REG(edx) = (unsigned long)(TheCPU.EMUtime>>32);
+		} else
+#endif
 		if (vm86s.cpu_type >= CPU_586) {
 		  __asm__ __volatile__ ("rdtsc" \
 			:"=a" (REG(eax)),  \
@@ -127,8 +148,8 @@ int cpu_trap_0f (unsigned char *csp, struct sigcontext_struct *scp)
 		  gettimeofday(&tv, NULL);
 		  t = (unsigned long long)tv.tv_sec*100000000 +
 		      (unsigned long long)tv.tv_usec*100;
-		  REG(eax)=((int *)&t)[0];
-		  REG(edx)=((int *)&t)[1];
+		  REG(eax)=t & 0xffffffff;
+		  REG(edx)=t >> 32;
 		}
 		(in_dpmi? scp->eip:LWORD(eip)) += 2;
 		return 1;
@@ -203,9 +224,6 @@ int cpu_trap_0f (unsigned char *csp, struct sigcontext_struct *scp)
  */
 void cpu_setup(void)
 {
-  extern void int_vector_setup(void);
-  extern void init_cpu (void);
-
   int_vector_setup();
 
   /* ax,bx,cx,dx,si,di,bp,fs,gs can probably can be anything */
@@ -230,18 +248,21 @@ void cpu_setup(void)
   REG(eflags) |= (VIF | VIP);
 #endif
 
-}
+  /* initialize user data & code selector values (used by DPMI code) */
+  __asm__ volatile (
+  " movw %%cs, %0\n"
+  " movw %%ds, %1\n"
+  :"=m"(ucodesel),"=m"(udatasel):);
 
-
-
-#ifdef USE_INT_QUEUE
-int
-do_hard_int(int intno)
-{
-  queue_hard_int(intno, NULL, NULL);
-  return (1);
-}
+#ifdef X86_EMULATOR
+  if (config.cpuemu) {
+    init_emu_cpu();
+  }
 #endif
+}
+
+
+
 
 int
 do_soft_int(int intno)
