@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -16,15 +16,17 @@
  * SIDOC_END_MODULE
  */
 
+#include "config.h"
 #include <stdlib.h>
 #include <stdio.h>
-#include <features.h>
 #include <sys/time.h>
 #include <unistd.h>
-#include "config.h"
 #include "emu.h"
+#include "video.h"
 #include "timers.h"
 #include "pic.h"
+#include "speaker.h"
+#include "dosemu_config.h"
 
 /* --------------------------------------------------------------------- */
 /*
@@ -81,8 +83,10 @@
 static hitimer_t (*RAWcpuTIME)(void);
 
 hitimer_u ZeroTimeBase = { 0 };
+hitimer_u ZeroTSCBase = { 0 };
 static hitimer_t LastTimeRead = 0;
 static hitimer_t StopTimeBase = 0;
+static hitimer_t StopTSCBase = 0;
 int cpu_time_stop = 0;
 
 /*
@@ -126,30 +130,29 @@ static hitimer_t getC4time(void)
 static hitimer_t getP5time(void)
 {
   if (cpu_time_stop) return LastTimeRead;
-  return CPUtoUS_Z();
+  return (rawP5time() - ZeroTimeBase.td);
 }
 
 /*
  * SIDOC_BEGIN_FUNCTION GETusTIME(sc)
  *
- * GETusTIME returns the DOS ('stretched') time with 1-usec resolution
+ * GETusTIME returns the DOS time with 1-usec resolution
  *  using GETcpuTIME to get the implementation-dependent CPU time.
- * The 'sc' parameter controls the granularity of the stretching
- *  algorithm (not yet there, see the docs)
+ * The 'sc' parameter is unused.
  *
  * SIDOC_END_FUNCTION
  */
 hitimer_t GETusTIME(int sc)
 {
-  return GETcpuTIME();	/* no 'stretching' at the moment */
+  return GETcpuTIME();
 }
 
 /*
  * SIDOC_BEGIN_FUNCTION GETtickTIME(sc)
  *
- * GETtickTIME returns the DOS (stretched) time with 838ns resolution
+ * GETtickTIME returns the DOS time with 838ns resolution
  *  using GETcpuTIME to get the implementation-dependent CPU time.
- * The 'sc' parameter works like in GETusTIME.
+ * The 'sc' parameter is unused.
  *
  * SIDOC_END_FUNCTION
  */
@@ -173,21 +176,23 @@ hitimer_t GETusSYSTIME(void)
 
 void get_time_init (void)
 {
-  extern int kernel_version_code;
   if ((config.realcpu > CPU_486) && config.rdtsc) {
     /* we are here if: a 586/686 was detected at startup, we are not
      * on a SMP machine and the user didn't say 'rdtsc off'. But
      * we could have an emulated CPU < 586, this doesn't affect timing
      * but only flags processing (& other features) */
-    RAWcpuTIME = rawP5time;
-    GETcpuTIME = getP5time;
+    RAWcpuTIME = rawP5time;		/* in usecs */
+    GETcpuTIME = getP5time;		/* in usecs */
+    ZeroTSCBase.td = GETTSC();
+    ZeroTimeBase.td = TSCtoUS(ZeroTSCBase.td);
     g_printf("TIMER: using pentium timing\n");
   }
   else {
     /* we are here on all other cases: real CPU < 586, SMP machines,
      * 'rdtsc off' into config file */
-    RAWcpuTIME = rawC4time;
-    GETcpuTIME = getC4time;
+    RAWcpuTIME = rawC4time;		/* in usecs */
+    GETcpuTIME = getC4time;		/* in usecs */
+    ZeroTimeBase.td = rawC4time();
     if (config.realcpu) {
       if (kernel_version_code < 0x2017e)
         g_printf("TIMER: using gettimeofday\n");
@@ -195,7 +200,6 @@ void get_time_init (void)
         g_printf("TIMER: using new gettimeofday with microsecond resolution\n");
     }
   }
-  ZeroTimeBase.td = RAWcpuTIME();
 }
 
 
@@ -208,6 +212,9 @@ int stop_cputime (int quiet)
   if (!quiet) dbug_printf("STOP TIME\n");
   StopTimeBase = RAWcpuTIME();
   LastTimeRead = StopTimeBase - ZeroTimeBase.td;
+  if (ZeroTSCBase.td) {
+    StopTSCBase = GETTSC();
+  }
   cpu_time_stop = 1;
   return 0;
 }
@@ -217,26 +224,52 @@ int restart_cputime (int quiet)
 {
   if (!cpu_time_stop) return 1;
   if (!quiet) dbug_printf("RESTART TIME\n");
-  ZeroTimeBase.td += (RAWcpuTIME() - StopTimeBase);
+  ZeroTimeBase.td += RAWcpuTIME() - StopTimeBase;
+  if (ZeroTSCBase.td) {
+    ZeroTSCBase.td += GETTSC() - StopTSCBase;
+  }
   cpu_time_stop = 0;
   return 0;
 }
 
 /* --------------------------------------------------------------------- */
 int dosemu_frozen = 0;
+int dosemu_user_froze = 0;
+
+void freeze_dosemu_manual(void)
+{
+  dosemu_user_froze = 2;
+  freeze_dosemu();
+}
 
 void freeze_dosemu(void)
 {
   if (dosemu_frozen) return;
+  
   stop_cputime(0);
   dosemu_frozen = 1;
+  if (dosemu_user_froze) dosemu_user_froze--;
+  dbug_printf("*** dosemu frozen\n");
+  
+  speaker_pause();
+  
+  if (Video->change_config)
+    Video->change_config (CHG_TITLE, NULL);
 }
 
 void unfreeze_dosemu(void)
 {
   if (!dosemu_frozen) return;
+  
+  speaker_resume ();
+  
   restart_cputime(0);
   dosemu_frozen = 0;
+  dosemu_user_froze = 0;
+  dbug_printf("*** dosemu unfrozen\n");
+
+  if (Video->change_config)
+    Video->change_config (CHG_TITLE, NULL);
 }
 
 
@@ -246,18 +279,19 @@ static int getmhz(void)
 {
 	struct timeval tv1,tv2;
 	hitimer_t a,b;
+        unsigned long a0, a1, b0, b1;
 
 	gettimeofday(&tv1, NULL);
 	__asm__ __volatile__ ("rdtsc"
-		:"=a" (((unsigned long*)&a)[0]),
-		 "=d" (((unsigned long*)&a)[1]));
+		:"=a" (a0),
+		 "=d" (a1));
 /*	for (j=0; j<10000000; j++);*/	/* 500ms on a P5-100 */
 	usleep(50000);
 	gettimeofday(&tv2, NULL);
 	__asm__ __volatile__ ("rdtsc"
-		:"=a" (((unsigned long*)&b)[0]),
-		 "=d" (((unsigned long*)&b)[1]));
-	b -= a;
+		:"=a" (b0),
+		 "=d" (b1));
+	b = (((hitimer_t)b1 << 32) | b0) - (((hitimer_t)a1 << 32) | a0);
 	a = (tv2.tv_sec*1000000 + tv2.tv_usec) - 
 	    (tv1.tv_sec*1000000 + tv1.tv_usec);
 	return (int)((b*4096)/a);
@@ -290,6 +324,7 @@ int bogospeed(unsigned long *spus, unsigned long *sptick)
 	/* speed division factor to get 838ns from CPU clocks */
 	*sptick = (LLF_TICKS*mlt)/dvs;
 
+	config.CPUSpeedInMhz = dvs/mlt;
 	fprintf (stderr,"CPU speed set to %d MHz\n",(dvs/mlt));
 /*	fprintf (stderr,"CPU speed factors %ld,%ld\n",*spus,*sptick); */
 	first = 0;
