@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -31,13 +31,11 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
-#include <malloc.h>
+#include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
-#if GLIBC_VERSION_CODE >= 2000
-  #include <sys/io.h>
-#endif
+#include <sys/io.h>
 
 #include "emu.h"
 #include "port.h"
@@ -50,45 +48,17 @@
 #include "shared.h"
 #include "serial.h"
 #include "bitops.h"
+#include "mapping.h"
+#ifdef X86_EMULATOR
+#include "cpu-emu.h"
+#include "bitops.h"
+#endif
 
-/*
- * maximum number of emulated devices allowed.  floppy, mda, etc...
- * an 8-bit handle is used for each device:
- */
-#define EMU_MAX_IO_DEVICES 254
-
-#define NO_HANDLE	0x00
-#define HANDLE_STD_IO	0x01
-#define HANDLE_STD_RD	0x02
-#define HANDLE_STD_WR	0x03
-#define HANDLE_VID_IO	0x04
-#define HANDLE_SPECIAL	0x05		/* catch-all */
-#define STD_HANDLES	6
-
-static struct _port_handler {
-	Bit8u  (*read_portb)  (ioport_t port_addr);
-	void   (*write_portb) (ioport_t port_addr, Bit8u byte);
-	Bit16u (*read_portw) (ioport_t port_addr);
-	void   (*write_portw) (ioport_t port_addr, Bit16u word);
-	Bit32u (*read_portd) (ioport_t port_addr);
-	void   (*write_portd) (ioport_t port_addr, Bit32u dword);
-	char   *handler_name;
-	int    irq, fd;
-} port_handler[EMU_MAX_IO_DEVICES];
-
-/*
- * If you are worried by the size of this array keep in mind that:
- * - it always existed, it was just not used
- * - ports above 0x400 could be better implemented using some form of
- *   dynamic storage
- *	00    = no device attached, port undefined
- *	01-FD = index into handle
- *	FE,FF = reserved
- * Ioperm could use a bitmap similar to set/reset_revectored
- */
-static unsigned char port_handle_table[0x10000];
-static unsigned char port_andmask[0x400];
-static unsigned char port_ormask[0x400];
+_port_handler port_handler[EMU_MAX_IO_DEVICES];
+unsigned char port_handle_table[0x10000];
+unsigned char port_andmask[0x400];
+unsigned char port_ormask[0x400];
+pid_t portserver_pid = 0;
 
 static unsigned char port_handles;	/* number of io_handler's */
 
@@ -96,16 +66,12 @@ static const char *irq_handler_name[EMU_MAX_IRQS];
 
 #define SET_HANDLE(p,h)		port_handle_table[(Bit16u)(p)]=(h)
 #define EMU_HANDLER(port)	port_handler[port_handle_table[(Bit16u)(port)]]
-
-/* This is the same code that was used in cpu.c, priv_iopl is really
-   very slow */
-#define PORT_IOPLON(p,n)	if ((p)<0x400) set_ioperm((p),(n),1); else priv_iopl(3)
-#define PORT_IOPLOFF(p,n)	if ((p)<0x400) set_ioperm((p),(n),0); else priv_iopl(0)
+enum{TYPE_INB, TYPE_OUTB, TYPE_INW, TYPE_OUTW, TYPE_IND, TYPE_OUTD, TYPE_EXIT};
 
 /* any user or system device which creates a new handle can't be later
  * remapped by extra_port_init()
  */
-void SET_HANDLE_COND(int p, int h)
+static void SET_HANDLE_COND(int p, int h)
 {
   if (port_handle_table[(p)] < STD_HANDLES)
 	port_handle_table[(p)]=(h);
@@ -143,8 +109,6 @@ static char *
 #define SIZE_PORTLOGMAP		(1 << (PORTLOG_MAXBITS -3))
 static unsigned long *portlog_map = 0;
 
-void init_port_traceing(void);
-
 void register_port_traceing(ioport_t firstport, ioport_t lastport)
 {
   firstport &= PORTLOG_MASK;
@@ -166,12 +130,12 @@ void init_port_traceing(void)
 {
   if (portlog_map) return;
   clear_port_traceing();
-  register_port_traceing(0x100, 0x2ff);
+  register_port_traceing(0x0, 0xffff);
 }
 
 #define TT_printf(p,f,v,m) ({ \
-  if ( d.io_trace && test_bit(p & PORTLOG_MASK, portlog_map)) { \
-    log_printf(1, "%x %c %x\n", p, f, v & m); \
+  if (debug_level('T') && test_bit(p & PORTLOG_MASK, portlog_map)) { \
+    log_printf(1, "%hx %c %x\n", p, f, v & m); \
   } \
 })
 
@@ -208,13 +172,13 @@ static void log_port_write_d(ioport_t port, Bit32u w)
   TT_printf(port, '[', w, 0xffffffff);
 }
 
-#define LOG_PORT_READ(port, r) (d.io_trace ? log_port_read(port, r) : r)
-#define LOG_PORT_READ_W(port, r) (d.io_trace ? log_port_read_w(port, r) : r)
-#define LOG_PORT_READ_D(port, r) (d.io_trace ? log_port_read_d(port, r) : r)
+#define LOG_PORT_READ(port, r) (debug_level('T') ? log_port_read(port, r) : r)
+#define LOG_PORT_READ_W(port, r) (debug_level('T') ? log_port_read_w(port, r) : r)
+#define LOG_PORT_READ_D(port, r) (debug_level('T') ? log_port_read_d(port, r) : r)
 
-#define LOG_PORT_WRITE(port, w) do{ if (d.io_trace) log_port_write(port, w); }while(0)
-#define LOG_PORT_WRITE_W(port, w) do{ if (d.io_trace) log_port_write_w(port, w); }while(0)
-#define LOG_PORT_WRITE_D(port, w) do{ if (d.io_trace) log_port_write_d(port, w); }while(0)
+#define LOG_PORT_WRITE(port, w) do{ if (debug_level('T')) log_port_write(port, w); }while(0)
+#define LOG_PORT_WRITE_W(port, w) do{ if (debug_level('T')) log_port_write_w(port, w); }while(0)
+#define LOG_PORT_WRITE_D(port, w) do{ if (debug_level('T')) log_port_write_d(port, w); }while(0)
 
 /* ---------------------------------------------------------------------- */
 /* SIDOC_BEGIN_REMARK
@@ -312,7 +276,7 @@ Bit32u port_ind(ioport_t port)
 		res = EMU_HANDLER(port).read_portd(port);
 	}
 	else {
-		res = std_port_ind(port);
+		res = (Bit32u) port_inw(port) | (((Bit32u) port_inw(port + 2)) << 16);
 	}
 	return LOG_PORT_READ_D(port, res);
 }
@@ -324,66 +288,12 @@ void port_outd(ioport_t port, Bit32u dword)
 		EMU_HANDLER(port).write_portd(port, dword);
 	}
 	else {
-		std_port_outd(port, dword);
+		port_outw(port, dword & 0xffff);
+		port_outw(port+2, (dword >> 16) & 0xffff);
 	}
 }
 
 
-/* ---------------------------------------------------------------------- */
-/* default port I/O access
- */
-
-Bit8u std_port_inb(ioport_t port)
-{
-	Bit8u res;
-
-	PORT_IOPLON(port,1);
-	res = port_real_inb(port);
-	PORT_IOPLOFF(port,1);
-	return res;
-}
-
-void std_port_outb(ioport_t port, Bit8u byte)
-{
-	PORT_IOPLON(port,1);
-	port_real_outb(port, byte);
-	PORT_IOPLOFF(port,1);
-}
-
-Bit16u std_port_inw(ioport_t port)
-{
-	Bit16u res;
-
-	PORT_IOPLON(port,2);
-	res = port_real_inw(port);
-	PORT_IOPLOFF(port,2);
-	return res;
-}
-
-void std_port_outw(ioport_t port, Bit16u word)
-{
-	PORT_IOPLON(port,2);
-	port_real_outw(port, word);
-	PORT_IOPLOFF(port,2);
-}
-
-Bit32u std_port_ind(ioport_t port)
-{
-	Bit32u res;
-	PORT_IOPLON(port,4);
-	res = port_real_ind(port);
-	PORT_IOPLOFF(port,4);
-	return res;
-}
-
-void std_port_outd(ioport_t port, Bit32u dword)
-{
-	PORT_IOPLON(port,4);
-	port_real_outd(port, dword);
-	PORT_IOPLOFF(port,4);
-}
-
-
 /* ---------------------------------------------------------------------- */
 /* the following functions are all static!				  */
 
@@ -397,38 +307,169 @@ static Bit8u port_not_avail_inb(ioport_t port)
 /* it is a fact of (hardware) life that unused locations return all
    (or almost all) the bits at 1; some software can try to detect a
    card basing on this fact and fail if it reads 0x00 - AV */
-	if (d.io) pna_emsg(port,'b',"read");
+	if (debug_level('i')) pna_emsg(port,'b',"read");
 	return 0xff;
 }
 
 static void port_not_avail_outb(ioport_t port, Bit8u byte)
 {
-	if (d.io) pna_emsg(port,'b',"write");
+	if (debug_level('i')) pna_emsg(port,'b',"write");
 }
 
 static Bit16u port_not_avail_inw(ioport_t port)
 {
-	if (d.io) pna_emsg(port,'w',"read");
+	if (debug_level('i')) pna_emsg(port,'w',"read");
 	return 0xffff;
 }
 
 static void port_not_avail_outw(ioport_t port, Bit16u value)
 {
-	if (d.io) pna_emsg(port,'w',"write");
+	if (debug_level('i')) pna_emsg(port,'w',"write");
 }
 
 static Bit32u port_not_avail_ind(ioport_t port)
 {
-	if (d.io) pna_emsg(port,'d',"read");
+	if (debug_level('i')) pna_emsg(port,'d',"read");
 	return 0xffffffff;
 }
 
 static void port_not_avail_outd(ioport_t port, Bit32u value)
 {
-	if (d.io) pna_emsg(port,'d',"write");
+	if (debug_level('i')) pna_emsg(port,'d',"write");
 }
 
 
+/* ---------------------------------------------------------------------- */
+/* default port I/O access
+ */
+
+struct portreq 
+{
+        ioport_t port;
+        int type;
+        unsigned long word;
+};
+
+static int port_fd_out[2] = {-1, -1};
+static int port_fd_in[2] = {-1, -1};
+
+Bit8u std_port_inb(ioport_t port)
+{
+        struct portreq pr;
+
+        if (current_iopl == 3) {
+		return port_real_inb(port);
+	}
+	if (!portserver_pid) {
+		error ("std_port_inb(0x%X): port server unavailable\n", port);
+		return port_not_avail_inb (port);
+	}
+	pr.port = port;
+	pr.type = TYPE_INB;
+	write(port_fd_out[1], &pr, sizeof(pr));
+	read(port_fd_in[0], &pr, sizeof(pr));
+	return pr.word;
+}
+
+void std_port_outb(ioport_t port, Bit8u byte)
+{
+        struct portreq pr;
+
+        if (current_iopl == 3) {
+		port_real_outb(port, byte);
+		return;
+        }
+	if (!portserver_pid) {
+		error ("std_port_outb(0x%X,0x%X): port server unavailable\n",
+		       port, byte);
+		port_not_avail_outb (port, byte);
+		return;
+	}
+        pr.word = byte;
+        pr.port = port;
+        pr.type = TYPE_OUTB;
+	write(port_fd_out[1], &pr, sizeof(pr));
+	read(port_fd_in[0], &pr, sizeof(pr));
+}
+
+Bit16u std_port_inw(ioport_t port)
+{
+        struct portreq pr;
+
+        if (current_iopl == 3) {
+		return port_real_inw(port);
+        }
+	if (!portserver_pid) {
+		error ("std_port_inw(0x%X): port server unavailable\n", port);
+		return port_not_avail_inw (port);
+	}
+        pr.port = port;
+        pr.type = TYPE_INW;
+	write(port_fd_out[1], &pr, sizeof(pr));
+	read(port_fd_in[0], &pr, sizeof(pr));
+	return pr.word;
+}
+
+void std_port_outw(ioport_t port, Bit16u word)
+{
+        struct portreq pr;
+
+        if (current_iopl == 3) {
+		port_real_outw(port, word);
+		return;
+        }
+	if (!portserver_pid) {
+		error ("std_port_outw(0x%X,0x%X): port server unavailable\n",
+		       port, word);
+		port_not_avail_outw (port, word);
+		return;
+	}
+        pr.word = word;
+        pr.port = port;
+        pr.type = TYPE_OUTW;
+	write(port_fd_out[1], &pr, sizeof(pr));
+	read(port_fd_in[0], &pr, sizeof(pr));
+}
+
+Bit32u std_port_ind(ioport_t port)
+{
+        struct portreq pr;
+
+        if (current_iopl == 3) {
+		return port_real_ind(port);
+        }
+	if (!portserver_pid) {
+		error ("std_port_ind(0x%X): port server unavailable\n", port);
+		return port_not_avail_ind (port);
+	}
+        pr.port = port;
+        pr.type = TYPE_IND;
+	write(port_fd_out[1], &pr, sizeof(pr));
+	read(port_fd_in[0], &pr, sizeof(pr));
+	return pr.word;
+}
+
+void std_port_outd(ioport_t port, Bit32u dword)
+{
+        struct portreq pr;
+
+        if (current_iopl == 3) {
+		port_real_outd(port, dword);
+		return;
+        }
+	if (!portserver_pid) {
+		error ("std_port_outd(0x%X,0x%X): port server unavailable\n",
+		       port, dword);
+		port_not_avail_outd (port, dword);
+		return;
+	}
+        pr.word = dword;
+        pr.port = port;
+        pr.type = TYPE_OUTD;
+	write(port_fd_out[1], &pr, sizeof(pr));
+	read(port_fd_in[0], &pr, sizeof(pr));
+}
+
 /* ---------------------------------------------------------------------- */
 /* SIDOC_BEGIN_REMARK
  *
@@ -450,12 +491,10 @@ int port_rep_inb(ioport_t port, Bit8u *base, int df, Bit32u count)
 	i_printf("Doing REP insb(%#x) %d bytes at %p, DF %d\n", port,
 		count, base, df);
 	if (EMU_HANDLER(port).read_portb == std_port_inb) {
-	    PORT_IOPLON(port,1);
 	    while (count--) {
-	      *dest = port_real_inb(port);
+	      *dest = std_port_inb(port);
 	      dest += incr;
 	    }
-	    PORT_IOPLOFF(port,1);
 	}
 	else {
 	  while (count--) {
@@ -463,7 +502,7 @@ int port_rep_inb(ioport_t port, Bit8u *base, int df, Bit32u count)
 	    dest += incr;
 	  }
 	}
-	if (d.io_trace) {
+	if (debug_level('T')) {
 		dest = base;
 		while (count_--) {
 			LOG_PORT_READ(port, *dest);
@@ -483,12 +522,10 @@ int port_rep_outb(ioport_t port, Bit8u *base, int df, Bit32u count)
 	i_printf("Doing REP outsb(%#x) %d bytes at %p, DF %d\n", port,
 		count, base, df);
 	if (EMU_HANDLER(port).write_portb == std_port_outb) {
-	    PORT_IOPLON(port,1);
 	    while (count--) {
-	      port_real_outb(port, *dest);
+	      std_port_outb(port, *dest);
 	      dest += incr;
 	    }
-	    PORT_IOPLOFF(port,1);
 	}
 	else {
 	  while (count--) {
@@ -496,7 +533,7 @@ int port_rep_outb(ioport_t port, Bit8u *base, int df, Bit32u count)
 	    dest += incr;
 	  }
 	}
-	if (d.io_trace) {
+	if (debug_level('T')) {
 		dest = base;
 		while (count_--) {
 			LOG_PORT_WRITE(port, *dest);
@@ -516,12 +553,10 @@ int port_rep_inw(ioport_t port, Bit16u *base, int df, Bit32u count)
 	i_printf("Doing REP insw(%#x) %d words at %p, DF %d\n", port,
 		count, base, df);
 	if (EMU_HANDLER(port).read_portw == std_port_inw) {
-	    PORT_IOPLON(port,2);
 	    while (count--) {
-	      *dest = port_real_inw(port);
+	      *dest = std_port_inw(port);
 	      dest += incr;
 	    }
-	    PORT_IOPLOFF(port,2);
 	}
 	else if (EMU_HANDLER(port).read_portw == NULL) {
 	  Bit16u res;
@@ -537,7 +572,7 @@ int port_rep_inw(ioport_t port, Bit16u *base, int df, Bit32u count)
 	    dest += incr;
 	  }
 	}
-	if (d.io_trace) {
+	if (debug_level('T')) {
 		dest = base;
 		while (count_--) {
 			LOG_PORT_READ_W(port, *dest);
@@ -557,12 +592,10 @@ int port_rep_outw(ioport_t port, Bit16u *base, int df, Bit32u count)
 	i_printf("Doing REP outsw(%#x) %d words at %p, DF %d\n", port,
 		count, base, df);
 	if (EMU_HANDLER(port).write_portw == std_port_outw) {
-	    PORT_IOPLON(port,2);
 	    while (count--) {
-	      port_real_outw(port, *dest);
+	      std_port_outw(port, *dest);
 	      dest += incr;
 	    }
-	    PORT_IOPLOFF(port,2);
 	}
 	else if (EMU_HANDLER(port).write_portw == NULL) {
 	  Bit16u res;
@@ -578,7 +611,7 @@ int port_rep_outw(ioport_t port, Bit16u *base, int df, Bit32u count)
 	    dest += incr;
 	  }
 	}
-	if (d.io_trace) {
+	if (debug_level('T')) {
 		dest = base;
 		while (count_--) {
 			LOG_PORT_WRITE_W(port, *dest);
@@ -616,8 +649,6 @@ int port_rep_outd(ioport_t port, Bit32u *base, int df, Bit32u count)
 	return (Bit8u *)dest-(Bit8u *)base;
 }
 
-
-/* ---------------------------------------------------------------------- */
 /* 
  * SIDOC_BEGIN_FUNCTION special_port_inb,special_port_outb
  *
@@ -644,22 +675,9 @@ static Bit8u special_port_inb(ioport_t port)
 {
 	Bit8u res = 0xff;
 
-	if (config.usesX) {
-	/* HGC stuff */
-	    if (port==0x3b8) {	/* HGC mode-reg */
-		res = std_port_inb (port);
-		return ((res & 0x7f) | (hgc_Mode & 0x80));
-	    }
-	    else
-	    if (port==0x3bf) {	/* HGC conf-reg */
-		res = std_port_inb (port);
-		return ((res & 0xfd) | (hgc_Konv & 0x02));
-	    }
-	}
 	if ((port==0x3ba)||(port==0x3da)) {
 		res = Misc_get_input_status_1();
-		if (!config.usesX && !r3da_pending && (config.emuretrace>1)) {
-			set_ioperm(0x3c0,1,0);
+		if (!r3da_pending && (config.emuretrace>1)) {
 			r3da_pending = port;
 		}
 	}
@@ -671,22 +689,6 @@ static Bit8u special_port_inb(ioport_t port)
 
 static void special_port_outb(ioport_t port, Bit8u byte)
 {
-	if (config.usesX) {
-	    if (port==0x3b8) {	/* HGC mode-reg */
-		if (byte & 0x80) set_hgc_page(1);
-			else set_hgc_page(0);
-		byte &= 0x7f;
-		goto defout;
-	    }
-	    else
-	    if (port==0x3bf) {	/* HGC conf-reg */
-		if (byte & 0x02) map_hgc_page(1);
-			else map_hgc_page(0);
-		byte &= 0xfd;
-		goto defout;
-	    }
-	}
-
 	/* Port writes for enable/disable blinking character mode */
 	if (port == 0x03c0) {
 		static int last_byte = -1;
@@ -709,12 +711,11 @@ static void special_port_outb(ioport_t port, Bit8u byte)
 		 *
 		 * SIDOC_END_REMARK
 		 */
-		if (config.vga && !config.usesX && (config.emuretrace>1)) {
+		if (config.vga && (config.emuretrace>1)) {
 		    if (r3da_pending) {
 			(void)std_port_inb(r3da_pending);
 			r3da_pending = 0;
-			set_ioperm(0x3c0,1,1);
-			port_real_outb(0x3c0, byte);
+			std_port_outb(0x3c0, byte);
 			return;
 		    }
 		    goto defout;
@@ -733,7 +734,7 @@ static void special_port_outb(ioport_t port, Bit8u byte)
 	}
 
 	/* Port writes for cursor position: 3b4-3b5 or 3d4-3d5 */
-	if (!config.usesX && ((port & 0xfffe)==READ_WORD(BIOS_VIDEO_PORT))) {
+	if ((port & 0xfffe)==READ_WORD(BIOS_VIDEO_PORT)) {
 		/* Writing to the 6845 */
 		static int last_port;
 		static int last_byte;
@@ -793,6 +794,9 @@ int port_init(void)
 	  port_handler[i].irq = EMU_NO_IRQ;
 	  port_handler[i].fd = -1;
 	}
+#ifdef X86_EMULATOR
+	memset (io_bitmap, 0, sizeof(io_bitmap));
+#endif
 
   /* handle 0 maps to the unmapped IO device handler.  Basically any
      ports which don't map to any other device get mapped to this
@@ -802,6 +806,8 @@ int port_init(void)
 	port_handler[NO_HANDLE].write_portb = port_not_avail_outb;
 	port_handler[NO_HANDLE].read_portw = port_not_avail_inw;
 	port_handler[NO_HANDLE].write_portw = port_not_avail_outw;
+	port_handler[NO_HANDLE].read_portd = port_not_avail_ind;
+	port_handler[NO_HANDLE].write_portd = port_not_avail_outd;
 	port_handler[NO_HANDLE].handler_name = "unknown port";
 
   /* the STD handles will be in use by many devices, and their fd
@@ -811,30 +817,40 @@ int port_init(void)
 	port_handler[HANDLE_STD_IO].write_portb = std_port_outb;
 	port_handler[HANDLE_STD_IO].read_portw = std_port_inw;
 	port_handler[HANDLE_STD_IO].write_portw = std_port_outw;
+	port_handler[HANDLE_STD_IO].read_portd = std_port_ind;
+	port_handler[HANDLE_STD_IO].write_portd = std_port_outd;
 	port_handler[HANDLE_STD_IO].handler_name = "std port io";
 
 	port_handler[HANDLE_STD_RD].read_portb = std_port_inb;
 	port_handler[HANDLE_STD_RD].write_portb = port_not_avail_outb;
 	port_handler[HANDLE_STD_RD].read_portw = std_port_inw;
 	port_handler[HANDLE_STD_RD].write_portw = port_not_avail_outw;
+	port_handler[HANDLE_STD_RD].read_portd = std_port_ind;
+	port_handler[HANDLE_STD_RD].write_portd = port_not_avail_outd;
 	port_handler[HANDLE_STD_RD].handler_name = "std port read";
 
 	port_handler[HANDLE_STD_WR].read_portb = port_not_avail_inb;
 	port_handler[HANDLE_STD_WR].write_portb = std_port_outb;
 	port_handler[HANDLE_STD_WR].read_portw = port_not_avail_inw;
 	port_handler[HANDLE_STD_WR].write_portw = std_port_outw;
+	port_handler[HANDLE_STD_WR].read_portd = port_not_avail_ind;
+	port_handler[HANDLE_STD_WR].write_portd = std_port_outd;
 	port_handler[HANDLE_STD_WR].handler_name = "std port write";
 
 	port_handler[HANDLE_VID_IO].read_portb = video_port_in;
 	port_handler[HANDLE_VID_IO].write_portb = video_port_out;
 	port_handler[HANDLE_VID_IO].read_portw = port_not_avail_inw;
 	port_handler[HANDLE_VID_IO].write_portw = port_not_avail_outw;
+	port_handler[HANDLE_VID_IO].read_portd = port_not_avail_ind;
+	port_handler[HANDLE_VID_IO].write_portd = port_not_avail_outd;
 	port_handler[HANDLE_VID_IO].handler_name = "video port io";
 
 	port_handler[HANDLE_SPECIAL].read_portb = special_port_inb;
 	port_handler[HANDLE_SPECIAL].write_portb = special_port_outb;
 	port_handler[HANDLE_SPECIAL].read_portw = port_not_avail_inw;
 	port_handler[HANDLE_SPECIAL].write_portw = port_not_avail_outw;
+	port_handler[HANDLE_SPECIAL].read_portd = port_not_avail_ind;
+	port_handler[HANDLE_SPECIAL].write_portd = port_not_avail_outd;
 	port_handler[HANDLE_SPECIAL].handler_name = "extra stuff";
 
 	port_handles = STD_HANDLES;
@@ -846,6 +862,56 @@ int port_init(void)
 	return port_handles;	/* unused but useful */
 }
 
+/* port server: this function runs in a seperate process from the main
+   DOSEMU. This enables the main DOSEMU to drop root privileges. The
+   server can do that as well: by setting iopl(3).
+   Maybe this server should wrap DOSEMU rather than be forked from
+   it.
+*/
+static void port_server(void)
+{
+        struct portreq pr;
+	struct sigaction sa;
+	_port_handler *ph;
+        priv_iopl(3);
+	priv_drop();
+        close(port_fd_in[0]);
+        close(port_fd_out[1]);
+        g_printf("server started\n");
+        /* free low memory: the server doesn't need it and it saves
+           a bit of memory (~128K out of the 328K fork overhead) */
+        libless_munmap(0, 0x100000);
+	SETSIG(SIG_RELEASE, SIG_IGN);
+	SETSIG(SIG_ACQUIRE, SIG_IGN);
+        for (;;) {
+                read(port_fd_out[0], &pr, sizeof(pr));
+                if (pr.type >= TYPE_EXIT)
+                        exit(0);
+		ph = &EMU_HANDLER(pr.port);
+                switch (pr.type) {
+                case TYPE_INB:
+                        pr.word = ph->read_portb(pr.port);
+                        break;
+                case TYPE_OUTB:
+                        ph->write_portb(pr.port, pr.word);
+                        break;
+                case TYPE_INW:
+                        pr.word = ph->read_portw(pr.port);
+                        break;
+                case TYPE_OUTW:
+                        ph->write_portw(pr.port, pr.word);
+                        break;
+                case TYPE_IND:
+                        pr.word = ph->read_portd(pr.port);
+                        break;
+                case TYPE_OUTD:
+                        ph->write_portd(pr.port, pr.word);
+                        break;
+                }
+                write(port_fd_in[1], &pr, sizeof(pr));
+        }
+}
+
 /* 
  * SIDOC_BEGIN_FUNCTION extra_port_init()
  *
@@ -913,46 +979,53 @@ int extra_port_init(void)
 		for (i=0x3b4; i<0x3df; i++)
 			SET_HANDLE_COND(i,HANDLE_VID_IO);
 	}
-#if 1		/* HGC ports */
-	if (config.usesX) {
-		SET_HANDLE_COND(0x3b4,HANDLE_STD_IO);
-		SET_HANDLE_COND(0x3b5,HANDLE_STD_IO);
-		SET_HANDLE_COND(0x3ba,HANDLE_STD_IO);
-	}
-	SET_HANDLE_COND(0x3b8,HANDLE_SPECIAL);
-	SET_HANDLE_COND(0x3bf,HANDLE_SPECIAL);
-#endif
-#if X_GRAPHICS
-	/* DANG_FIXTHIS this code needs to be removed - it collides with vgaemu
-	 */
-	if (config.X) {
-#if 0
-		for (i=0x3c0; i<0x3c1; i++)
-			SET_HANDLE_COND(i,HANDLE_STD_IO);
-		for (i=0x3c4; i<0x3c9; i++)
-			SET_HANDLE_COND(i,HANDLE_STD_IO);
-#endif
-		for (i=0x3ce; i<0x3dd; i++)
-			SET_HANDLE_COND(i,HANDLE_STD_IO);
-	}
-#endif
-
 	if (!config.X) {
+ 	  SET_HANDLE_COND(0x3b8,HANDLE_SPECIAL);
+	  SET_HANDLE_COND(0x3bf,HANDLE_SPECIAL);
 	  SET_HANDLE_COND(0x3c0,HANDLE_SPECIAL);	/* W */
+  	  SET_HANDLE_COND(0x3ba,HANDLE_SPECIAL);		/* R */
+	  SET_HANDLE_COND(0x3da,HANDLE_SPECIAL);		/* R */
+	  SET_HANDLE_COND(0x3db,HANDLE_SPECIAL);		/* R */
 	}
-	SET_HANDLE_COND(0x3ba,HANDLE_SPECIAL);		/* R */
-	SET_HANDLE_COND(0x3da,HANDLE_SPECIAL);		/* R */
-	SET_HANDLE_COND(0x3db,HANDLE_SPECIAL);		/* R */
 
 	i = READ_WORD(BIOS_VIDEO_PORT);
-	if (i) {	/* !config.vga */
+	if (i && !config.X) {	/* !config.vga */
 	    SET_HANDLE_COND(i,HANDLE_SPECIAL);		/* W */
 	    SET_HANDLE_COND(i+1,HANDLE_SPECIAL);	/* W */
 	}
 
+        if (can_do_root_stuff) {
+                for (i = 0; i < sizeof(port_handle_table); i++) {
+                        if (config.pci || config.speaker == SPKR_NATIVE || (
+			    port_handle_table[i] >= HANDLE_STD_IO &&
+                            port_handle_table[i] <= HANDLE_STD_WR)) {
+                                /* fork the privileged port server */
+                                g_printf("starting port server\n");
+                                pipe(port_fd_out);
+                                pipe(port_fd_in);
+                                portserver_pid = fork();
+                                if (portserver_pid == 0) {
+                                        port_server();
+                                }
+                                close(port_fd_in[1]);
+                                close(port_fd_out[0]);
+                                break;
+                        }
+                }
+        }
+
  	return 0;
 }
 
+void port_exit(void)
+{
+	struct portreq pr;
+	if (!portserver_pid) return;
+        pr.type = TYPE_EXIT;
+	write(port_fd_out[1], &pr, sizeof(pr));
+	portserver_pid = 0;
+}
+
 void release_ports (void)
 {
 	int i;
@@ -1001,7 +1074,7 @@ int port_register_handler(emu_iodev_t device, int flags)
     if (handle >= port_handles) {
 	/* no existing handle found, create new one */
 	if (port_handles >= EMU_MAX_IO_DEVICES) {
-		error("PORT: too many IO devices, increase EMU_MAX_IO_DEVICES");
+		error("PORT: too many IO devices, increase EMU_MAX_IO_DEVICES\n");
 		leavedos(77);
 	}
 
@@ -1048,7 +1121,7 @@ int port_register_handler(emu_iodev_t device, int flags)
 
     i_printf("PORT: registered \"%s\" handle 0x%02x [0x%04x-0x%04x] fd=%d\n",
 	port_handler[handle].handler_name, handle, device.start_addr,
-	device.end_addr, (device.fd>=0? devstat.st_dev:device.fd));
+	device.end_addr, (int)(device.fd>=0? devstat.st_dev:device.fd));
 
     if (flags & PORT_FAST) {
 	if (device.start_addr < 0x400) {
@@ -1081,6 +1154,9 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 	char *devrname;
 	int fd, usemasks = 0;
 
+        if (!can_do_root_stuff)
+                return FALSE;
+
 	i_printf("PORT: allow_io for port 0x%04x:%d perm=%x or=%x and=%x\n",
 		 start, size, permission, ormask, andmask);
 
@@ -1113,7 +1189,7 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 		sscanf(line, "%x-%x : %s", &beg, &end, portname);
 		if ((start <= end) && ((start+size) > beg)) {
 			/* ports are besetzt, try to open the according device */
-			i_printf("PORT: in range 0x%04x-0x%04x already registered as %s\n",
+			i_printf("PORT: range 0x%04x-0x%04x already registered as %s\n",
 				 beg, end, portname);
 			if (!strncasecmp(portname,"dosemu",6)) return FALSE;
 			mapped = 1;
@@ -1140,7 +1216,6 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 		 * port access was granted.
 		 * SIDOC_END_REMARK
 		 */
-		PRIV_SAVE_AREA
 		int devperm;
 
 		devrname=strrchr(device,'/');
@@ -1157,9 +1232,7 @@ Boolean port_allow_io(ioport_t start, Bit16u size, int permission, Bit8u ormask,
 			default:	devperm = O_RDWR;
 					flags |= (PORT_DEV_RD|PORT_DEV_WR);
 		}
-		enter_priv_on();
 		io_device.fd = open(device, devperm);
-		leave_priv_setting();
 		if (io_device.fd == -1) {
 			switch (errno) {
 			case EBUSY:
@@ -1222,7 +1295,7 @@ set_ioperm(int start, int size, int flag)
 	PRIV_SAVE_AREA
 	int tmp;
 
-	if (!can_do_root_stuff || (start>0x3ff))
+	if ((!can_do_root_stuff && flag == 1) || (start>0x3ff))
 	    return -1;		/* don't bother */
 	if ((start+size)>0x3ff) size=0x400-start;
 
@@ -1235,7 +1308,16 @@ set_ioperm(int start, int size, int flag)
 	tmp = DOS_SYSCALL(ioperm(start, size, flag));
 	leave_priv_setting();
 
-	i_printf ("nPORT: set_ioperm [%4x:%2d:%d] returns %d\n",start,size,flag,tmp);
+#ifdef X86_EMULATOR
+	if (config.cpuemu && (tmp==0)) {
+	    int i;
+	    for (i=start; i<(start+size); i++)
+		(flag? set_bit(i,io_bitmap) : clear_bit(i,io_bitmap));
+	    i_printf("ePORT: set_ioperm [%x:%d:%d]\n",start,size,flag);
+	}
+#else
+	i_printf ("nPORT: set_ioperm [%x:%d:%d] returns %d\n",start,size,flag,tmp);
+#endif
 	return tmp;
 }
 
