@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -60,6 +60,8 @@
 #include "config.h"
 #include "emu.h"
 #include "vgaemu.h"
+#include "video.h"
+#include "memory.h"
 
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -80,7 +82,6 @@ static unsigned char crtc_ival[16][CRTC_MAX_INDEX + 1] = {
   {0x5f, 0x4f, 0x50, 0x82, 0x54, 0x80, 0x0b, 0x3e, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xea, 0x8c, 0xdf, 0x28, 0x00, 0xe7, 0x04, 0xc3, 0xff},
   {0x5f, 0x4f, 0x50, 0x82, 0x54, 0x80, 0x0b, 0x3e, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xea, 0x8c, 0xdf, 0x28, 0x00, 0xe7, 0x04, 0xe3, 0xff},
   {0x5f, 0x4f, 0x50, 0x82, 0x54, 0x80, 0xbf, 0x1f, 0x00, 0x41, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x9c, 0x8e, 0x8f, 0x28, 0x40, 0x96, 0xb9, 0xa3, 0xff},
-
   {0x5f, 0x4f, 0x50, 0x82, 0x54, 0x80, 0x0b, 0x3e, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xea, 0x8c, 0xdf, 0x28, 0x00, 0xe7, 0x04, 0xe3, 0xff}
 };
 
@@ -160,41 +161,64 @@ void CRTC_write_value(unsigned char data)
     case 0x04:
     case 0x05:
       if(NEWBITS(0xFF)) {
-        todo[todo_ind++] = CFG_CRTC_HEIGHT;
+        todo[todo_ind++] = CFG_CRTC_WIDTH;
       }
       crtc_deb("CRTC_write_value: crtc[0x%02x] = 0x%02x (guessed)\n", ind, u);
       break;
 
     case 0x06:
-    case 0x07:
-    case 0x09:
       if(NEWBITS(0xFF)) {
         todo[todo_ind++] = CFG_CRTC_HEIGHT;
       }
       crtc_deb("CRTC_write_value: crtc[0x%02x] = 0x%02x (guessed)\n", ind, u);
       break;
 
+    case 0x07:
+      if(NEWBITS(0x10)) {
+        todo[todo_ind++] = CFG_CRTC_LINE_COMPARE;
+      }
+      if(NEWBITS(0xEF)) {
+        todo[todo_ind++] = CFG_CRTC_HEIGHT;
+      }
+      crtc_deb("CRTC_write_value: crtc[0x%02x] = 0x%02x (guessed)\n", ind, u);
+      break;
+
+    case 0x09:
+      if(NEWBITS(0x40)) {
+        todo[todo_ind++] = CFG_CRTC_LINE_COMPARE;
+      }
+      if(NEWBITS(0xBF)) {
+        todo[todo_ind++] = CFG_CRTC_HEIGHT;
+      }
+      crtc_deb("CRTC_write_value: crtc[0x%02x] = 0x%02x (guessed)\n", ind, u);
+      break; 
+     
     case 0x0c:		/* Start Address High */
-      vga.display_start = (vga.display_start & 0xff) | (u << 8);
-      crtc_deb("CRTC_write_value: Start Address = 0x%04x\n", vga.display_start);
+      /* these shifts involving vga.crtc.addr_mode should really be rotations, 
+         depending on mode control bit 5 */
+      vga.display_start = (vga.crtc.data[0x0d] + (u << 8)) << vga.crtc.addr_mode;
+      screen_adr = SCREEN_ADR(0) + vga.display_start/2;
+      crtc_deb("CRTC_write_value: Start Address = 0x%04x, high changed\n", vga.display_start);
       break;
 
     case 0x0d:		/* Start Address Low */
-      vga.display_start = (vga.display_start & 0xff00) | u;
-      crtc_deb("CRTC_write_value: Start Address = 0x%04x\n", vga.display_start);
+      vga.display_start = (u + (vga.crtc.data[0x0c] << 8)) << vga.crtc.addr_mode;
+      /* this shift should really be a rotation, depending on mode control bit 5 */
+      screen_adr = SCREEN_ADR(0) + vga.display_start/2;
+      crtc_deb("CRTC_write_value: Start Address = 0x%04x, low changed\n", vga.display_start);
       break;
 
     case 0x0e:		/* Cursor Location High */
-      vga.crtc.cursor_location = (vga.crtc.cursor_location & 0xff) | (u << 8);
-      cursor_row = vga.crtc.cursor_location / (vga.scan_len/2);
-      cursor_col = vga.crtc.cursor_location % (vga.scan_len/2);
+      vga.crtc.cursor_location = (vga.crtc.data[0x0f] + (u << 8)) << vga.crtc.addr_mode; 
+      cursor_row = (vga.crtc.cursor_location - vga.display_start) / vga.scan_len;
+      cursor_col = ((vga.crtc.cursor_location - vga.display_start) % vga.scan_len) / 2;
       crtc_deb("CRTC_write_value: Cursor Location = 0x%04x\n", vga.crtc.cursor_location);
       break;
 
     case 0x0f:		/* Cursor Location  Low */
-      vga.crtc.cursor_location = (vga.crtc.cursor_location & 0xff00) | u;
-      cursor_row = vga.crtc.cursor_location / (vga.scan_len/2);
-      cursor_col = vga.crtc.cursor_location % (vga.scan_len/2);
+      vga.crtc.cursor_location = (u + (vga.crtc.data[0x0e] << 8)) << vga.crtc.addr_mode;
+      cursor_row = (vga.crtc.cursor_location - vga.display_start) / vga.scan_len;
+      cursor_col = ((vga.crtc.cursor_location - vga.display_start) % vga.scan_len) / 2;
       crtc_deb("CRTC_write_value: Cursor Location = 0x%04x\n", vga.crtc.cursor_location);
       break;
 
@@ -208,9 +232,10 @@ void CRTC_write_value(unsigned char data)
       break;
   
     case 0x13:          /* Number of bytes in a scanline */
-      vga.scan_len = data << (vga.crtc.addr_mode + 1);
-      vga.reconfig.re_init = 1;
-      crtc_deb("CRTC_write_value: scan_len = 0x%04x\n", vga.scan_len);
+      if(NEWBITS(0xFF)) {
+        todo[todo_ind++] = CFG_CRTC_ADDR_MODE;
+      }
+      crtc_deb("CRTC_write_value: crtc[0x%02x] = 0x%02x (guessed)\n", ind, u);
       break;
 
     case 0x14:		/* Underline Location */
@@ -242,6 +267,12 @@ void CRTC_write_value(unsigned char data)
       }
       break;
 
+    case 0x18: /* line compare */
+      if(NEWBITS(0xFF)) {
+        todo[todo_ind++] = CFG_CRTC_LINE_COMPARE;
+      }
+      break;
+      
     default:
       crtc_deb("CRTC_write_value: crtc[0x%02x] = 0x%02x (ignored)\n", ind, u);
   }
