@@ -1,6 +1,6 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 1999 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -18,7 +18,7 @@
  *			dawes@XFree86.Org
  *
  * Changes for DOSEMU:
- *		9/97 Alberto Vignani <alberto.vignani@torino.alpcom.it>
+ *		9/97 Alberto Vignani <vignani@tin.it>
  *
  */
 
