@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -27,7 +27,7 @@
  *
  */
 
-#include <features.h>
+#include "config.h"
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/ioctl.h>
@@ -41,11 +41,7 @@
 #include <signal.h>
 #include <sys/stat.h>
 #ifdef __linux__
-#if GLIBC_VERSION_CODE >= 2000
 #include <sys/vt.h>
-#else
-#include "Linux/vt.h"
-#endif
 #include <sys/kd.h>
 #endif
 
@@ -71,27 +67,15 @@
 #include "mapping.h"
 #include "timers.h"
 
-extern void child_close_mouse ();
-extern void child_open_mouse ();
-extern void clear_screen (int, int);
-extern inline void console_update_cursor (int, int, int, int);
-static void set_dos_video ();
-void put_video_ram ();
+static void set_dos_video (void);
 
-extern int
-  dosemu_sigaction (int sig, struct sigaction *, struct sigaction *);
 static void SIGRELEASE_call (void);
 static void SIGACQUIRE_call (void);
 
-void dump_video_regs (void);
-
 #ifdef WANT_DUMP_VIDEO
 static void dump_video (void);
 #endif
 
-void dump_video_linux (void);
-
-
 static int  color_text;
 
 #define MAXDOSINTS 032
@@ -100,15 +84,6 @@ static int  color_text;
 
 static void release_vt (int sig, struct sigcontext_struct context);
 static void acquire_vt (int sig, struct sigcontext_struct context);
-void get_video_ram (int);
-
-#if 1
-#undef SETSIG
-#define SETSIG(sa, sig, fun)	sa.sa_handler = fun; \
-				sa.sa_flags = 0; \
-				sa.sa_mask = 0; \
-				sigaction(sig, &sa, NULL);
-#endif
 
 static inline void
 forbid_switch (void)
@@ -133,8 +108,10 @@ void check_console(void) {
     major = chkbuf.st_rdev >> 8;
     minor = chkbuf.st_rdev & 0xff;
 
+    c_printf("major = %d minor = %d\n",
+	    major, minor);
     /* console major num is 4, minor 64 is the first serial line */
-    if (can_do_root_stuff && (major == 4) && (minor < 64)) {
+    if ((major == 4) && (minor < 64)) {
        scr_state.console_no = minor;
        config.console=1;
     }
@@ -162,36 +139,6 @@ allow_switch (void)
     }
 }
 
-static void
-parent_close_mouse (void)
-{
-  if (mice->intdrv)
-     {
-	if (mice->fd > 0) {
-   	   remove_from_io_select(mice->fd, mice->add_to_io_select);
-           DOS_SYSCALL(close (mice->fd));
-	}
-    }
-  else
-    child_close_mouse ();
-}
-
-static void
-parent_open_mouse (void)
-{
-  PRIV_SAVE_AREA
-  if (mice->intdrv)
-    {
-      enter_priv_on(); /* The mouse may not be a resource everyone can open. */
-      mice->fd = DOS_SYSCALL (open (mice->dev, O_RDWR | O_NONBLOCK));
-      leave_priv_setting();
-      if (mice->fd > 0)
-	add_to_io_select(mice->fd, mice->add_to_io_select);
-    }
-  else
-    child_open_mouse ();
-}
-
 static inline void
 SIGACQUIRE_call (void)
 {
@@ -270,7 +217,6 @@ set_linux_video (void)
 #endif
 	  restore_vga_state (&linux_regs);
 	}
-      release_perm ();
     }
 }
 
@@ -310,10 +256,17 @@ SIGRELEASE_call (void)
 #endif
 
 	  set_linux_video ();
-	  put_video_ram ();
+	  if (can_do_root_stuff) 
+	    { 
+	      release_perm ();
+	      put_video_ram ();
+	    }
 
 	  /*      if (config.vga) dos_pause(); */
 	  scr_state.current = 0;
+	  /* NOTE: if DOSEMU is not frozen then a DOS program in the background
+	     is capable of changing the screen appearance, even while in another
+	     console or X */
 	  freeze_dosemu();
 	}
     }
@@ -367,6 +320,7 @@ get_video_ram (int waitflag)
 
   char *textbuf = NULL, *vgabuf = NULL;
 
+  if (!can_do_root_stuff) return;
   v_printf ("get_video_ram STARTED\n");
   if (config.vga)
     {
@@ -549,15 +503,13 @@ set_process_control (void)
 void
 clear_process_control (void)
 {
-  PRIV_SAVE_AREA
   struct vt_mode vt_mode;
+  struct sigaction sa;
 
   vt_mode.mode = VT_AUTO;
-  enter_priv_on();
   ioctl (console_fd, VT_SETMODE, (int) &vt_mode);
-  leave_priv_setting();
-  signal (SIG_RELEASE, SIG_IGN);
-  signal (SIG_ACQUIRE, SIG_IGN);
+  SETSIG (SIG_RELEASE, SIG_IGN);
+  SETSIG (SIG_ACQUIRE, SIG_IGN);
 }
 
 
@@ -640,11 +592,11 @@ set_vc_screen_page (int page)
 int
 get_perm (void)
 {
-  permissions++;
-  if (permissions > 1)
+  if (permissions > 0)
     {
       return 0;
     }
+  permissions++;
   if (config.vga)
     {				/* hope this will not lead to problems with ega/cga */
       /* get I/O permissions for VGA registers */
@@ -697,7 +649,7 @@ get_perm (void)
 	  FCR_W = FCR_WM;
 	}
     }
-  else if (config.usesX || (config.console_video && (config.cardtype == CARD_MDA)))
+  else if (config.console_video && (config.cardtype == CARD_MDA))
     {
       if (set_ioperm (0x3b4, 1, 1) ||
 	  set_ioperm (0x3b5, 1, 1) ||
@@ -759,7 +711,7 @@ release_perm (void)
 	      leavedos (-1);
 	    }
 	}
-      else if (config.usesX || (config.console_video && (config.cardtype == CARD_MDA)))
+      else if (config.console_video && (config.cardtype == CARD_MDA))
 	{
 	  if (set_ioperm (0x3b4, 1, 0) ||
 	      set_ioperm (0x3b5, 1, 0) ||
@@ -1138,56 +1090,12 @@ void
 install_int_10_handler (void)
 {
   unsigned char *ptr;
-
-#if 0
-  if (!config.usesX)
-    {
-      /* Wrapper around call to video init c000:0003 */
-      ptr = (u_char *) INT10_ADD;
-
-      *ptr++ = 0x50;		/* push ax          */
-      *ptr++ = 0xfb;		/* start interrupts STI  */
-
-      *ptr++ = 0xb0;
-      *ptr++ = 0x08;		/* mov  al,08  ; Start Video init     */
-
-      *ptr++ = 0xcd;
-      *ptr++ = DOS_HELPER_INT;		/* int  e6          */
-
-      if (config.vbios_seg == 0xe000)
-	{
-	  *ptr++ = 0x9a;
-	  *ptr++ = 0x03;
-	  *ptr++ = 0x00;
-	  *ptr++ = 0x00;
-	  *ptr++ = 0xe0;	/* call e000:0003   */
-	}
-      else
-	{
-	  *ptr++ = 0x9a;
-	  *ptr++ = 0x03;
-	  *ptr++ = 0x00;
-	  *ptr++ = 0x00;
-	  *ptr++ = 0xc0;	/* call c000:0003   */
-	}
-      *ptr++ = 0xb0;
-      *ptr++ = 0x09;		/* move al,09   ; Finished video init    */
-
-      *ptr++ = 0xcd;
-      *ptr++ = DOS_HELPER_INT;		/* int  e6          */
-
-      *ptr++ = 0x58;		/* pop ax           */
-      *ptr++ = 0xfb;		/* start interrupts STI  */
-      *ptr++ = 0xcb;		/* retf             */
-    }
-#else
-  if (config.vbios_seg == 0xe000) {
-    extern void bios_f000_int10ptr();
+  
+  if (config.vbios_seg == 0xe000 && config.vbios_post) {
     ptr = (u_char *)((BIOSSEG << 4) + ((long)bios_f000_int10ptr - (long)bios_f000));
     *((long *)ptr) = 0xe0000003;
     v_printf("VID: new int10 handler at %p\n",ptr);
   }
   else
     v_printf("VID: install_int_10_handler: do nothing\n");
-#endif
 }
