@@ -1,6 +1,6 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -53,16 +53,13 @@
 #define	DEBUG_VBE		/* general info */
 
 
-#include <features.h>
 #include <string.h>
 #include "cpu.h"
 #include "emu.h"
 #include "video.h"
 #include "remap.h"
 #include "vgaemu.h"
-
-/* from dosext/dpmi/dpmi.c */
-unsigned long dpmi_GetSegmentBaseAddress(unsigned short);
+#include "dpmi.h"
 
 #define VBE_BIOS_MAXPAGES	2	/* max. 8k BIOS size, more than enough */
 
@@ -164,21 +161,6 @@ static int vbe_pm_interface(unsigned);
 static int vbe_power_state(unsigned, unsigned);
 
 
-/*
- * The actual VGA/VBE BIOS is in vesabios.S and vesabios_pm.S. These are
- * some labels pointing to interesting code portions. They are used in
- * vbe_init() to setup the struct vgaemu_bios which holds all relevant
- * info about our BIOS.
- */
-
-extern void vgaemu_bios_start();
-extern void vgaemu_bios_prod_name();
-extern void vgaemu_bios_win_func();
-extern void vgaemu_bios_end();
-extern void vgaemu_bios_pm_interface();
-extern void vgaemu_bios_pm_interface_end();
-
-
 /*
  * DANG_BEGIN_FUNCTION vbe_init
  *
