@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -187,14 +187,6 @@
 #define	DEBUG_BANK	0	/* (<= 2) bank switching */
 #define	DEBUG_COL	0	/* (<= 1) color interpretation changes */
 
-/*
- * whether we emulate only writes to VGA memory or read & write
- * (0 -> no instruction emulation, the 'normal' case)
- * cf. vga.inst_emu
- */
-#define EMU_WRITE_INST	1
-#define EMU_ALL_INST	2
-
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 #if !defined True
 #define False 0
@@ -264,10 +256,7 @@
 #endif
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-#include <features.h>
-#if GLIBC_VERSION_CODE == 2000
-#include <sigcontext.h>
-#endif
+#include <signal.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -278,6 +267,7 @@
 
 #include "cpu.h"		/* root@sjoerd: for context structure */
 #include "emu.h"
+#include "int.h"
 #include "dpmi.h"
 #include "port.h"
 #include "video.h"
@@ -291,20 +281,12 @@
 /* table with video mode definitions */
 #include "vgaemu_modelist.h"
 
-/* from dosext/dpmi/dpmi.c */
-unsigned long dpmi_GetSegmentBaseAddress(unsigned short);
-
-/* env/video/video.c */
-int load_file(char *name, int foffset, char *mstart, int msize);
-
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /*
  * functions local to this file
  */
 
-static int vga_emu_protect_page(unsigned, int);
 static int vga_emu_protect(unsigned, unsigned, int);
-static int vga_emu_adjust_protection(unsigned, unsigned);
 static int vga_emu_map(unsigned, unsigned);
 #if 0
 static int vgaemu_unmap(unsigned);
@@ -344,12 +326,18 @@ volatile static int my_fault = 0;
 
 static void chk_ro()
 {
-  volatile char *p = (char *) 0xa0077;
-  my_fault = 1;
-  *p = 0x99;
-  if(my_fault) {
-    vga_deb_map("vga_emu_fault: IS WRITABLE\n");
+#ifdef X86_EMULATOR
+  if (config.cpuemu<1) {
+#endif
+    volatile char *p = (char *) 0xa0077;
+    my_fault = 1;
+    *p = 0x99;
+    if(my_fault) {
+      vga_deb_map("vga_emu_fault: IS WRITABLE\n");
+    }
+#ifdef X86_EMULATOR
   }
+#endif
   my_fault = 0;
 }
 #endif
@@ -464,6 +452,13 @@ int VGA_emulate_outb(ioport_t port, Bit8u value)
 }
 
 
+int VGA_emulate_outw(ioport_t port, Bit16u value)
+{
+    if (VGA_emulate_outb(port,value)==0) return 0;
+    return VGA_emulate_outb(port+1,value>>8);
+}
+
+
 /*
  * DANG_BEGIN_FUNCTION VGA_emulate_inb
  *
@@ -576,6 +571,13 @@ Bit8u VGA_emulate_inb(ioport_t port)
   return uc;
 }
 
+Bit16u VGA_emulate_inw(ioport_t port)
+{
+    Bit16u v = VGA_emulate_inb(port);
+    return v | (VGA_emulate_inb(port+1)<<8);
+}
+
+
 /* 
  * This routine emulates a logical VGA read (CPU read  within the
  * VGA memory range a0000-bffff).  This usually doesn't just read the
@@ -683,9 +685,8 @@ static Bit32u rasterop(Bit32u value)
   return 0;
 }
 
-void Logical_VGA_write(unsigned offset, unsigned char value)
+static Bit32u Logical_VGA_CalcNewVal (unsigned char value)
 {
-  unsigned vga_page;
   Bit32u enable_set_reset, new_val = 0; 
   Bit8u bitmask;
 
@@ -696,17 +697,15 @@ void Logical_VGA_write(unsigned offset, unsigned char value)
       new_val |= new_val<<8;
       new_val |= new_val<<16; /* value extended over 4 bytes */
       enable_set_reset = color2pixels[EnableSetReset];
-      new_val = rasterop((new_val & ~enable_set_reset) | 
+      return rasterop((new_val & ~enable_set_reset) | 
                          (color2pixels[SetReset] & enable_set_reset));
-      break;
 
     case 1: /* write mode 1 */
-      new_val = *((unsigned *)VGALatch);;
+      return *((unsigned *)VGALatch);
       break;
 
     case 2: /* write mode 2 */
-      new_val = rasterop(color2pixels[value & 0xf]);
-      break;
+      return rasterop(color2pixels[value & 0xf]);
 
     case 3: /* write mode 3 */
       bitmask = BitMask; 
@@ -721,22 +720,31 @@ void Logical_VGA_write(unsigned offset, unsigned char value)
       break;
     }
   
+  return new_val;
+}
+
+void Logical_VGA_write(unsigned offset, unsigned char value)
+{
+  unsigned vga_page;
+  unsigned char *p;
+  Bit32u new_val;
+
+  new_val = Logical_VGA_CalcNewVal(value);
+
   vga_page = offset >> 12;
+  p = (unsigned char *)(vga.mem.base + offset);
 
   if(MapMask & 0x01) {
-    *(vga.mem.base + offset) = new_val & 0xff;
+    *p = new_val;
   }
-  
   if(MapMask & 0x02) {
-    *(vga.mem.base + offset + 0x10000) = (new_val >> 8) & 0xff;
+    p[0x10000] = new_val >> 8;
   }
-  
   if(MapMask & 0x04) {
-    *(vga.mem.base + offset + 0x20000) = (new_val >> 16) & 0xff;
+    p[0x20000] = new_val >> 16;
   }
-  
   if(MapMask & 0x08) {
-    *(vga.mem.base + offset + 0x30000) = new_val >> 24;
+    p[0x30000] = new_val >> 24;
   }
 
   if(MapMask) {
@@ -779,7 +787,7 @@ void Logical_VGA_write(unsigned offset, unsigned char value)
  *
  */     
 
-int vga_emu_fault(struct sigcontext_struct *scp)
+int vga_emu_fault(struct sigcontext_struct *scp, int pmode)
 {
   int i, j;
   unsigned page_fault, vga_page = 0, u, access_type, lin_addr;
@@ -809,7 +817,7 @@ int vga_emu_fault(struct sigcontext_struct *scp)
   }
 
   vga_deb_map("vga_emu_fault: %s%s access to %s region, address 0x%05x, page 0x%x, vga page 0x%x\n",
-    in_dpmi ? "dpmi " : "", access_type? "write" : "read",
+    pmode ? "dpmi " : "", access_type? "write" : "read",
     txt1[i], lin_addr, page_fault, vga_page
   );
 
@@ -819,7 +827,7 @@ int vga_emu_fault(struct sigcontext_struct *scp)
     (unsigned) REG(cs), (unsigned) REG(eip)
   );
 
-  if(in_dpmi) {
+  if(pmode) {
 #if DEBUG_MAP >= 1
     cs_ip = (unsigned char *) (dpmi_GetSegmentBaseAddress(_cs) + _eip);
 #endif
@@ -841,9 +849,15 @@ int vga_emu_fault(struct sigcontext_struct *scp)
 
   if(i == VGAEMU_MAX_MAPPINGS) {
     if(page_fault >= 0xa0 && page_fault < 0xc0) {	/* unmapped VGA area */
-      u = instr_len(SEG_ADR((unsigned char *), cs, ip));
-      LWORD(eip) += u;
-      if(!in_dpmi && u) {
+      if (pmode) {
+        u = instr_len((unsigned char *)SEL_ADR(_cs, _eip));
+        _eip += u;
+      }
+      else {
+        u = instr_len(SEG_ADR((unsigned char *), cs, ip));
+        LWORD(eip) += u;
+      }
+      if(u) {
         vga_msg("vga_emu_fault: attempted write to unmapped area (cs:ip += %u)\n", u);
       }
       else {
@@ -853,9 +867,15 @@ int vga_emu_fault(struct sigcontext_struct *scp)
       return True;
     }
     else if(page_fault >= 0xc0 && page_fault < (0xc0 + vgaemu_bios.pages)) {	/* ROM area */
-      u = instr_len(SEG_ADR((unsigned char *), cs, ip));
-      LWORD(eip) += u;
-      if(!in_dpmi && u) {
+      if (pmode) {
+        u = instr_len((unsigned char *)SEL_ADR(_cs, _eip));
+        _eip += u;
+      }
+      else {
+        u = instr_len(SEG_ADR((unsigned char *), cs, ip));
+        LWORD(eip) += u;
+      }
+      if(u) {
         vga_msg("vga_emu_fault: attempted write to ROM area (cs:ip += %u)\n", u);
       }
       else {
@@ -875,16 +895,34 @@ int vga_emu_fault(struct sigcontext_struct *scp)
  
   if(vga_page < vga.mem.pages) {
     vga.mem.dirty_map[vga_page] = 1;
+#ifdef X86_EMULATOR
+    if (config.cpuemu>1) {
+	error("VGAEmu: CPU emulation collision, should not be here\n");
+	leavedos(0x4945);
+    }
+#endif
     if(!vga.inst_emu) {
       /* Normal: make the display page writeable after marking it dirty */
       vga_emu_adjust_protection(vga_page, page_fault);
     }  
-    else  
+    else {
+#if 1
+      /* XXX Hack: dosemu touched the protected page of video mem, which is
+       * a bug. However for the video modes that do not require instremu, we
+       * can allow that access. For the planar modes we cant and will fail :(
+       */
+      if(pmode && _cs == UCODESEL) {
+	error("BUG: dosemu touched the protected video memory!!!\n");
+	return False;
+      }
+#endif
+
       /* A VGA mode PL4/PL2 video memory access has been trapped 
        * while we are using X.  Leave the display page read/write-protected
        * so that each instruction that accesses it can be trapped and
        * simulated. */
-            instr_emu(scp);
+        instr_emu(scp, pmode);
+    }
   }
   return True;
 }
@@ -931,7 +969,7 @@ static int vgaemu_prot_ok(unsigned page, int prot)
  *
  */
 
-static int vga_emu_protect_page(unsigned page, int prot)
+int vga_emu_protect_page(unsigned page, int prot)
 {
   int i;
   int sys_prot;
@@ -1039,7 +1077,7 @@ static int vga_emu_protect(unsigned page, unsigned mapped_page, int prot)
  *
  */
 
-static int vga_emu_adjust_protection(unsigned page, unsigned mapped_page)
+int vga_emu_adjust_protection(unsigned page, unsigned mapped_page)
 {
   int prot;
   int i, err, j, k;
@@ -1049,7 +1087,11 @@ static int vga_emu_adjust_protection(unsigned page, unsigned mapped_page)
     return 1;
   }
 
-  prot = vga.inst_emu == EMU_ALL_INST ? NONE : RO;
+  prot = (vga.inst_emu==EMU_ALL_INST
+#ifdef X86_EMULATOR
+	|| config.cpuemu>1
+#endif
+	? NONE : RO);
 
   if(!vga.inst_emu) {
     if(vga.mem.dirty_map[page]) prot = RW;
@@ -1166,6 +1208,11 @@ static int vga_emu_map(unsigned mapping, unsigned first_page)
   if(vmt->pages + first_page > vga.mem.pages) return 2;
 
   /* default protection for dirty pages */
+#ifdef X86_EMULATOR
+  if (config.cpuemu>1)
+      prot = VGA_EMU_NONE_PROT;
+  else
+#endif
   switch(vga.inst_emu) {
     case EMU_WRITE_INST:
       prot = VGA_EMU_RO_PROT;
@@ -1381,8 +1428,8 @@ int vga_emu_init(vgaemu_display_type *vedt)
   /* register VGA ports */
   io_device.read_portb = VGA_emulate_inb;
   io_device.write_portb = (void (*)(ioport_t, Bit8u)) VGA_emulate_outb;
-  io_device.read_portw = NULL;
-  io_device.write_portw = NULL;
+  io_device.read_portw = VGA_emulate_inw;
+  io_device.write_portw = (void (*)(ioport_t, Bit16u)) VGA_emulate_outw;
   io_device.read_portd = NULL;
   io_device.write_portd = NULL;
   io_device.irq = EMU_NO_IRQ;
@@ -1454,6 +1501,10 @@ int vga_emu_init(vgaemu_display_type *vedt)
   }
 #endif
 
+  /* point int 1f to the default 8x8 graphics font for high characters */
+  SETIVEC(0x1f, 0xc000, vgaemu_bios.font_8 + 128 * 8);
+  video_ints[0x1f] = video_ints[0x43] = 1;
+
   vga_msg(
     "vga_emu_init: memory: %u kbyte at 0x%x (lfb at 0x%x); %ssupport for mono modes\n",
     vga.mem.size >> 10, (unsigned) vga.mem.base,
@@ -2029,6 +2080,7 @@ int vga_emu_setmode(int mode, int width, int height)
   vga.mode_type = vmi->type;
   vga.width = vmi->width;
   vga.height = vmi->height;
+  vga.line_compare = vmi->height;
   vga.scan_len = (vmi->width + 3) & ~3;	/* dword aligned */
   vga.text_width = vmi->text_width;
   vga.text_height = vmi->text_height;
@@ -2040,7 +2092,7 @@ int vga_emu_setmode(int mode, int width, int height)
     vga.pixel_size = (vga.pixel_size + 7) & ~7;		/* assume byte alignment for these modes */
     vga.scan_len *= vga.pixel_size >> 3;
   }
-  vga.inst_emu = vga.mode_type == PL4 || vga.mode_type == PL2 ? EMU_ALL_INST : 0;
+  vga.inst_emu = ((vga.mode_type==PL4 || vga.mode_type==PL2) ? EMU_ALL_INST : 0);
 
   vga_msg("vga_emu_setmode: scan_len = %d\n", vga.scan_len);
   i = vga.scan_len;
@@ -2266,7 +2318,7 @@ void dirty_all_video_pages()
  * Set visible text page.
  *
  * `vga.display_start' is set to `page' * `page_size'.
- * This function works only in text modes.
+ * In graphics modes the page_size is determined from memory/vga settings
  *
  * arguments:
  * page      - Number of the text page.
@@ -2279,8 +2331,13 @@ void dirty_all_video_pages()
 int vga_emu_set_text_page(unsigned page, unsigned page_size)
 {
   if(vga.mode_class != TEXT) {
-    vga_msg("vga_emu_set_text_page: not in text mode\n");
-    return 1;
+    page_size = vga.scan_len * vga.height;
+    if (page_size != 0) {
+      page_size = vga.mem.bank_pages * 4096 / page_size;
+      if (page_size != 0) {
+	page_size = vga.mem.bank_pages * 4096 / page_size;
+      }
+    }
   }
 
   if((page + 1) * page_size > vga.mem.size) {
@@ -2444,23 +2501,34 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
       u = vga.seq.data[4] & 4 ? 0 : 1;
       u = vga.seq.data[4] & 8 ? 2 : u;
       vga.seq.addr_mode = u;
-      if(u != u0 && vga.mode_type == P8) {	/* ++HACK++ */
-        u1 = vga.seq.addr_mode == 0 ? 4 : 1;
-        vga.scan_len = (vga.scan_len * vga.mem.planes) / u1;
-        if(u1 != vga.mem.planes) {
-          vga.mem.planes = u1; vga.reconfig.mem = 1;   
-          vga_msg("vgaemu_adj_cfg: mem reconfig (%u planes)\n", u1);
-        }
+      u1 = vga.seq.addr_mode == 0 ? 4 : 1;
+      if(u1 != vga.mem.planes) {
+        vga.mem.planes = u1; vga.reconfig.mem = 1;   
+        vga_msg("vgaemu_adj_cfg: mem reconfig (%u planes)\n", u1);
       }
-      if(msg || u != u0) vga_msg("vgaemu_adj_cfg: seq.addr_mode = %s\n", txt1[u]);
+      if(msg || u != u0) vga_msg("vgaemu_adj_cfg: seq.addr_mode = %s\n", txt1[u]); 
     break;
 
     case CFG_CRTC_ADDR_MODE:
       u0 = vga.crtc.addr_mode;
+      u1 = vga.scan_len;
       u = vga.crtc.data[0x17] & 0x40 ? 0 : 1;
       u = vga.crtc.data[0x14] & 0x40 ? 2 : u;
       vga.crtc.addr_mode = u;
-      if(msg || u != u0) vga_msg("vgaemu_adj_cfg: crtc.addr_mode = %s\n", txt2[u]);
+      vga.display_start = (vga.crtc.data[0x0d] + (vga.crtc.data[0x0c] << 8)) << 
+	      vga.crtc.addr_mode;
+      /* this shift should really be a rotation, depending on mode control bit 5 */
+      screen_adr = SCREEN_ADR(0) + vga.display_start;
+      vga.crtc.cursor_location =  (vga.crtc.data[0x0f] + (vga.crtc.data[0x0e] << 8)) <<
+              vga.crtc.addr_mode;
+      vga.scan_len = vga.crtc.data[0x13] << (vga.crtc.addr_mode + 1);
+      if (vga.scan_len == 0) vga.scan_len = 160;
+      if (vga.mode_class == TEXT) co = vga.scan_len / 2;
+      cursor_row = (vga.crtc.cursor_location - vga.display_start) / vga.scan_len;
+      cursor_col = ((vga.crtc.cursor_location - vga.display_start) % vga.scan_len) / 2;
+      if (u1 != vga.scan_len) vga.reconfig.mem = 1;
+      if(msg || u != u0) vga_msg("vgaemu_adj_cfg: crtc.addr_mode = %s, scan_len = %d\n",
+	txt2[u], vga.scan_len);
     break;
 
     case CFG_CRTC_HEIGHT:
@@ -2503,8 +2571,15 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
       vga_msg("vgaemu_adj_cfg: vertical_display_end = %d\n", vertical_display_end);
       vga_msg("vgaemu_adj_cfg: vertical_multiplier = %d\n", vertical_multiplier);
       vga_msg("vgaemu_adj_cfg: height = %d\n", height);
-      vga.height = height;
-      vga.reconfig.display = 1;
+      if (vga.line_compare != vga.crtc.line_compare / vertical_multiplier) {
+        vga.line_compare = vga.crtc.line_compare / vertical_multiplier;
+        dirty_all_video_pages();
+      }
+      if (vga.height != height) {
+        vga.height = height;
+        vga.reconfig.display = 1;
+      }
+      if (vga.line_compare == 0) vga.line_compare = vga.height;
       break;
     }
     case CFG_CRTC_WIDTH:
@@ -2524,8 +2599,7 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
       horizontal_blanking_end = vga.crtc.data[0x3] & 0xF;
       horizontal_retrace_start = vga.crtc.data[0x4];
       horizontal_retrace_end = vga.crtc.data[0x5] & 0xF;
-      multiplier = 8; /* Can this be 9? */
-      /* Do I need to do something special for text modes? */
+      multiplier = (vga.mode_class == TEXT) ? 9 : (vga.mode_type == P8) ? 4 : 8;
       width = horizontal_display_end * multiplier;
       vga_msg("vgaemu_adj_cfg: horizontal_total = %d\n", horizontal_total);
       vga_msg("vgaemu_adj_cfg: horizontal_retrace_start = %d\n", horizontal_retrace_start);
@@ -2535,10 +2609,28 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
       vga_msg("vgaemu_adj_cfg: horizontal_display_end = %d\n", horizontal_display_end);
       vga_msg("vgaemu_adj_cfg: multiplier = %d\n", multiplier);
       vga_msg("vgaemu_adj_cfg: width = %d\n", width);
-      vga.width = width;
-      vga.reconfig.display = 1;
+      if (vga.width != width) {
+	 vga.width = width;
+	 vga.reconfig.display = 1;
+      }
       break;
     }
+    case CFG_CRTC_LINE_COMPARE:
+    {
+      int vertical_multiplier;
+      vga.crtc.line_compare =
+              vga.crtc.data[0x18] +
+              ((vga.crtc.data[0x7] & 0x10) << (8 - 4)) +
+              ((vga.crtc.data[0x9] & 0x40) << (9 - 6));
+      vertical_multiplier = ((vga.crtc.data[0x9] & 0x1F) +1) <<
+	      ((vga.crtc.data[0x9] & 0x80) >> 7);
+      vga_msg("vgaemu_adj_cfg: line_compare = %d\n", vga.crtc.line_compare);
+      if (vga.line_compare != vga.crtc.line_compare / vertical_multiplier) {
+        vga.line_compare = vga.crtc.line_compare / vertical_multiplier;
+        dirty_all_video_pages();
+      }
+      if (vga.line_compare == 0) vga.line_compare = vga.height;
+    }
     default:
       vga_msg("vgaemu_adj_cfg: unknown item %u\n", what);
   }
@@ -2596,7 +2688,7 @@ unsigned vgaemu_xy2ofs(unsigned x, unsigned y)
 
 void vgaemu_move_vga_mem(unsigned dst, unsigned src, unsigned len)
 {
-  void *dp, *sp;
+  char *dp, *sp;
 
   if(len == 0) return;
 
