@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -70,6 +70,8 @@
 
 #define DOS_HELPER_MOUSE_HELPER     0x33
 
+#define DOS_HELPER_PAUSE_KEY        0x35
+
 #define DOS_HELPER_CDROM_HELPER     0x40
 
 #define DOS_HELPER_ASPI_HELPER      0x41
@@ -98,7 +100,10 @@
 
 #define DOS_HELPER_GETCWD           0x80
 #define DOS_HELPER_CHDIR            0x81
+#define DOS_HELPER_GETPID           0x82
 
+#define DOS_HELPER_CPUEMUON         0x90
+#define DOS_HELPER_CPUEMUOFF        0x91
 
 #define DOS_HELPER_XCONFIG          0xa0
 
