@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -12,26 +12,36 @@
 #ifndef EMU_H
 #define EMU_H
 
-#include <features.h>
+#include "config.h"
+
 #include <sys/types.h>
 #include <setjmp.h>
-#if GLIBC_VERSION_CODE == 2000
-#include <sigcontext.h>
-#endif
 #include <signal.h> 
-#include "config.h"
+
+#if defined(HAVE_KEYBOARD_V1) && (HAVE_KEYBOARD_V1 > 1)
+  #error "Sorry, wrong keyboard code version for this DOSEMU version"
+#endif
+
+#if defined(HAVE_UNICODE_KEYB) && (HAVE_UNICODE_KEYB != 2)
+  #error "Sorry, wrong unicode keyboard code version for this DOSEMU version"
+#endif
+
 #include "types.h"
 #include "extern.h"
 #include "machcompat.h"
 #include "cpu.h"
 #include "vm86plus.h"
 #include "priv.h"
+#include "mouse.h"
 
 #include "extern.h"
 
 #if 1 /* Set to 1 to use Silly Interrupt generator */
 #define SIG 1
 typedef struct { int fd; int irq; } SillyG_t;
+extern SillyG_t *SillyG;
+extern int SillyG_do_irq(void);
+extern inline void irq_select(void);
 #endif
 
 #define inline __inline__
@@ -52,8 +62,8 @@ EXTERN struct vm86plus_struct vm86s INIT ( {
 } );
 
 EXTERN volatile sig_atomic_t signal_pending INIT(0);
+EXTERN volatile int fault_cnt INIT(0);
 EXTERN fd_set fds_sigio, fds_no_sigio;
-EXTERN unsigned int use_sigio INIT(0);
 EXTERN unsigned int not_use_sigio INIT(0);
 EXTERN int terminal_pipe;
 EXTERN int terminal_fd INIT(-1);
@@ -99,16 +109,9 @@ EXTERN int console_fd INIT(-1);
 EXTERN int mem_fd INIT(-1);
 EXTERN int in_readkeyboard;
 
-/* X-pipes */
-EXTERN int keypipe;
-EXTERN int mousepipe;
+EXTERN volatile int in_vm86 INIT(0);
 
-EXTERN int in_vm86 INIT(0);
-
-EXTERN int li, co;	/* lines, columns */
 EXTERN int scanseq;
-EXTERN int cursor_row;
-EXTERN int cursor_col;
 
 #if 0
 void dos_ctrl_alt_del(void);	/* disabled */
@@ -117,9 +120,10 @@ extern jmp_buf NotJEnv;
 
 EXTERN void run_vm86(void);
 EXTERN void loopstep_run_vm86(void);
-EXTERN void     vm86_GP_fault();
+EXTERN void vm86_GP_fault(void);
 
 EXTERN void do_call_back(Bit32u codefarptr);
+EXTERN void do_intr_call_back(int intno, int inter);
 
 #define NOWAIT  0
 #define WAIT    1
@@ -141,14 +145,7 @@ void getKeys(void);
      int queue_ioctl(int, int, int), do_ioctl(int, int, int);
      void keybuf_clear(void);
 
-     int set_ioperm(int, int, int);
-
-EXTERN struct debug_flags d INIT({0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0});
-#ifdef DONT_DEBUG_BOOT
-EXTERN struct debug_flags d_save;
-#endif
- 
-EXTERN u_char in_sighandler, in_ioctl;
+EXTERN u_char in_sighandler;
 /* one-entry queue ;-( for ioctl's */
 EXTERN struct ioctlq iq INIT({0, 0, 0, 0}); 
 EXTERN u_char in_ioctl INIT(0);
@@ -198,25 +195,6 @@ EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
       } while ((s_tmp == -1) ); \
   s_tmp; })
 
-#ifndef FALSE
-#define FALSE	0
-#endif
-#ifndef TRUE
-#define TRUE	1
-#endif
-
-#ifndef True
-#define True  TRUE
-#endif
-#ifndef False
-#define False FALSE
-#endif
-
-#ifndef DOSEMU_BOOL_IS_CHAR
-typedef unsigned char boolean;
-#define DOSEMU_BOOL_IS_CHAR
-#endif
-
 /* would like to have this in memory.h (but emu.h is included before memory.h !) */
 #define HARDWARE_RAM_START 0xc8000
 #define HARDWARE_RAM_STOP  0xf0000
@@ -230,12 +208,16 @@ typedef struct vesamode_type_struct {
      typedef struct config_info {
        int hdiskboot;
 
+#ifdef X86_EMULATOR
+       boolean cpuemu;
+#endif
+       int CPUSpeedInMhz;
        /* for video */
        boolean console;
        boolean console_video;
-       boolean graphics;
        boolean vga;
        boolean X;
+       boolean X_fullscreen;
        u_short cardtype;
        u_short chipset;
        boolean pci;
@@ -247,6 +229,7 @@ typedef struct vesamode_type_struct {
        u_short term_updatefreq;		/* Terminal update frequency */
        u_short term_charset;		/* Terminal Character set */
        u_short term_esc_char;	        /* ASCII value used to access slang help screen */
+       char    *xterm_title;	        /* xterm/putty window title */
        /* u_short term_corner; */       /* Update char at lower-right corner */
        u_short X_updatelines;           /* Amount to update at a time */
        u_short X_updatefreq;            /* X update frequency */
@@ -271,57 +254,52 @@ typedef struct vesamode_type_struct {
        vesamode_type *vesamode_list;	/* chained list of VESA modes */
        int     X_lfb;			/* support VESA LFB modes */
        int     X_pm_interface;		/* support protected mode interface */
+       int     X_background_pause;	/* pause xdosemu if it loses focus */
        boolean fullrestore;
        boolean force_vt_switch;         /* in case of console_video force switch to emu VT at start */
        int     dualmon;
 
-       u_short usesX;  /* !=0 if dosemu owns an X window */
-
        boolean console_keyb;
        boolean kbd_tty;
        boolean X_keycode;	/* use keycode field of event structure */
        boolean exitearly;
        int     realcpu;
-       boolean mathco, smp;
+       boolean mathco, smp, cpummx;
        boolean ipxsup;
-       boolean vnet;
-       boolean keybint;
+       int     vnet;
+       char   *netdev;
+       boolean pktdrv;
        boolean dosbanner;
-       boolean allowvideoportaccess;
        boolean emuretrace;
        boolean rdtsc;
-       boolean timers;
-       boolean mouse_flag;
        boolean mapped_bios;	/* video BIOS */
        char *vbios_file;	/* loaded VBIOS file */
        boolean vbios_copy;
        int vbios_seg;           /* VGA-BIOS-segment for mapping */
        int vbios_size;          /* size of VGA-BIOS (64K for vbios_seg=0xe000
        						     32K for vbios_seg=0xc000) */
+       boolean vbios_post;
 
        boolean bootdisk;	/* Special bootdisk defined */
        int  fastfloppy;
        char *emusys;		/* map CONFIG.SYS to CONFIG.EMU */
-       char *emubat;		/* map AUTOEXEC.BAT to AUTOEXEC.EMU */
        char *emuini;           /* map system.ini to  system.EMU */
 
        u_short speaker;		/* 0 off, 1 native, 2 emulated */
        u_short fdisks, hdisks;
        u_short num_lpt;
        u_short num_ser;
-       u_short num_mice;
+       mouse_t mouse;
 
        int pktflags;		/* global flags for packet driver */
 
        unsigned int update, freq;	/* temp timer magic */
-       unsigned int wantdelta, realdelta;
        unsigned long cpu_spd;		/* (1/speed)<<32 */
        unsigned long cpu_tick_spd;	/* (1.19318/speed)<<32 */
 
        unsigned int hogthreshold;
 
        int mem_size, xms_size, ems_size, dpmi, max_umb;
-       int secure;
        unsigned int ems_frame;
        char must_spare_hardware_ram;
        char hardware_pages[ ((HARDWARE_RAM_STOP-HARDWARE_RAM_START) >> 12)+1 ];
@@ -331,6 +309,7 @@ typedef struct vesamode_type_struct {
        				   (bitmask, bit3..15 ==> IRQ3 .. IRQ15) */
 
        struct keytable_entry *keytable;
+       struct keytable_entry *altkeytable;
 
        unsigned short detach;
        unsigned char *debugout;
@@ -338,6 +317,7 @@ typedef struct vesamode_type_struct {
        unsigned char *pre_stroke_mem;
 
        /* Lock File business */
+       boolean full_file_locks;
        char *tty_lockdir;	/* The Lock directory  */
        char *tty_lockfile;	/* Lock file pretext ie LCK.. */
        boolean tty_lockbinary;	/* Binary lock files ? */
@@ -363,12 +343,34 @@ typedef struct vesamode_type_struct {
        int features[1];
 
        /* Sound emulation */
-       __u16 sb_base;
-       __u8 sb_dma;
-       __u8 sb_irq;
+       int sound;
+       uint16_t sb_base;
+       uint8_t sb_dma;
+       uint8_t sb_hdma;
+       uint8_t sb_irq;
        char *sb_dsp;
        char *sb_mixer;
-       __u16 mpu401_base;
+       uint16_t mpu401_base;
+       char *sound_driver;
+       /* OSS-specific options */
+       int oss_min_frags;
+       int oss_max_frags;
+       int oss_stalled_frags;
+       int oss_do_post;
+       int oss_min_extra_frags;
+
+       /* joystick */
+       char *joy_device[2];
+       
+       /* range for joystick axis values */
+       int joy_dos_min;		/* must be > 0 */
+       int joy_dos_max;		/* avoid setting this to > 250 */
+       
+       int joy_granularity;	/* the higher, the less sensitive - for wobbly joysticks */
+       int joy_latency;		/* delay between nonblocking linux joystick reads */
+
+       int cli_timeout;		/* cli timeout hack */
+       int pic_watchdog;        /* pic watchdog reschedule hack */      
      }
 
 config_t;
@@ -383,9 +385,6 @@ config_t;
  */
 #define SIG_SER		SIGTTIN
 
-#define SIG_TIME	SIGALRM
-#define TIMER_TIME	ITIMER_REAL
-
 #define IO_READ  1
 #define IO_WRITE 2
 #define IO_RDWR	 (IO_READ | IO_WRITE)
@@ -402,29 +401,24 @@ EXTERN __inline__ void parent_nextscan(void);
 EXTERN __inline__ void disk_close(void);
 EXTERN void cpu_setup(void);
 EXTERN void real_run_int(int);
-#ifdef USE_NEW_INT
-  #define run_int do_int
-#else
-  #define run_int real_run_int
-#endif
+#define run_int do_int
 EXTERN int mfs_redirector(void);
-EXTERN void int10(void);
-EXTERN void int13(u_char);
-EXTERN void int14(u_char);
-EXTERN void int17(u_char);
+EXTERN int int10(void);
+EXTERN int int13(void);
+EXTERN int int16(void);
+EXTERN int int17(void);
 EXTERN void io_select(fd_set);
+EXTERN void io_select_init(void);
 EXTERN int pd_receive_packet(void);
 EXTERN int printer_tick(u_long);
-EXTERN int printer_tick(u_long);
 EXTERN void floppy_tick(void);
 EXTERN void open_kmem(void);
 EXTERN void close_kmem(void);
-EXTERN void CloseNetworkLink(int);
 EXTERN int parse_config(char *, char *);
+EXTERN void prepare_dexe_load(char *name);
 EXTERN void disk_init(void);
 EXTERN void serial_init(void);
 EXTERN void close_all_printers(void);
-EXTERN void release_ports(void);
 EXTERN void serial_close(void);
 EXTERN void disk_close_all(void);
 EXTERN void init_all_printers(void);
@@ -432,18 +426,22 @@ EXTERN int mfs_inte6(void);
 EXTERN void pkt_helper(void);
 EXTERN short pop_word(struct vm86_regs *);
 EXTERN void ems_init(void);
-EXTERN int GetDebugFlagsHelper(char *, int);
-EXTERN int SetDebugFlagsHelper(char *);
 EXTERN void leavedos(int) NORETURN;
-EXTERN void add_to_io_select(int, unsigned char);
-EXTERN void remove_from_io_select(int, unsigned char);
+EXTERN void add_to_io_select(int, u_char, void(*)(void));
+EXTERN void remove_from_io_select(int, u_char);
 EXTERN void sigquit(int);
 #ifdef __linux__
 EXTERN void sigalrm(int, struct sigcontext_struct);
+EXTERN void e_sigalrm(struct sigcontext_struct *context);
 EXTERN void sigio(int, struct sigcontext_struct);
 EXTERN int dosemu_sigaction(int sig, struct sigaction *new, struct sigaction *old);
+EXTERN void SIG_init(void);
+EXTERN void SIG_close(void);
 #endif
 
+/* set if sigaltstack(2) is available */
+EXTERN int have_working_sigaltstack;
+
 /* signals for Linux's process control of consoles */
 #define SIG_RELEASE     SIGWINCH
 #define SIG_ACQUIRE     SIGUSR2
@@ -479,41 +477,55 @@ EXTERN int dosemu_sigaction(int sig, struct sigaction *new, struct sigaction *ol
 #define ADDSET_SIGNALS_THAT_QUEUE(x) \
 do { \
        sigaddset(x, SIGIO); \
-       sigaddset(x, SIG_TIME); \
+       sigaddset(x, SIGALRM); \
        sigaddset(x, SIG_RELEASE); \
        sigaddset(x, SIG_ACQUIRE); \
 } while(0)
 
 #ifdef __linux__
-#if GLIBC_VERSION_CODE
-#define SignalHandler __sighandler_t
+#ifndef SA_ONSTACK
+#define SA_ONSTACK 0
+#undef HAVE_SIGALTSTACK
 #endif
-#define NEWSETQSIG(sig, fun)	sa.sa_handler = (__sighandler_t)fun; \
-			/* Point to the top of the stack, minus 4 \
-			   just in case, and make it aligned  */ \
-			sa.sa_restorer = \
-			(void (*)()) (((unsigned int)(cstack) + sizeof(cstack) - 4) & ~3); \
-					sa.sa_flags = SA_RESTART ; \
-					sigemptyset(&sa.sa_mask); \
-					ADDSET_SIGNALS_THAT_QUEUE(&sa.sa_mask); \
-					dosemu_sigaction(sig, &sa, NULL);
+#define SignalHandler __sighandler_t
+#define NEWSETQSIG(sig, fun) \
+	sa.sa_handler = (__sighandler_t)fun; \
+	sa.sa_flags = SA_RESTART; \
+	sigemptyset(&sa.sa_mask); \
+	ADDSET_SIGNALS_THAT_QUEUE(&sa.sa_mask); \
+	if (have_working_sigaltstack) { \
+		sa.sa_flags |= SA_ONSTACK; \
+		sigaction(sig, &sa, NULL); \
+	} else { \
+		/* Point to the top of the stack, minus 4 \
+		   just in case, and make it aligned  */ \
+		sa.sa_restorer = \
+		(void (*)(void)) (((unsigned int)(cstack) + sizeof(cstack) - 4) & ~3); \
+		dosemu_sigaction(sig, &sa, NULL); \
+	}
 
-#define SETSIG(sig, fun)	sa.sa_handler = (SignalHandler)fun; \
-					sa.sa_flags = SA_RESTART; \
-					sigemptyset(&sa.sa_mask); \
-					sigaddset(&sa.sa_mask, SIG_TIME); \
-					sigaction(sig, &sa, NULL);
+#define SETSIG(sig, fun) \
+	sa.sa_handler = (SignalHandler)fun; \
+	sa.sa_flags = SA_RESTART; \
+	sigemptyset(&sa.sa_mask); \
+	sigaddset(&sa.sa_mask, SIGALRM); \
+	sigaction(sig, &sa, NULL);
 
 #define NEWSETSIG(sig, fun) \
-			sa.sa_handler = (__sighandler_t) fun; \
-			/* Point to the top of the stack, minus 4 \
-			   just in case, and make it aligned  */ \
-			sa.sa_restorer = \
-			(void (*)()) (((unsigned int)(cstack) + sizeof(cstack) - 4) & ~3); \
-			sa.sa_flags = SA_RESTART; \
-			sigemptyset(&sa.sa_mask); \
-			sigaddset(&sa.sa_mask, SIG_TIME); \
-			dosemu_sigaction(sig, &sa, NULL);
+	sa.sa_handler = (__sighandler_t) fun; \
+	sa.sa_flags = SA_RESTART; \
+	sigemptyset(&sa.sa_mask); \
+	sigaddset(&sa.sa_mask, SIGALRM); \
+	if (have_working_sigaltstack) { \
+		sa.sa_flags |= SA_ONSTACK; \
+		sigaction(sig, &sa, NULL); \
+	} else { \
+		/* Point to the top of the stack, minus 4 \
+		   just in case, and make it aligned  */ \
+		sa.sa_restorer = \
+		(void (*)(void)) (((unsigned int)(cstack) + sizeof(cstack) - 4) & ~3); \
+		dosemu_sigaction(sig, &sa, NULL); \
+	}
 #endif
 
 EXTERN inline void SIGNAL_save( void (*signal_call)(void) );
@@ -549,11 +561,18 @@ EXTERN int fatalerr INIT(0);
 EXTERN int running_DosC INIT(0);
 EXTERN int dosc_interface(void);
 
+EXTERN void dump_config_status(void *);
 EXTERN void signal_init(void);
 EXTERN void device_init(void);
 EXTERN void hardware_setup(void);
 EXTERN void memory_init(void);
+EXTERN void map_hardware_ram(void);
+EXTERN void map_video_bios(void);
+EXTERN void stdio_init(void);
+EXTERN void time_setting_init(void);
 EXTERN void timer_interrupt_init(void);
+EXTERN void low_mem_init(void);
+EXTERN void print_version(void);
 EXTERN void keyboard_flags_init(void);
 EXTERN void video_config_init(void);
 EXTERN void printer_init(void);
@@ -561,14 +580,16 @@ EXTERN void video_close(void);
 EXTERN void hma_exit(void);
 EXTERN void ems_helper(void);
 EXTERN boolean_t ems_fn(struct vm86_regs *);
-EXTERN void serial_helper(void);
 EXTERN void mouse_helper(void);
 EXTERN void cdrom_helper(void);
 EXTERN void boot(void);
 EXTERN int pkt_int(void);
 EXTERN void read_next_scancode_from_queue (void);
 EXTERN unsigned short detach (void);
+EXTERN void disallocate_vt (void);
+EXTERN void restore_vt (unsigned short vt);
 EXTERN void HMA_init(void);
+EXTERN void HMA_MAP(int HMA);
 
 extern char *Path_cdrom[];
 
