@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -10,6 +10,7 @@
 #ifndef MOUSE_H
 #define MOUSE_H
 
+#include "config.h"
 #include <termios.h>
 
 #define MOUSE_BASE_VERSION	0x0700	/* minimum driver version 7.00 */
@@ -29,6 +30,8 @@
 #define MOUSE_HITACHI 8
 #define MOUSE_X 9
 #define MOUSE_IMPS2 10
+#define MOUSE_XTERM 11
+#define MOUSE_GPM 12
 
 /* types of mouse events */
 #define DELTA_CURSOR		1
@@ -42,17 +45,16 @@
 #define MICKEY			9	/* mickeys per move */
 #define M_DELTA			8
 
-#define MAX_MOUSE 1
 #define HEIGHT 16
 #define PLANES 4
 
 typedef struct  {
-  char dev[255];
+  char *dev;
   int fd;
   int type;
   int flags;
-  boolean add_to_io_select;
   boolean intdrv;
+  boolean use_absolute; /* use absolute mouse addressing */
   boolean emulate3buttons;
   boolean has3buttons;
   boolean cleardtr;
@@ -61,14 +63,14 @@ typedef struct  {
   int lastButtons;
   int chordMiddle;
 
-  struct termios oldset;
+  struct termios *oldset;
 
 } mouse_t;
 
 /* this entire structure gets saved on a driver state save */
 /* therefore we try to keep it small where appropriate */
-/* the 'volatile' is there to cover some bug in gcc -O -g3 */
-EXTERN volatile struct  {
+struct mouse_struct {
+  boolean enabled;
   unsigned char lbutton, mbutton, rbutton;
   unsigned char oldlbutton, oldmbutton, oldrbutton;
 
@@ -106,7 +108,7 @@ EXTERN volatile struct  {
   short virtual_minx, virtual_maxx, virtual_miny, virtual_maxy;
 
   /* these are for sensitivity options */
-  short speed_x, speed_y;
+  short speed_x, speed_y, init_speed_x, init_speed_y;
   short threshold;
   short language;
 
@@ -137,19 +139,42 @@ EXTERN volatile struct  {
     boolean state;
     unsigned short pkg;
   } ps2;
-} mouse;
+};
 
+struct mouse_client {
+  const char *name;
+  int    (*init)(void);
+  void   (*close)(void);
+  void   (*run)(void);         /* handle mouse events */
+  void   (*set_cursor)(int action, int mx, int my, int x_range, int y_range);
+};
+
+extern struct mouse_client *Mouse;
+extern struct mouse_client Mouse_serial;
+extern struct mouse_client Mouse_raw;
+extern struct mouse_client Mouse_X;
+extern struct mouse_client Mouse_xterm;
+extern struct mouse_client Mouse_gpm;
+extern struct mouse_client Mouse_none;
+
+#ifdef HAVE_UNICODE_KEYB
+#include "keyboard.h"
+void mouse_keyboard(Boolean make, t_keysym key);
+#else
 void mouse_keyboard(int);
+#endif
 void mouse_curtick(void);
-void mouse_sethandler(void *, unsigned short *, unsigned short *);
-
-EXTERN mouse_t mice[MAX_MOUSE] ;
+void mouse_sethandler(void);
 
 EXTERN int keyboard_mouse;
 
-extern void mouse_init(void);
-extern void mouse_int(void);
-extern void mouse_close(void);
+extern void dosemu_mouse_init(void);
+extern void mouse_ps2bios(void);
+extern int mouse_int(void);
+extern void dosemu_mouse_close(void);
+extern void parent_close_mouse(void);
+extern int parent_open_mouse(void);
+extern void mouse_post_boot(void);
 extern void int74(void);
 
 extern void mouse_move(void);
@@ -158,11 +183,15 @@ extern void mouse_mb(void);
 extern void mouse_rb(void);
 
 extern void DOSEMUMouseEvents(void);
+extern void DOSEMUSetupMouse(void);
 
 extern void mouse_event(void);
+extern void do_mouse_irq(void);
+extern void mouse_io_callback(void);
 
 extern void mouse_move_buttons(int lbutton, int mbutton, int rbutton);
 extern void mouse_move_relative(int dx, int dy);
 extern void mouse_move_absolute(int x, int y, int x_range, int y_range);
+extern void mouse_reset_to_current_video_mode(void);
 
 #endif /* MOUSE_H */
