@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -31,6 +31,7 @@
 
 #define EMU_NO_IRQ   0xffff
 
+/* Hey, stranger, there is one too many of us in this town! */
 typedef struct {
   Bit8u         (* read_portb)(ioport_t port);
   void          (* write_portb)(ioport_t port, Bit8u byte);
@@ -42,11 +43,22 @@ typedef struct {
   ioport_t      start_addr;
   ioport_t      end_addr;
   int           irq, fd;
-  } emu_iodev_t;
+} emu_iodev_t;
+
+typedef struct {
+  Bit8u  (*read_portb)  (ioport_t port_addr);
+  void   (*write_portb) (ioport_t port_addr, Bit8u byte);
+  Bit16u (*read_portw) (ioport_t port_addr);
+  void   (*write_portw) (ioport_t port_addr, Bit16u word);
+  Bit32u (*read_portd) (ioport_t port_addr);
+  void   (*write_portd) (ioport_t port_addr, Bit32u dword);
+  char   *handler_name;
+  int    irq, fd;
+} _port_handler;
+
 
 static __inline__ void port_real_outb(ioport_t port, Bit8u value)
 {
-  if (!can_do_root_stuff) return;
   __asm__ __volatile__ ("outb %0,%1"
 		    ::"a" ((Bit8u) value), "d"((Bit16u) port));
 }
@@ -54,7 +66,6 @@ static __inline__ void port_real_outb(ioport_t port, Bit8u value)
 static __inline__ Bit8u port_real_inb(ioport_t port)
 {
   Bit8u _v;
-  if (!can_do_root_stuff) return 0xff;
   __asm__ __volatile__ ("inb %1,%0"
 		    :"=a" (_v):"d"((Bit16u) port));
   return _v;
@@ -62,7 +73,6 @@ static __inline__ Bit8u port_real_inb(ioport_t port)
 
 static __inline__ void port_real_outw(ioport_t port, Bit16u value)
 {
-  if (!can_do_root_stuff) return;
   __asm__ __volatile__ ("outw %0,%1" :: "a" ((Bit16u) value),
 		"d" ((Bit16u) port));
 }
@@ -70,14 +80,12 @@ static __inline__ void port_real_outw(ioport_t port, Bit16u value)
 static __inline__ Bit16u port_real_inw(ioport_t port)
 {
   Bit16u _v;
-  if (!can_do_root_stuff) return 0xffff;
   __asm__ __volatile__ ("inw %1,%0":"=a" (_v) : "d" ((Bit16u) port));
   return _v;
 }
 
 static __inline__ void port_real_outd(ioport_t port, Bit32u value)
 {
-  if (!can_do_root_stuff) return;
   __asm__ __volatile__ ("outl %0,%1" : : "a" (value),
 		"d" ((Bit16u) port));
 }
@@ -85,7 +93,6 @@ static __inline__ void port_real_outd(ioport_t port, Bit32u value)
 static __inline__ Bit32u port_real_ind(ioport_t port)
 {
   Bit32u _v;
-  if (!can_do_root_stuff) return 0xffffffff;
   __asm__ __volatile__ ("inl %1,%0":"=a" (_v) : "d" ((Bit16u) port));
   return _v;
 }
@@ -121,15 +128,15 @@ extern int port_rep_outw(ioport_t port, Bit16u *dest, int df, Bit32u count);
 extern int port_rep_ind(ioport_t port, Bit32u *dest, int df, Bit32u count);
 extern int port_rep_outd(ioport_t port, Bit32u *dest, int df, Bit32u count);
 
-#if GLIBC_VERSION_CODE >= 2000
-  /* arrgh... we need this because <sys/io.h> in glibc includes asm/io.h :-( */
-  #undef inb
-  #undef inw
-  #undef ind
-  #undef outb
-  #undef outw
-  #undef outd
-#endif
+extern void port_exit(void);
+
+/* avoid potential clashes with <sys/io.h> */
+#undef inb
+#undef inw
+#undef ind
+#undef outb
+#undef outw
+#undef outd
 #define inb			port_inb
 #define inw			port_inw
 #define ind			port_ind
@@ -146,6 +153,33 @@ extern int port_rep_outd(ioport_t port, Bit32u *dest, int df, Bit32u count);
 extern int extra_port_init(void);
 extern void release_ports(void);
 
+/*
+ * maximum number of emulated devices allowed.  floppy, mda, etc...
+ * an 8-bit handle is used for each device:
+ */
+#define EMU_MAX_IO_DEVICES 254
+
+#define NO_HANDLE	0x00
+#define HANDLE_STD_IO	0x01
+#define HANDLE_STD_RD	0x02
+#define HANDLE_STD_WR	0x03
+#define HANDLE_VID_IO	0x04
+#define HANDLE_SPECIAL	0x05		/* catch-all */
+#define STD_HANDLES	6
+
+extern _port_handler port_handler[];
+/*
+ * Codes:
+ *	00    = no device attached, port undefined
+ *	01-FD = index into handle
+ *	FE,FF = reserved
+ */
+extern unsigned char port_handle_table[];
+extern unsigned char port_andmask[];
+extern unsigned char port_ormask[];
+
+extern pid_t portserver_pid;
+
 #define PORT_FAST	1
 #define PORT_DEV_RD	2
 #define PORT_DEV_WR	4
@@ -157,4 +191,10 @@ extern Boolean port_allow_io(ioport_t, Bit16u, int, Bit8u, Bit8u, unsigned int,
 	char *);
 extern int     set_ioperm(int start, int size, int flag);
 
+extern void init_port_traceing(void);
+extern void register_port_traceing(ioport_t firstport, ioport_t lastport);
+extern void clear_port_traceing(void);
+
+extern void do_r3da_pending (void);
+
 #endif /* _PORT_H */
