@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -26,10 +26,6 @@
 #include "extern.h"
 
 
-extern void set_screen_origin(int), set_vc_screen_page(int);
-
-extern int vc_active(void);
-
 struct screen_stat {
   int console_no,		/* our console number */
    vt_allow,			/* whether to allow VC switches */
