@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -7,6 +7,7 @@
 #ifndef DOSEMU_TIMERS_H
 #define DOSEMU_TIMERS_H
 
+#include "config.h"
 #ifdef __linux__
 #include <sys/param.h>		/* for HZ */
 #endif
@@ -87,10 +88,6 @@ extern void pit_outp(ioport_t, Bit8u);
 #define INT15_IDLE_USECS	(JIFFIE_TIME*8)
 #define INT28_IDLE_USECS	(JIFFIE_TIME/2)
 
-/* --------------------------------------------------------------------- */
-
-#define TIMER_TIME		ITIMER_REAL
-
 /* --------------------------------------------------------------------- */
 /*	New unified timing macros with/without Pentium rdtsc - AV 8/97	 */
 /* --------------------------------------------------------------------- */
@@ -108,29 +105,31 @@ extern void pit_outp(ioport_t, Bit8u);
 static __inline__ hitimer_t _mul64x32_(hitimer_t v, unsigned long f)
 {
 	hitimer_u t;
+	unsigned long v0 = v & 0xffffffff;
+	unsigned long v1 = v >> 32;
+
 
 #ifdef ASM_PEDANTIC
 	unsigned long int d0;
 #endif
 	__asm__ __volatile__ (
 #ifndef ASM_PEDANTIC
-	       "movl	%2,%%eax;"
+		"movl	%2,%%eax\n\t"
 #endif
-	       "mull	%%ebx
-		movl	%%edx,%%ecx
-		movl	%3,%%eax
-		mull	%%ebx
-		addl	%%ecx,%%eax
-		adcl	$0,%%edx
-		"
+		"mull	%%ebx\n\t"
+		"movl	%%edx,%%ecx\n\t"
+		"movl	%3,%%eax\n\t"
+		"mull	%%ebx\n\t"
+		"addl	%%ecx,%%eax\n\t"
+		"adcl	$0,%%edx"
 #ifndef ASM_PEDANTIC
 		: "=a"(t.t.tl),"=d"(t.t.th)
-		: "g"(((int *)&v)[0]),"g"(((int *)&v)[1]),\
+		: "g"(v0),"g"(v1),\
 		  "b"(f)
 		: "%eax","%edx","%ecx","%ebx","memory" );
 #else
 		:"=&a"(t.t.tl),"=&d"(t.t.th), "=&b"(d0)
-		:"g"(((int *)&v)[1]),"2"(f),"0"(((int *)&v)[0])
+		:"g"(v1),"2"(f),"0"(v0)
 		:"%ecx","memory" );
 #endif
 	return t.td;
@@ -152,10 +151,13 @@ static __inline__ hitimer_t _mul64x32_(hitimer_t v, unsigned long f)
 
 /* --------------------------------------------------------------------- */
 
+extern hitimer_u ZeroTimeBase, ZeroTSCBase;
+extern hitimer_t t_vretrace;
+
 EXTERN hitimer_t (*GETcpuTIME)(void) INIT(0);
 
-static inline unsigned long long GETTSC(void) {
-	unsigned long long d;
+static inline hitimer_t GETTSC(void) {
+	hitimer_t d;
 	__asm__ __volatile__ ("rdtsc" : "=A" (d));
 	return d;
 }
@@ -186,6 +188,8 @@ static inline unsigned long long GETTSC(void) {
 	hitimer_u t; \
 	__asm__ __volatile__ (" \
 		rdtsc\n \
+		subl	%3,%%eax\n \
+		sbbl	%4,%%edx\n \
 		movl	%%edx,%%ebx\n \
 		mull	%2\n \
 		movl	%%edx,%%ecx\n \
@@ -193,8 +197,6 @@ static inline unsigned long long GETTSC(void) {
 		mull	%2\n \
 		addl	%%ecx,%%eax\n \
 		adcl	$0,%%edx\n \
-		subl	%3,%%eax\n \
-		sbbl	%4,%%edx\n \
 		" \
 		: "=a"(t.t.tl),"=d"(t.t.th) \
 		: "m"(config.cpu_spd), \
@@ -203,6 +205,25 @@ static inline unsigned long long GETTSC(void) {
 	t.td; \
 })
 
+static inline hitimer_t TSCtoUS(register hitimer_t t) {
+	unsigned long __low, __high;
+	__asm__ ("":"=a"(__low),"=d"(__high):"A"(t));
+	__asm__ __volatile__ (" \
+		movl	%1,%%ebx\n \
+		mull	%2\n \
+		movl	%1,%%ecx\n \
+		movl	%%ebx,%0\n \
+		mull	%2\n \
+		addl	%%ecx,%0\n \
+		adcl	$0,%1\n \
+		" \
+		: "=a"(__low), "=d"(__high) \
+		: "m"(config.cpu_spd) \
+		: "%eax","%edx","%ecx","%ebx","memory" );
+	__asm__ ("":"=A"(t):"a"(__low),"d"(__high));
+	return t;
+}
+
 #else /* ASM_PEDANTIC */
 
 #define CPUtoUS() \
@@ -229,6 +250,8 @@ static inline unsigned long long GETTSC(void) {
 	hitimer_u t; \
 	__asm__ __volatile__ (" \
 		rdtsc\n \
+		subl	%3,%%eax\n \
+		sbbl	%4,%%edx\n \
 		movl	%%edx,%%ebx\n \
 		mull	%2\n \
 		movl	%%edx,%%ecx\n \
@@ -236,8 +259,6 @@ static inline unsigned long long GETTSC(void) {
 		mull	%2\n \
 		addl	%%ecx,%%eax\n \
 		adcl	$0,%%edx\n \
-		subl	%3,%%eax\n \
-		sbbl	%4,%%edx\n \
 		" \
 		: "=&a"(t.t.tl),"=&d"(t.t.th) \
 		: "m"(config.cpu_spd), \
@@ -246,6 +267,25 @@ static inline unsigned long long GETTSC(void) {
 	t.td; \
 })
 
+static inline hitimer_t TSCtoUS(register hitimer_t t) {
+	unsigned long __low, __high;
+	__asm__ ("":"=a"(__low),"=d"(__high):"A"(t));
+	__asm__ __volatile__ (" \
+		movl	%1,%%ebx\n \
+		mull	%2\n \
+		movl	%1,%%ecx\n \
+		movl	%%ebx,%0\n \
+		mull	%2\n \
+		addl	%%ecx,%0\n \
+		adcl	$0,%1\n \
+		" \
+		: "=&a"(__low), "=&d"(__high) \
+		: "m"(config.cpu_spd) \
+		: "%ecx","%ebx","memory" );
+	__asm__ ("":"=A"(t):"a"(__low),"d"(__high));
+	return t;
+}
+
 #endif /* ASM_PEDANTIC */
 
 /* 1 us granularity */
@@ -260,9 +300,11 @@ int restart_cputime (int);
 extern int cpu_time_stop;	/* for dosdebug */
 int bogospeed(unsigned long *spus, unsigned long *sptick);
 
+void freeze_dosemu_manual(void);
 void freeze_dosemu(void);
 void unfreeze_dosemu(void);
 extern int dosemu_frozen;
+extern int dosemu_user_froze;
 
 /* --------------------------------------------------------------------- */
 
@@ -270,5 +312,6 @@ extern Bit8u pit_control_inp(ioport_t);
 extern void pit_control_outp(ioport_t port, Bit8u val);
 extern void initialize_timers(void);
 extern void get_time_init(void);
+extern void do_sound(Bit16u period);
 
 #endif /* DOSEMU_TIMERS_H */
