@@ -1,5 +1,5 @@
 /* 
- * (C) Copyright 1992, ..., 2001 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2004 the "DOSEMU-Development-Team".
  *
  * for details see file COPYING in the DOSEMU distribution
  */
@@ -8,12 +8,12 @@
 #define VIDEO_H
 
 #include "extern.h"
+#include "port.h"
 typedef unsigned char byte;
 
 typedef struct { byte end, start; } cshape;
 
 extern void gettermcap(int);
-extern void scr_state_init(void);
 
 /* if you set this to 1, the video memory dirty bit will be checked 
    before updating the screen.
@@ -34,6 +34,7 @@ extern void scr_state_init(void);
      from 0xa0000 to 0xbffff.
 */
 #define USE_DUALMON 1
+void init_dualmon(void);
 
 #if USE_DUALMON && VIDEO_CHECK_DIRTY
   #error "Currently USE_DUALMON can't be used together with VIDEO_CHECK_DIRTY"
@@ -43,8 +44,13 @@ extern void scr_state_init(void);
 #define CURSOR_END(c)   ((int)((cshape*)&c)->end)
 #define NO_CURSOR 0x0100
 
-#define MAX_COLUMNS 132
-#define MAX_LINES 60
+/* 
+ * xterm allows to set virtually any screen size, but DOS applications
+ * and the BIOS really can't handle more than these numbers of rows and
+ * columns (128*255*2 < 65535)
+ */
+#define MAX_COLUMNS 255
+#define MAX_LINES 128
 
 #if 0
 #define SCREEN_ADR(s)	((ushort *)(virt_text_base + (s*TEXT_SIZE)))
@@ -66,39 +72,6 @@ extern void scr_state_init(void);
 #define ATTR_FG(attr) (attr & 0x0F)
 #define ATTR_BG(attr) (attr >> 4)
 
-/**********************************************************************/
-/* scroll queue */
-
-#define USE_SCROLL_QUEUE 0
-
-#if USE_SCROLL_QUEUE
-
-struct scroll_entry {
-   short x0,y0,x1,y1;
-   short n;
-   byte attr;
-};
-
-#define SQ_MAXLENGTH 5
-
-struct scroll_entry scroll_queue[SQ_MAXLENGTH+1];
-int sq_head,sq_tail;
-
-struct scroll_entry *get_scroll_queue();
-void clear_scroll_queue();
-
-extern int video_update_lock;
-
-#define VIDEO_UPDATE_LOCK() video_update_lock++;
-#define VIDEO_UPDATE_UNLOCK() video_update_lock--;
-
-#else
-#define video_update_lock 0
-#define VIDEO_UPDATE_LOCK()
-#define VIDEO_UPDATE_UNLOCK()
-#define clear_scroll_queue()
-#endif
-
 /***********************************************************************/
 
 /* Here's an idea to clean up the video code: Build a 'virtual' video
@@ -139,10 +112,14 @@ struct video_system {
    void (*update_cursor)(void);    /* update cursor position&shape. Called by sigalrm
                                   handler *only* if update_screen does not exist
                                   or is not done because the video mem is clean */
+   int (*change_config)(unsigned item, void *buf); /* change configuration
+                                  e.g. window title (optional) */
 };
 
 extern struct video_system *Video;
-
+#if USE_DUALMON
+extern struct video_system *Video_default;
+#endif
 extern struct video_system Video_graphics,  Video_X, Video_console, Video_hgc, Video_term;
 
 EXTERN ushort *screen_adr;   /* pointer to video memory of current page */
@@ -172,8 +149,6 @@ EXTERN int text_scanlines INIT(400);   /* # of scan lines in textmodes */
                              /* these have effect only on video mode sets! */
 
 EXTERN unsigned char video_initialized INIT(0);
-extern int vga_initialize(void);
-extern void mda_initialize(void);
 extern void install_int_10_handler(void);
 extern void clear_screen(int s,int att);
 extern boolean set_video_mode(int);
@@ -200,6 +175,7 @@ EXTERN int v_8514_base INIT(0);
 #define CHARSET_FULLIBM	3
 #define CHARSET_LATIN1  4
 #define CHARSET_LATIN2  5
+#define CHARSET_KOI8    6
 
 
 /* Color set defines */
@@ -262,7 +238,24 @@ EXTERN int v_8514_base INIT(0);
 #define MATROX		8
 #define WDVGA		9
 #define SIS		10
-#define MAX_CARDTYPE	SIS
+#define SVGALIB	11
+#define MAX_CARDTYPE	SVGALIB
+
+/* title and change config definitions */
+#define TITLE_EMUNAME_MAXLEN 128
+#define TITLE_APPNAME_MAXLEN 25
+
+#define CHG_TITLE	1
+#define CHG_FONT	2
+#define CHG_MAP		3
+#define CHG_UNMAP	4
+#define CHG_WINSIZE	5
+#define CHG_TITLE_EMUNAME	6
+#define CHG_TITLE_APPNAME	7
+#define CHG_TITLE_SHOW_APPNAME	8
+#define CHG_BACKGROUND_PAUSE	9
+#define GET_TITLE_APPNAME	10
+#define CHG_FULLSCREEN	11
 
 EXTERN void get_screen_size (void);
 EXTERN void set_video_bios_size(void);
@@ -271,6 +264,12 @@ EXTERN void X_blink_cursor(void);
 EXTERN void init_vga_card(void);
 EXTERN void scr_state_init(void);
 
+extern void set_console_video(void);
+extern void clear_console_video(void);
+extern inline void console_update_cursor (int, int, int, int);
+extern int load_file(char *name, int foffset, char *mstart, int msize);
+
+
 /* moved here from s3.c --AV  */
 #define BASE_8514_1	0x2e8
 #define BASE_8514_2	0x148
